/*
 * @(#) Layout.java	1.4 98/09/04
 *
 * Copyright (c) 1997,1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package pnuts.awt;

import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.util.Hashtable;
import java.util.Stack;
import java.awt.*;
import java.io.*;

/**
 * Manager class of <a href="/pnuts/doc/hierarchicalLayout.html">Hierarchical Layout</a>.
 * @version	1.4
 * @author	Toyokazu Tomatsu
 */
public abstract class Layout {
    private static Hashtable mappingTable = new Hashtable();
    private final static String swingPkgNames[] = {"com.sun.java.swing",
						   "javax.swing"};

    static {
	// default mapping
	registerLayoutManager(BorderLayout.class, BorderLayoutMapping.class);
	registerLayoutManager(CardLayout.class, CardLayoutMapping.class);
	registerLayoutManager(GridLayout.class, GridLayoutMapping.class);
	registerLayoutManager(GridBagLayout.class, GridBagLayoutMapping.class);
	registerLayoutManager(FlowLayout.class, FlowLayoutMapping.class);
	registerLayoutManager(PnutsLayout.class, PnutsLayoutMapping.class);
    }

    private static Class jpanelClass = null;
    private static boolean jPanelFlg = false;

    private static Class findJPanelClass(String pkgName){
	if (!jPanelFlg){
	    jPanelFlg = true;
	    try {
		jpanelClass = Class.forName(pkgName + ".JPanel");
		return jpanelClass;
	    } catch (ClassNotFoundException e){ /* ignore */ }
	}
	return jpanelClass;
    }

    /**
     * Register a Layout Mapping
     *
     * @param	clazz	Class of the mapping
     * @param	layout	The definition of the mapping
     *                  This class should be subclass of Layout class.
     */
    public static void registerLayoutManager(Class clazz, Class layout){
	if (Layout.class.isAssignableFrom(layout)){
	    mappingTable.put(clazz, layout);
	}
    }

    protected static boolean isArray(Object obj){
	return obj.getClass().isArray();
    }

    protected Layout(){}

    /**
     * Define how to make a container.
     * This class should be defined in a subclass of Layout class.
     */
    public abstract Container createContainer(Container container, Object[] fmt);

    /**
     * Layout components using format
     */
    public static Container layout(Object[] format){
	return layout(new Panel(), format);
    }

    /**
     * Layout components in the <em>container</em> using format fmt.
     */
    public static Container layout(Container container, Object[] fmt){
	if (fmt[0] instanceof Class){
	    Layout layout = null;
	    try {
		Object map = mappingTable.get(fmt[0]);
		if (map instanceof Layout){
		    return ((Layout)map).createContainer(container, fmt);
		} else if (map instanceof Class){
		    layout = (Layout)((Class)map).newInstance();
		    mappingTable.put(fmt[0], layout);
		    return layout.createContainer(container, fmt);
		} else {
		    throw new LayoutException("mapping not defined: " + fmt[0]);
		}
	    } catch (Exception e){
		throw new LayoutException(e.getMessage());
	    }
	} else {
	    throw new LayoutException("illegal format");
	}
    }

    protected static Container makePanel(Container prototype) {
	try {
	    if (prototype == null){
		return new Panel();
	    }
	    Class cl = prototype.getClass();
	    int guiType = 0; // 0:AWT, 1:com.sun.java.swing, 2: javax.swing
	    String name = cl.getName();
	    if (name.startsWith(swingPkgNames[0])){
		guiType = 1;
	    } else if (name.startsWith(swingPkgNames[1])){
		guiType = 2;
	    }

	    if (prototype instanceof Window){
		if (guiType == 1 || guiType == 2){
		    cl = findJPanelClass(swingPkgNames[guiType]);
		    if (cl == null){
			cl = Panel.class;
		    }
		} else {
		    cl = Panel.class;
		}
	    }
	    Container ret = (Container)cl.newInstance();
	    ret.setForeground(prototype.getForeground());
	    ret.setBackground(prototype.getBackground());
	    ret.setCursor(prototype.getCursor());
	    ret.setFont(prototype.getFont());
	    if (guiType == 1){
		Method m = null;
		Class t_noarg[] = new Class[]{};
		Object noarg[] = new Object[]{};
		Class t_boolean[] = new Class[]{Boolean.TYPE};
		Class t_string[] = new Class[]{String.class};
		Class clazz = prototype.getClass();
		Object val = null;

		m = clazz.getMethod("isDoubleBuffered", t_noarg);
		val = m.invoke(prototype, noarg);
		m = cl.getMethod("setDoubleBuffered", t_boolean);
		m.invoke(ret, new Object[]{val});

		m = clazz.getMethod("getToolTipText", t_noarg);
		val = m.invoke(prototype, noarg);
		m = cl.getMethod("setToolTipText", t_string);
		m.invoke(ret, new Object[]{val});

		m = clazz.getMethod("isOpaque", t_noarg);
		val = m.invoke(prototype, noarg);
		m = cl.getMethod("setOpaque", t_boolean);
		m.invoke(ret, new Object[]{val});

		m = clazz.getMethod("getAutoscrolls", t_noarg);
		val = m.invoke(prototype, noarg);
		m = cl.getMethod("setAutoscrolls", t_boolean);
		m.invoke(ret, new Object[]{val});
	    }
	    return ret;
	} catch (Exception e){
	    return new Panel();
	}
    }
}
