/*
 * @(#) Complex.java	1.4 99/03/30
 *
 * Copyright (c) 1997,1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package pnuts.ext;

import pnuts.lang.Numeric;
import pnuts.lang.Arithmetic;
import pnuts.lang.Pnuts;
import pnuts.lang.QuantityFactory;

/**
 * An implementation of complex number.
 * This class is an examples of pnuts.lang.Numeric and
 * pnuts.lang.QuantityFactory.
 *
 * A sample script to use this class is
 * <a href="/pnuts/examples/numeric/complex.pnut">here</a>.
 *
 * @version	1.4
 * @author	Toyokazu Tomatsu
 */
public class Complex extends Number implements Numeric {

    final static Integer zero = new Integer(0);
    static {
	Pnuts.registerQuantityFactory("i", new QuantityFactory(){
	    public Object make(Number n, String unit){
		return new Complex(new Integer(0), n);
	    }
	});
    }

    /**
     * @serial
     */
    Number re;

    /**
     * @serial
     */
    Number im;

    public Complex(Number re, Number im){
	this.re = re;
	this.im = im;
    }

    static Object make(Number re, Number im){
	if (Arithmetic.compareTo(im, zero) == Arithmetic.EQUAL){
	    return re;
	} else {
	    return new Complex(re, im);
	}
    }

    public Number getReal(){
	return re;
    }

    public Number getImaginary(){
	return im;
    }

    public Object add(Object o){
	if (o instanceof Complex){
	    Complex c = (Complex)o;
	    return make(add(re, c.re), add(im, c.im));
	} else if (o instanceof Numeric){
	    Numeric n = (Numeric)o;
	    return make((Number)n.add(re), im);
	} else if (o instanceof Number){
	    return make(add(re, (Number)o), im);
	}
	throw new RuntimeException("illegal operand type: " + o.getClass());
    }

    public Object subtract(Object o){
	if (o instanceof Complex){
	    Complex c = (Complex)o;
	    return make(subtract(re, c.re), subtract(im, c.im));
	} else if (o instanceof Numeric){
	    return add(((Numeric)o).negate());
	} else if (o instanceof Number){
	    return make(subtract(re, (Number)o), im);
	}
	throw new RuntimeException("illegal operand type: " + o.getClass());
    }

    /**
     * @serial
     */
    boolean reverse;

    public Object multiply(Object o){
	if (o instanceof Complex){
	    Complex c = (Complex)o;
	    return make(subtract(multiply(re, c.re),
				 multiply(im, c.im)),
			add(multiply(re, c.im),
			    multiply(im, c.re)));
	} else if (o instanceof Numeric){
	    Numeric n = (Numeric)o;
	    if (!reverse){
		reverse = true;
		return n.multiply(this);
	    } else {
		reverse = false;
		return make((Number)n.multiply(re), (Number)n.multiply(im));
	    }
	} else if (o instanceof Number){
	    return make(multiply(re, (Number)o),
			multiply(im, (Number)o));
	}
	throw new RuntimeException("illegal operand type: " + o.getClass());
    }

    public Object divide(Object o){
	if (o instanceof Complex){
	    Complex c = (Complex)o;
	    Number de = add(multiply(c.re, c.re),
			    multiply(c.im, c.im));
	    return make(divide(add(multiply(re, c.re),
				   multiply(im, c.im)),
			       de),
			divide(subtract(multiply(im, c.re),
					multiply(re, c.im)),
			       de));
	} else if (o instanceof Numeric){
	    return multiply(((Numeric)o).inverse());
	} else if (o instanceof Number){
	    return make(divide(re, (Number)o),
			divide(im, (Number)o));
	}
	throw new RuntimeException("illegal operand type: " + o.getClass());
    }

    public Object negate(){
	return new Complex((Number)Arithmetic.negate(re), (Number)Arithmetic.negate(im));
    }

    public Object inverse(){
	Number de = (Number)subtract(multiply(re, re),
				     multiply(im, im));
	return make(divide(re, de),
		    (Number)Arithmetic.negate(divide(im, de)));
    }

    public int compareTo(Object o){
	throw new RuntimeException("not applicable");
    }

    public int intValue(){
	return re.intValue();
    }

    public long longValue(){
	return re.longValue();
    }

    public float floatValue(){
	return re.floatValue();
    }

    public double doubleValue(){
	return re.doubleValue();
    }

    static Number subtract(Number n1, Number n2){
	return (Number)Arithmetic.subtract(n1, n2);
    }

    static Number add(Number n1, Number n2){
	return (Number)Arithmetic.add(n1, n2);
    }

    static Number multiply(Number n1, Number n2){
	return (Number)Arithmetic.multiply(n1, n2);
    }

    static Number divide(Number n1, Number n2){
	return (Number)Arithmetic.divide(n1, n2);
    }

    public String toString(){
	if (Arithmetic.compareTo(im, zero) == Arithmetic.RIGHT_IS_BIGGER){
	    return "(" + re + im + "i)";
	} else {
	    return "(" + re + "+" + im + "i)";
	}
    }
}
