/*
 * @(#) DynamicClass.java	1.8 99/06/03
 *
 * Copyright (c) 1997,1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package pnuts.ext;

import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;
import pnuts.lang.Pnuts;
import pnuts.lang.Package;
import pnuts.lang.PnutsFunction;
import pnuts.lang.Context;
import pnuts.lang.AbstractData;
import java.io.Serializable;
import java.io.ObjectOutputStream;
import java.io.ObjectInputStream;
import java.io.IOException;

/**
 * An implementation of class-based inheritance.
 * See <a href="/pnuts/doc/class-oo.html">Pnuts User's Guide</a> for details.
 *
 * @see		pnuts.lang.AbstractData
 * @author	Toyokazu Tomatsu
 * @version	1.8
 */
public class DynamicClass extends PnutsFunction implements Serializable {

    protected static Hashtable typeSpace = new Hashtable(32);
    protected static PnutsFunction nop = (PnutsFunction)Pnuts.eval("function () null");
    protected static Hashtable primitiveTypes = new Hashtable();
    static {
	primitiveTypes.put(Integer.TYPE, Integer.class);
	primitiveTypes.put(Short.TYPE, Short.class);
	primitiveTypes.put(Byte.TYPE, Byte.class);
	primitiveTypes.put(Character.TYPE, Character.class);
	primitiveTypes.put(Long.TYPE, Long.class);
 	primitiveTypes.put(Float.TYPE, Float.class);
 	primitiveTypes.put(Double.TYPE, Double.class);
 	primitiveTypes.put(Boolean.TYPE, Boolean.class);
    }

    /**
     * @serial
     */
    DynamicClass superType;

    /**
     * @serial
     */
    String name;

    /**
     * @serial
     */
    Hashtable methods;

    /**
     * @serial
     */
    Vector NULL = new Vector(0);

    Hashtable attributes;

    protected DynamicClass(String name){
	super(name);
    }

    /**
     * Declare a type which behaves as a "class".
     * It inherits attributes and methods from superType
     */
    public static DynamicClass declare(String name, DynamicClass superType){
	DynamicClass t = (DynamicClass)typeSpace.get(name);
	if (t == null){
	    t = new DynamicClass(name);
	    typeSpace.put(name, t);
	}
	t.name = name;
	t.superType = superType;
	t.attributes = new Hashtable(32);
	t.methods = new Hashtable(32);
	
	t.method("new", nop);
	return t;
    }

    /**
     * get the supertype
     */
    public DynamicClass getSuperType(){
	return superType;
    }

    /**
     * get the name of type
     */
    public String getName(){
	return name;
    }

    public Hashtable getAttributes(){
	Hashtable a = (Hashtable)attributes.clone();
	DynamicClass t = superType;
	while (t != null){
	    for (Enumeration e = t.attributes.keys(); e.hasMoreElements(); ){
		Object next = e.nextElement();
		if (a.get(next) == null){
		    a.put(next, t.attributes.get(next));
		}
	    }
	    t = t.superType;
	}
	return a;
    }

    public Enumeration getMethods(){
	Hashtable a = (Hashtable)methods.clone();
	DynamicClass t = superType;
	while (t != null){
	    for (Enumeration e = t.methods.keys(); e.hasMoreElements(); ){
		Object next = e.nextElement();
		if (a.get(next) == null){
		    a.put(next, t.methods.get(next));
		}
	    }
	    t = t.superType;
	}
	return a.keys();
    }

    /**
     * @deprecated As of Pnuts 1.0beta1
     */
    public void add(String name, Class clazz){
	var(name, clazz);
    }

    /**
     * @deprecated As of Pnuts 1.0beta1
     */
    public void add(String name, DynamicClass type){
	var(name, type);
    }

    /**
     * @deprecated As of Pnuts 1.0beta1
     */
    public void add(String name){
	var(name);
    }

    /**
     * declare an attribute of the specified class
     */
    public void var(String name, Class clazz){
	attributes.put(name, clazz);
    }

    /**
     * declare an attribute of the specified type
     */
    public void var(String name, DynamicClass type){
	attributes.put(name, type);
    }

    /**
     * declare an attribute of the Object class
     */
    public void var(String name){
	var(name, Object.class);
    }

    /**
     * remove an attribute
     */
    public void removeVar(String name){
	attributes.remove(name);
    }

    public void remove(String name){
    }

    static int minArgs(PnutsFunction func){
	int i = 0;
	for (; i < Integer.MAX_VALUE; i++){
	    if (func.defined(i)){
		return i;
	    }
	}
	return -1;
    }

    static String packageStatement(Package pkg){
	return "package(\"" + pkg.getName() + "\")\n";
    }

    static String importStatements(Object imp[]){
	String s = "";
	for (int i = imp.length - 1; i >= 0; i--){
	    if (imp[i] == null){
		s += "import(null)\n";
	    } else if (imp[i] instanceof Class){
		s += "import(\"" + ((Class)imp[i]).getName() + "\")\n";
	    } else {
		s += "import(\"" + imp[i] + "\")\n";
	    }
	}
	return s;
    }

    /**
     * define a method
     */
    public void method(String name, PnutsFunction func){
	int narg = minArgs(func);

	String expr = packageStatement(func.getPackage(narg)) +
	    importStatements(func.getImportEnv(narg)) +
	    "function (this, super) " + func.unparse(narg);

	PnutsFunction f = (PnutsFunction)Pnuts.eval(expr);
	Vector vec = (Vector)methods.get(name);
	if (vec == null){
	    vec = new Vector();
	    methods.put(name, vec);
	}
	if (vec.size() < narg + 1){
	    vec.setSize(narg + 1);
	}
	vec.setElementAt(f, narg);
    }

    /**
     * remove a method
     */
    public void removeMethod(String name){
	methods.remove(name);
    }


    PnutsFunction getMethod(String name, int narg, Instance instance){
	return getMethod(name, narg, instance, this);
    }

    PnutsFunction getMethod(String name, int narg, Instance instance, DynamicClass start){
	DynamicClass t = start;
	while (t != null){
	    Vector v = (Vector)t.methods.get(name);
	    if (v != null){
		PnutsFunction c = (PnutsFunction)v.elementAt(narg);
		if (c != null){
		    return (PnutsFunction)c.call(new Object[]{instance, new Super(instance)});
		}
	    }
	    t = t.superType;
	}
	return null;
    }

    public Object attributeType(String name){
	DynamicClass t = this;
	while (t != null){
	    Object c = t.attributes.get(name);
	    if (c != null){
		return c;
	    }
	    t = t.superType;
	}
	return null;
    }

    boolean isAssignable(String name, Object obj){
	Object t = attributeType(name);
	if (t == null){
	    return false;
	}
	if (obj instanceof Instance){
	    DynamicClass type = ((Instance)obj).getType();
	    if (t instanceof DynamicClass){
		return type.isSubTypeOf((DynamicClass)t);
	    } else if (t instanceof Class){
		return ((Class)t).isAssignableFrom(obj.getClass());
	    } else {
		return false;
	    }
	} else if (obj != null){
	    if (t instanceof Class){
		Class c = (Class)t;
		if (c.isPrimitive()){
		    c = (Class)primitiveTypes.get(c);
		}
		return c.isAssignableFrom(obj.getClass());
	    } else {
		return false;
	    }
	} else {
	    return true;
	}
    }

    /**
     * check if the type is subtype of the specified type
     */
    public boolean isSubTypeOf(DynamicClass type){
	DynamicClass t = this;
	while (t != null){
	    if (t == type){
		return true;
	    }
	    t = t.superType;
	}
	return false;
    }

    protected Object exec(Object[] args, Context context){
	Instance instance = new Instance();
	PnutsFunction constructor = getMethod("new", args.length, instance);
	if (constructor != null){
	    constructor.call(args, context);
	}
	return instance;
    }

    public String toString(){
	return "class " + name;
    }

    public class Instance implements AbstractData, Serializable {
	Hashtable attr = new Hashtable(32);

	/**
	 * assign a value to the specified variable
	 */
	public void set(String name, Object obj){
	    if (attributeType(name) == null){
		throw new RuntimeException("no such field in " + DynamicClass.this + ": " + name);
	    }
	    if (!isAssignable(name, obj)){
		throw new RuntimeException("can not assign: " + attributeType(name));
	    }
	    if (obj != null){
		attr.put(name, obj);
	    } else {
		attr.put(name, NULL);
	    }
	}

	/**
	 * get a value of the specified variable
	 */
	public Object get(String name){
	    if (attributeType(name) == null){
		throw new RuntimeException("no such field in " + DynamicClass.this + ": " + name);
	    }
	    Object o = attr.get(name);
	    if (o == NULL){
		return null;
	    } else {
		return o;
	    }
	}

	/**
	 * call a method
	 */
	public Object invoke(String name, Object[] args){
	    if ("instanceOf".equals(name)){
		if (args[0] instanceof DynamicClass){
		    return new Boolean(instanceOf((DynamicClass)args[0]));
		} else {
		    return Boolean.FALSE;
		}
	    }
	    if ("getType".equals(name)){
		return getType();
	    }
	    if ("getClass".equals(name)){
		return getClass();
	    }
	    PnutsFunction func = (PnutsFunction)DynamicClass.this.getMethod(name, args.length, this);
	    if (func != null){
		return func.call(args);
	    }
	    throw new RuntimeException("no such method:" + name);
	}

	/**
	 * get its type
	 */
	DynamicClass getType(){
	    return DynamicClass.this;
	}

	boolean instanceOf(DynamicClass type){
	    return getType().isSubTypeOf(type);
	}

	public String toString(){
	    return getType().getName() + "@" + hashCode();
	}
    }

    class Super extends Instance {
	Instance instance;

	Super(Instance instance){
	    this.instance = instance;
	}

	public Object invoke(String name, Object[] args){
	    DynamicClass s = instance.getType().getSuperType();
	    PnutsFunction func = (PnutsFunction)getMethod(name, args.length, instance, s);
	    if (func != null){
		return func.call(args);
	    }
	    throw new RuntimeException("no such method:" + name);
	}
    }
}
