/*
 * @(#) Fraction.java	1.3 99/03/30
 *
 * Copyright (c) 1997,1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package pnuts.ext;

import java.math.BigInteger;
import java.math.BigDecimal;
import pnuts.lang.Numeric;
import pnuts.lang.Arithmetic;

/**
 * An implementation of fractional number.
 * This class is an examples of pnuts.lang.Numeric and
 * pnuts.lang.QuantityFactory.
 *
 * A sample script to use this class is
 * <a href="/pnuts/examples/numeric/fraction.pnut">here</a>.
 *
 */
public class Fraction extends Number implements Numeric {

    final static int ADD = 0;
    final static int SUB = 1;
    final static int MUL = 2;
    final static int DIV = 3;
    final static int COM = 4;
    final static int NEG = 5;
    final static int INV = 6;
    final static Integer zero = new Integer(0);

    /**
     * @serial
     */
    Number denominator;

    /**
     * @serial
     */
    Number numerator;

    protected Fraction(int num, int den){
	this(new Integer(num), new Integer(den));
    }

    protected Fraction(Number num, Number den){
	numerator = num;
	denominator = den;
    }

    public static Number make(int num, int den){
	return make(new Integer(num), new Integer(den));
    }

    public static Number make(Number num, Number den){
	if (Arithmetic.compareTo(den, zero) == Arithmetic.EQUAL){
	    throw new RuntimeException("denominator is zero");
	}
	Number g = (Number)Arithmetic.gcd(num, den);
	if (Arithmetic.compareTo(g, zero) == Arithmetic.RIGHT_IS_BIGGER){
	    g = (Number)Arithmetic.negate(g);
	}
	if (Arithmetic.compareTo(den, zero) == Arithmetic.RIGHT_IS_BIGGER){
	    num = (Number)Arithmetic.negate(num);
	    den = (Number)Arithmetic.negate(den);
	}
	if (g.intValue() != 1){
	    num = divide(num, g);
	    den = divide(den, g);
	}
	if (den.intValue() == 1){
	    return num;
	} else {
	    return new Fraction(num, den);
	}
    }

    public Number getDenominator(){
	return denominator;
    }

    public Number getNumerator(){
	return numerator;
    }

    public int intValue(){
	return (int)longValue();
    }

    public long longValue(){
	return (long)doubleValue();
    }

    public float floatValue(){
	return (float)doubleValue();
    }

    public double doubleValue(){
	return numerator.doubleValue() / denominator.doubleValue();
    }

    static Object op1(Number n1, Number d1, Number n2, Number d2, int op){
	switch (op){
	case ADD:
	    return make(add(multiply(n1, d2),
			    multiply(n2, d1)),
		        multiply(d1, d2));
	case SUB:
	    return make(subtract(multiply(n1, d2),
				 multiply(n2, d1)),
		        multiply(d1, d2));
	case MUL:
	    return make(multiply(n1, n2),
		        multiply(d1, d2));
	case DIV:
	    return make(multiply(n1, d2),
		        multiply(d1, n2));
	}
	return null;
    }

    static Object op2(Number n1, Number d1, Number n2, int op){
	switch (op){
	case ADD:
	    return make(add(n1, multiply(d1, n2)), d1);
	case SUB:
	    return make(subtract(n1, multiply(d1, n2)), d1);
	case MUL:
	    return make(multiply(n1, n2), d1);
	case DIV:
	    return make(n1, multiply(d1, n2));
	}
	return null;
    }

    static Object op3(Number n1, Number d1, Number n2, int op){
	switch (op){
	case ADD:
	    return divide(add(n1, multiply(d1, n2)), d1);
	case SUB:
	    return divide(subtract(n1, multiply(d1, n2)), d1);
	case MUL:
	    return divide(multiply(n1, n2), d1);
	case DIV:
	    return divide(n1, multiply(d1, n2));
	}
	return null;
    }

    Object op(Object o, int type){
	if (o instanceof Fraction){
	    Fraction f = (Fraction)o;
	    Number n = f.getNumerator();
	    Number d = f.getDenominator();
	    return op1(numerator, denominator, n, d, type);
	} else if ((o instanceof Integer) ||
		   (o instanceof Short) ||
		   (o instanceof Long) ||
		   (o instanceof BigInteger)){
	    return op2(numerator, denominator, (Number)o, type);
	} else if ((o instanceof Float) ||
		   (o instanceof Double)){
	    return op3(numerator, denominator, (Number)o, type);
	} else if (o instanceof BigDecimal){
	    BigDecimal d = (BigDecimal)o;
	    if (d.scale() == 0){
		return op2(numerator, denominator, d.toBigInteger(), type);
	    } else {
		return op3(numerator, denominator, (Number)o, type);
	    }
	} else if (o instanceof Numeric){
	    return op4((Numeric)o, type);
	}
	throw new RuntimeException("illegal operand type: " + o.getClass());
    }

    Object op4(Numeric n, int op){
	switch (op){
	case ADD:
	    return n.add(this);
	case SUB:
	    return add(n.negate());
	case MUL:
	    return n.multiply(this);
	case DIV:
	    return multiply(n.inverse());
	}
	return null;
    }

    public Object add(Object o){
	return op(o, ADD);
    }

    public Object subtract(Object o){
	return op(o, SUB);
    }

    public Object multiply(Object o){
	return op(o, MUL);
    }

    public Object divide(Object o){
	return op(o, DIV);
    }

    public Object negate(){
	return make((Number)Arithmetic.negate(numerator), denominator);
    }

    public Object inverse(){
	return make(denominator, numerator);
    }

    public int compareTo(Object o){
	if (o instanceof Fraction){
	    Number n = (Number)subtract(o);
	    if (n instanceof Fraction){
		Fraction f = (Fraction)n;
		return Arithmetic.compareTo(f.numerator, zero);
	    } else {
		return Arithmetic.compareTo(n, zero);
	    }
	} else if (o instanceof Number){

	}
	Number n = (Number)subtract(o);
	if (n instanceof Fraction){
	    Fraction f = (Fraction)n;
	    return Arithmetic.compareTo(f.numerator, zero);
	} else if (n instanceof Number){
	    return Arithmetic.compareTo(n, zero);
	}
	throw new RuntimeException("illegal type for compare operation");
    }

    static Number subtract(Number n1, Number n2){
	return (Number)Arithmetic.subtract(n1, n2);
    }

    static Number add(Number n1, Number n2){
	return (Number)Arithmetic.add(n1, n2);
    }

    static Number multiply(Number n1, Number n2){
	return (Number)Arithmetic.multiply(n1, n2);
    }

    static Number divide(Number n1, Number n2){
	return (Number)Arithmetic.divide(n1, n2);
    }

    public String toString(){
	return numerator + "/" + denominator;
    }
}
