/*
 * @(#) JarClassLoader.java	1.3 98/09/04
 *
 * Copyright (c) 1997,1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package pnuts.ext;

import java.net.URL;
import java.net.MalformedURLException;
import java.io.InputStream;
import java.io.ByteArrayOutputStream;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.EOFException;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Properties;
import java.util.zip.ZipFile;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

/**
 * This classloader can read JAR files dynamically.
 * <a href="/pnuts/examples/classloader/jarClassLoader.pnut">jarClassLoader.pnut</a> is a sample script in Pnuts.
 *
 * Besides, after this class is loaded, jar URL can be used as defined in JDK1.2. 
 * For example, URL("jar:http://www.etale.com/pnuts/pnuts.jar!/lib/init.pnut").
 *
 * @version	1.3
 * @author	Toyokazu Tomatsu
 */
public class JarClassLoader extends ClassLoader {
    public Hashtable classes = new Hashtable();
    public Hashtable resources = new Hashtable();
    private static byte[] buf = new byte[1024];
    private static ByteArrayOutputStream bout = new ByteArrayOutputStream();
    private static String resourceProtocolName = "jar";

    static Hashtable loaders = new Hashtable();

    public static JarClassLoader getLoader(Object key){
	return (JarClassLoader)loaders.get(key);
    }

    /**
     * Load a JAR file from the specified URL.
     */
    public void loadJar(URL url) throws IOException {
	String f = url.getFile();
	if (f.startsWith("file:")){
	    loadJar(f.substring(5));
	} else {
	    loaders.put(url.toString(), this);
	    ZipEntry entry = null;
	    ZipInputStream zin = new ZipInputStream(url.openStream());
	    int idx = 0;
	    while ((entry = zin.getNextEntry()) != null){
		String name = entry.getName();
		if (name.startsWith("META-INF/")){
		} else if (name.endsWith(".class")){
		    classes.put(name, new UrlEntry(url, idx));
		} else {
		    if (name.charAt(0) == '/'){
			name = name.substring(1);
		    }
		    resources.put(name, new UrlEntry(url, idx));
		}
		++idx;
	    }
	}
    }

    /**
     * Load a JAR file through ZipInputStream. The contents of entries
     * are kept in a hashtable (, so this operation is heavy weight).
     * When JAR file is randomly accessible, use loadJar(String filename).
     */
    public void loadJar(ZipInputStream zin) throws IOException {
	ZipEntry entry = null;
	while ((entry = zin.getNextEntry()) != null){
	    String name = entry.getName();
	    int n = 0;
	    if (name.startsWith("META-INF/")){
	    } else if (name.endsWith(".class")){
		bout.reset();
		while ((n = zin.read(buf, 0, buf.length)) > 0){
		    bout.write(buf, 0, n);
		}
		classes.put(name, bout.toByteArray());
	    } else {
		bout.reset();
		while ((n = zin.read(buf, 0, buf.length)) > 0){
		    bout.write(buf, 0, n);
		}
		if (name.charAt(0) == '/'){
		    name = name.substring(1);
		}
		resources.put(name, bout.toByteArray());
	    }
	}
    }

    static class UrlEntry {
	URL url;
	int index;

	UrlEntry(URL url, int index){
	    this.url = url;
	    this.index = index;
	}

	byte[] getData(){
	    try {
		ZipInputStream zin = new ZipInputStream(url.openStream());
		zin.getNextEntry();
		for (int i = 0; i < index; i++){
		    zin.getNextEntry();
		}
		bout.reset();
		int n;
		while ((n = zin.read(buf, 0, buf.length)) > 0){
		    bout.write(buf, 0, n);
		}
		return bout.toByteArray();
	    } catch (IOException e){
		return null;
	    }
	}
    }

    /**
     * Load a JAR file specified with the filename.  The corresponding
     * ZipFile object is kept in a hashtable.
     */
    public void loadJar(String filename) throws IOException {
	loadJar(new ZipFile(filename));
    }

    protected void loadJar(ZipFile file) throws IOException {
	loaders.put("file:" + file.getName(), this);
	ZipEntry entry = null;
	Enumeration e = file.entries();
	while (e.hasMoreElements()){
	    entry = (ZipEntry)e.nextElement();
	    String name = entry.getName();
	    if (name.startsWith("META-INF/")){
	    } else if (name.endsWith(".class")){
		classes.put(name, file);
	    } else {
		if (name.charAt(0) == '/'){
		    name = name.substring(1);
		}
		resources.put(name, file);
	    }
	}
    }

    /**
     * Resolves the specified name to a Class. The method loadClass() is 
     * called by the virtual machine.
     * <p>
     * If the <code>resolve</code> flag is true, the method should call 
     * the <code>resolveClass</code> method on the resulting class object.
     * <p>
     * @param	   name     the name of the desired Class.
     * @param      resolve  true if the Class needs to be resolved.
     * @return	   the resulting Class, or null if it was not found.
     * @exception  ClassNotFoundException  if the class loader cannot find
     *               a definition for the class.
     */
    public Class loadClass(String name, boolean resolve)  throws ClassNotFoundException{
	Class clazz = null;
	try {
	    return super.findSystemClass(name);
	} catch (ClassNotFoundException e1){
	    /* ignore */
	} catch (NoClassDefFoundError e2){
	    /* ignore */
	}
	clazz = findLoadedClass(name);
	if (clazz != null){
	    return clazz;
	}
	String str = name.replace('.', '/') + ".class";
	Object o = classes.get(str);
	if (o == null){
	    throw new ClassNotFoundException();
	}
	byte[] b = null;
	if (o instanceof ZipFile){
	    b = getData((ZipFile)o, str);
	} else if (o instanceof UrlEntry){
	    b = ((UrlEntry)o).getData();
	} else {
	    b = (byte[])o;
	}
	if (b == null){
	    throw new ClassFormatError(str);
	}
	clazz = defineClass(name, b, 0, b.length);
	if (clazz == null){
	    throw new ClassFormatError();
	}
	classes.remove(str);
	if (resolve){
	    resolveClass(clazz);
	}
	return clazz;
    }

    /**
     * URL scheme for a resource in a JAR file is:
     */
    public URL getResource(String name) {
	if (name.startsWith("/")){
	    name = name.substring(1);
	}
	URL r = getSystemResource(name);
	if (r != null) {
	    return r;
	}
	try {
	    Object o = resources.get(name);
	    if (o instanceof ZipFile){
		return new URL(resourceProtocolName, "", "file:" + ((ZipFile)o).getName() + "!/" + name);
	    } else if (o instanceof UrlEntry){
		UrlEntry entry = (UrlEntry)o;
		return new URL(resourceProtocolName, "", entry.url.toString() + "!/" + name);
	    } else {
		return null;
	    }
	} catch (MalformedURLException e){
	    /* ignore */
	}
	return null;
    }

    /**
     * Get an InputStream on a given resource.  Will return null if no
     * resource with this name is found. <p>
     * 
     * @param	name	the name of the resource, to be used as is.
     * @return	an InputStream on the resource, or null if not found.
     */
    public InputStream getResourceAsStream(String name) {
	if (name.startsWith("/")){
	    name = name.substring(1);
	}
	InputStream in = getSystemResourceAsStream(name);
	if (in != null){
	    return in;
	}
	Object o = resources.get(name);
	byte[] b = null;
	if (o instanceof ZipFile){
	    b = getData((ZipFile)o, name);
	} else if (o instanceof UrlEntry){
	    b = ((UrlEntry)o).getData();
	} else {
	    b = (byte[])o;
	}
	if (b != null){
	    return new ByteArrayInputStream(b);
	} else {
	    return null;
	}
    }

    static byte[] getData(ZipFile file, String name) {
	try {
	    ZipEntry entry = file.getEntry(name);
	    InputStream in = file.getInputStream(entry);
	    int n = 0;
	    bout.reset();
	    while ((n = in.read(buf, 0, buf.length)) > 0){
		bout.write(buf, 0, n);
	    }
	    return bout.toByteArray();
	} catch (EOFException e0){
	    System.err.print(e0);
	    System.err.println(", BUGID: 4040920");
	} catch (IOException e){
	    /* ignore */
	}
	return null;
    }

    /**
     * Initialization of statics
     */
    private static final String protocolPathProp = "java.protocol.handler.pkgs";

    static {
	Properties p = new Properties(System.getProperties());
	String s = p.getProperty(protocolPathProp);
	if (s != null){
	    p.put(protocolPathProp, s + "|pnuts.ext");
	} else {
	    p.put(protocolPathProp, "pnuts.ext");
	}
	System.setProperties(p);
    }
}
