/*
 * @(#) PSEPackage.java	1.5 99/06/03
 *
 * Copyright (c) 1997,1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package pnuts.ext;

import pnuts.lang.Package;
import pnuts.lang.Context;
import pnuts.lang.PackageFactory;
import pnuts.lang.PnutsException;
import pnuts.lang.Value;
import java.util.Enumeration;
import java.util.Vector;
import java.io.File;
import COM.odi.ObjectStore;
import COM.odi.Database;
import COM.odi.DatabaseNotFoundException;
import COM.odi.DatabaseRootNotFoundException;
import COM.odi.Transaction;
import COM.odi.util.OSHashMap;

/**
 * This class is a persisnt Package class.  Global variables
 * are automatically saved in a PSE odb file.
 * <PRE>
 * Usage:
 *     import("pnuts.ext.PSEPackage")
 *     package(PSEPackage("foo"))
 * </PRE>
 * @see		pnuts.lang.Package
 * @author	Toyokazu Tomatsu
 * @version	1.5
 */
public class PSEPackage extends Package {

    private transient String dbName;
    private transient OSHashMap map;
    private transient Database db;

    static {
	System.runFinalizersOnExit(true);
    }

    public static class Factory implements PackageFactory {
	public Package createPackage(String name){
	    return new PSEPackage(name);
	}
    }

    /**
     * creates a persistent package
     *
     * @param name the package name
     */
    public PSEPackage(String name){
	this(name, System.getProperty("user.home") + "/_pnuts_" + name + ".odb");
    }

    /**
     * creates a persistent package
     *
     * @param name the package name
     * @param dbName the database name
     */
    public PSEPackage(String name, String dbName){
	this(name, dbName, false);
    }

    /**
     * creates a persistent package
     *
     * @param name the package name
     * @param lock if PSE makes a lock directory
     */
    public PSEPackage(String name, boolean lock){
	this(name, System.getProperty("user.home") + "/_pnuts_" + name + ".odb", lock);
    }

    /**
     * creates a persistent package
     *
     * @param name the package name
     * @param dbName the database name
     * @param lock if PSE makes a lock directory
     */
    public PSEPackage(String name, String dbName, boolean lock){
	super(name);
	System.getProperties().put("COM.odi.useDatabaseLocking", "" + lock);
	this.dbName = dbName;
	ObjectStore.initialize(null, null);
	try {
	    db = Database.open(dbName, ObjectStore.OPEN_UPDATE);
	} catch (DatabaseNotFoundException e1){
	    db = Database.create(dbName,
				 ObjectStore.ALL_READ |
				 ObjectStore.ALL_WRITE);
	}
	Transaction tr = Transaction.begin(ObjectStore.UPDATE);
	try {
	    map = (OSHashMap)db.getRoot("root");
	} catch (DatabaseRootNotFoundException e){
	    db.createRoot("root", map = new OSHashMap());
	}
	tr.commit();
    }

    public void set(String sym, Object obj, Context context){
	Transaction tr = Transaction.begin(ObjectStore.UPDATE);
	try {
	    map = (OSHashMap)db.getRoot("root");
	    map.put(sym, obj);
	    tr.commit();
	} catch (Throwable t){
	    throw new PnutsException(t);
	}
    }

    protected Value lookup(String sym, Context context){
	Transaction tr = Transaction.begin(ObjectStore.READONLY);
	try {
	    map = (OSHashMap)db.getRoot("root");
	    final Object o = map.get(sym);
	    tr.commit();
	    if (o != null){
		return new Value(){
		    public Object get(){
			return o;
		    }
		};
	    } else {
		return null;
	    }
	} catch (Throwable t){
	    throw new PnutsException(t);
	}
    }

    public void clear(String sym, Context context){
	Transaction tr = Transaction.begin(ObjectStore.UPDATE);
	try {
	    map = (OSHashMap)db.getRoot("root");
	    map.remove(sym);
	    tr.commit();
	} catch (Throwable t){
	    throw new PnutsException(t);
	}
    }

    public Enumeration keys(){
	Transaction tr = Transaction.begin(ObjectStore.READONLY);
	try {
	    Enumeration e;
	    map = (OSHashMap)db.getRoot("root");
	    final Object[] array = map.keySet().toArray();
	    for (int i = 0; i < array.length; i++){
		ObjectStore.fetch(array[i]);
	    }
	    tr.commit();
	    e = new Enumeration(){
		Object[] a = array;
		int idx = 0;
		public boolean hasMoreElements(){
		    return idx < a.length;
		}
		public Object nextElement(){
		    return a[idx++];
		}
	    };
	    return e;
	} catch (Throwable t){
	    throw new PnutsException(t);
	}
    }

    public void finalize(){
	db.close();
    }
}
