/*
 * @(#) PnutsClient.java	1.2 99/04/12
 *
 * Copyright (c) 1997-1999 Sun Microsystems, Inc. All Rights Reserved.
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package pnuts.ext;

import java.io.InputStream;
import java.io.Reader;
import java.io.InputStreamReader;
import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.net.MalformedURLException;
import java.util.WeakHashMap;
import java.security.AccessController;
import java.security.PrivilegedAction;
import pnuts.lang.Pnuts;
import pnuts.lang.Context;

/**
 * This class is a sample implementation of a secure script environment based on
 * jdk1.2 security.   The security policy is almost same as that of a Java Applet.
 * The JAR files which include this class should be located in $JRE/lib/ext/.
 * <p>
 * e.g.
 * <pre>
 *  % java -Djava.security.manager pnuts.ext.PnutsClient http://host/xxx/script.pnut
 * </pre>
 * "script.pnut" is executed as safely as Java applets.
 *
 * @version	1.2
 * @see		pnuts.ext.IPnutsProxy
 * @see		pnuts.ext.PnutsProxy
 */
public final class PnutsClient {

    private static WeakHashMap classloaders = new WeakHashMap();

    private IPnutsProxy proxy;

    protected PnutsClient(){
    }

    /**
     * Creates a PnutsClient object.
     *
     * @param codebase the codebase of scripts
     */
    public PnutsClient(final URL codebase){
	AccessController.doPrivileged(new PrivilegedAction() {
	    public Object run() {
		try{
		    ProxyClassLoader loader = getClassLoader(codebase);
		    Class c = loader.getPnutsProxy();
		    proxy = (IPnutsProxy)c.newInstance();
		} catch (Throwable t){
		    t.printStackTrace();
		}
		return null;
	    }
	});
    }

    /**
     * Executes a script at the specified URL
     *
     * @param url the URL
     */
    public static Object load(URL url) throws IOException {
	return load(url, new Context());
    }

    /**
     * Executes a script at the specified URL
     *
     * @param url the URL
     * @param context the Context
     */
    public static Object load(URL url, Context context) throws IOException {
	URL codebase = getCodebase(url);
	PnutsClient client = new PnutsClient(codebase);
	return client.load(url.openStream(), context);
    }

    static ProxyClassLoader getClassLoader(URL codebase) {
	ProxyClassLoader c = (ProxyClassLoader)classloaders.get(codebase);
	if (c == null) {
	    ProxyClassLoader ac = createClassLoader(codebase);
	    classloaders.put(codebase, ac);
	    return ac;
	}
	return c;
    }

    static ProxyClassLoader createClassLoader(URL codebase) {
	return new ProxyClassLoader(codebase);
    }

    /**
     * load a script from InputStream "in"
     *
     * @see pnuts.lang.Pnuts#load(InputStream)
     */
    public Object load(InputStream in){
	return load(new InputStreamReader(in));
    }

    /**
     * load a script from InputStream "in" in "context"
     *
     * @see pnuts.lang.Pnuts#load(InputStream, pnuts.lang.Context)
     */
    public Object load(InputStream in, Context context){
	return load(new InputStreamReader(in), context);
    }

    /**
     *
     * @see pnuts.lang.Pnuts#load(Reader)
     */
    public Object load(Reader in){
	return load(in, new Context());
    }

    /**
     *
     * @see pnuts.lang.Pnuts#load(Reader, pnuts.lang.Context)
     */
    public Object load(Reader in, Context context){
	return proxy.load(in, context);
    }

    /**
     *
     * @see pnuts.lang.Pnuts#eval(String)
     */
    public Object eval(String expr){
	return eval(expr, new Context());
    }

    /**
     *
     * @see pnuts.lang.Pnuts#eval(String, pnuts.lang.Context)
     */
    public Object eval(String expr, Context context){
	return proxy.eval(expr, context);
    }

    static URL getCodebase(URL url) {
	String file = url.getFile();
	String host = url.getHost();
	String protocol = url.getProtocol();
	int i = file.lastIndexOf('/');
	if (i < 0){
	    return null;
	} else {
	    try {
		return new URL(protocol, host, file.substring(0, i + 1));
	    } catch (MalformedURLException e){
		return null;
	    }
	}
    }

    public static void main(String a[]) throws Throwable {
	Pnuts.load("init");
	URL url = new URL(a[0]);
	PnutsClient.load(url);
    }
}
