/*
 * @(#) Quantity.java	1.3 99/03/30
 *
 * Copyright (c) 1997,1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package pnuts.ext;

import java.math.BigDecimal;
import java.io.Serializable;
import pnuts.lang.Arithmetic;
import pnuts.lang.Numeric;

/**
 * This class represents a quantity with a unit.
 * <a href="/pnuts/apidoc/pnuts.ext.UnitSystem.html">pnuts.ext.UnitSystem</a> class is
 * "factory" class for this class.
 *
 * @author	Toyokazu Tomatsu
 * @version	1.3
 */
public class Quantity implements Numeric, Serializable {

    /**
     * @serial
     */
    Number number;

    /**
     * @serial
     */
    Units units;

    /**
     * @serial
     */
    int prec;

    public static Quantity make(UnitSystem sys, Number n, String unit){
	Units u = new Units();
	u.addDimension(sys, unit, 1);
	return new Quantity(n, u, sys.prec);
    }

    Quantity(Number number, Units units, int prec){
	this.number = number;
	this.units = units;
	this.prec = prec;
    }

    public Units units(){
	return units;
    }

    public Number number(){
	return number;
    }

    public Object multiply(Object o){
	if (o instanceof Quantity){
	    Quantity u = (Quantity)o;
	    Units uni = units.multiply(u.units);
	    Number num = (Number)Arithmetic.multiply(number, u.number);
	    int p = (prec > u.prec ? u.prec : prec);
	    Number s = units.relativeScale(u.units);
	    num = (Number)Arithmetic.multiply(num, s);
	    if (num instanceof Double || num instanceof Float){
		BigDecimal d = new BigDecimal(num.doubleValue());
		num = d.setScale(p, BigDecimal.ROUND_HALF_UP);
	    }
	    return new Quantity(num, uni, p);
	} else if (o instanceof Numeric){
	    Number n = (Number)((Numeric)o).multiply(number);
	    Quantity u = new Quantity(n, units, prec);
	    return u;
	} else if (o instanceof Number){
	    Number n = (Number)Arithmetic.multiply(number, o);
	    Quantity u = new Quantity(n, units, prec);
	    return u;
	} else {
	    throw new RuntimeException("illegal type combination");
	}
    }

    public Object divide(Object o){
	if (o instanceof Quantity){
	    Quantity u = (Quantity)o;
	    Units uni = units.divide(u.units);
	    Number num = (Number)Arithmetic.divide(number, u.number);
	    int p = (prec > u.prec ? u.prec : prec);
	    Number s = u.units.relativeScale(units);
	    num = (Number)Arithmetic.multiply(num, s);
	    if (num instanceof Double || num instanceof Float){
		BigDecimal d = new BigDecimal(num.doubleValue());
		num = d.setScale(p, BigDecimal.ROUND_HALF_UP);
	    }
	    return new Quantity(num, uni, p);
	} else if (o instanceof Numeric){
	    return multiply(((Numeric)o).inverse());
	} else if (o instanceof Number){
	    Number n = (Number)Arithmetic.divide(number, o);
	    Quantity u = new Quantity(n, units, prec);
	    return u;
	} else {
	    throw new RuntimeException("illegal type combination");
	}
    }

    public Object add(Object o){
	if (!compatible(o)){
	    throw new RuntimeException("incompatible unit or dimension");
	}
	Quantity u = (Quantity)o;
	Number num = (Number)Arithmetic.multiply(u.number, units.relativeScale(u.units));
	int p = (prec > u.prec ? u.prec : prec);
	if (num instanceof Double || num instanceof Float){
	    BigDecimal d = new BigDecimal(num.doubleValue());
	    num = d.setScale(p, BigDecimal.ROUND_HALF_UP);
	}
	num = (Number)Arithmetic.add(number, num);
	return new Quantity(num, units, prec);
    }

    public Object subtract(Object o){
	if (!compatible(o)){
	    throw new RuntimeException("incompatible unit or dimension");
	}
	Quantity u = (Quantity)o;
	Number num = (Number)Arithmetic.multiply(u.number, units.relativeScale(u.units));
	num = (Number)Arithmetic.subtract(number, num);
	int p = (prec > u.prec ? u.prec : prec);
	if (num instanceof Double || num instanceof Float){
	    BigDecimal d = new BigDecimal(num.doubleValue());
	    num = d.setScale(p, BigDecimal.ROUND_HALF_UP);
	}
	return new Quantity(num, units, prec);
    }

    boolean compatible(Object o){
	if (o instanceof Quantity){
	    Quantity mu = (Quantity)o;
	    return units.equals(mu.units);
	} else {
	    return false;
	}
    }

    public int compareTo(Object o){
	if (!compatible(o)){
	    throw new RuntimeException("incompatible unit or dimension");
	}
	Quantity u = (Quantity)o;
	return Arithmetic.compareTo(number, u.number);
    }

    public Object negate(){
	return new Quantity((Number)Arithmetic.negate(number), units, prec);
    }

    public Object inverse(){
	BigDecimal d = new BigDecimal(1.0 / number.doubleValue());
	Number n = d.setScale(prec, BigDecimal.ROUND_HALF_UP);
	return new Quantity(n, units.inverse(), prec);
    }

    public String toString(){
	return number + units.toString();
    }
}
