/*
 * @(#) Arithmetic.java	1.7 99/05/19
 *
 * Copyright (c) 1997-1999 Sun Microsystems, Inc. All Rights Reserved.
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package pnuts.lang;

import java.math.BigInteger;
import java.math.BigDecimal;
import java.lang.reflect.Array;
import java.lang.reflect.Method;

/**
 * The Arithmetic class is geared for transparent multi-precision arithmetic operations.
 * <p>
 * Here is the type conversion table for binary operation:
 * </p>
 * <table border>
 * <tr><td></td><td>BigDecimal</td><td>BigInteger</td><td>Double</td><td>Float</td><td>Long</td><td>Integer</td></tr>
 * <tr><td>BigDecimal</td><td>BigDecimal</td><td>BigDecimal</td><td>BigDecimal</td><td>BigDecimal</td><td>BigDecimal</td><td>BigDecimal</td></tr>
 * <tr><td>BigInteger</td><td>BigDecimal</td><td>BigInteger</td><td>BigDecimal</td><td>BigDecimal</td><td>BigInteger</td><td>BigInteger</td></tr>
 * <tr><td>Double</td><td>BigDecimal</td><td>BigDecimal</td><td>Double</td><td>Double</td><td>Double</td><td>Double</td></tr>
 * <tr><td>Float</td><td>BigDecimal</td><td>BigDecimal</td><td>Double</td><td>Float</td><td>Float</td><td>Float</td></tr>
 * <tr><td>Long</td><td>BigDecimal</td><td>BigInteger</td><td>Double</td><td>Float</td><td>Long</td><td>Long</td></tr>
 * <tr><td>Integer</td><td>BigDecimal</td><td>BigInteger</td><td>Double</td><td>Float</td><td>Long</td><td>Integer</td></tr>
 * </table>
 */
public final class Arithmetic {
    /*
     * operators
     */
    private static final int ADD = 0;
    private static final int SUBTRACT = 1;
    private static final int MULTIPLY = 2;
    private static final int DIVIDE = 3;
    private static final int REMAINDER = 4;
    private static final int GCD = 5;
    private static final int MOD = 6;
    private static final int COMPARE = 7;
    private static final int NEGATE = 10;
    private static final int ABS = 11;
    private static final int AND = 12;
    private static final int OR = 13;
    private static final int XOR = 14;
    private static final int NOT = 15;
    private static final int SHIFT_R = 16;
    private static final int SHIFT_L = 17;
    private static final int SHIFT_A = 18;
    private static final int POW = 19;

    private static Integer zero = new Integer(0);

    private static BigInteger maxLong = BigInteger.valueOf(Long.MAX_VALUE);
    private static BigInteger minLong = BigInteger.valueOf(Long.MIN_VALUE);

    public final static int NOT_EQUAL = Numeric.NOT_EQUAL;
    public final static int LEFT_IS_BIGGER = Numeric.LEFT_IS_BIGGER;
    public final static int RIGHT_IS_BIGGER = Numeric.RIGHT_IS_BIGGER;
    public final static int EQUAL = Numeric.EQUAL;

    private final static Integer LEFT = new Integer(LEFT_IS_BIGGER);
    private final static Integer RIGHT = new Integer(RIGHT_IS_BIGGER);
    private final static Integer EQ = new Integer(EQUAL);
    private final static Integer NE = new Integer(NOT_EQUAL);

    static Object numeric_object(int operator, Numeric n1, Object n2){
	switch (operator){
	case ADD:
	    return n1.add(n2);
	case SUBTRACT:
	    return n1.subtract(n2);
	case MULTIPLY:
	    return n1.multiply(n2);
	case DIVIDE:
	    return n1.divide(n2);
	case COMPARE:
	    return new Integer(n1.compareTo(n2));
	}
	throw new IllegalArgumentException("" + operator);
    }

    static Object object_numeric(int operator, Object n1, Numeric n2){
	switch (operator){
	case ADD:
	    return n2.add(n1);
	case SUBTRACT:
	    return add(n2.negate(), n1);
	case MULTIPLY:
	    return n2.multiply(n1);
	case DIVIDE:
	    return multiply(n2.inverse(), n1);
	case COMPARE:
	    int c = ((Numeric)n2).compareTo(n1);
	    if (c == LEFT_IS_BIGGER){
		return RIGHT;
 	    } else if (c == RIGHT_IS_BIGGER){
		return LEFT;
	    } else if (c == EQUAL){
		return EQ;
	    } else {
		return NE;
	    }
	}
	throw new IllegalArgumentException("" + operator);
    }

    static Object unary(int operator, Object n1){
	if (n1 instanceof BigDecimal){
	    switch (operator){
	    case NEGATE: return ((BigDecimal)n1).negate();
	    case ABS:    return ((BigDecimal)n1).abs();
	    default:   	 throw new IllegalArgumentException("" + operator);

	    }
	}
	if (n1 instanceof BigInteger){
	    switch (operator){
	      case NEGATE: return ((BigInteger)n1).negate();
	      case ABS:    return ((BigInteger)n1).abs();
	      case NOT:    return ((BigInteger)n1).not();
	      default:     throw new IllegalArgumentException("" + operator);
	    }
	}
	if (n1 instanceof Double){
	    double d = Double.NaN;
	    switch (operator){
	      case NEGATE: d = -((Number)n1).doubleValue(); break;
	      case ABS:    d = Math.abs(((Number)n1).doubleValue()); break;
	      default:     throw new IllegalArgumentException("" + operator);
	    }
	    return new Double(d);
	}
	if (n1 instanceof Float){
	    float d = Float.NaN;
	    switch (operator){
	      case NEGATE: d = -((Number)n1).floatValue(); break;
	      case ABS:    d = Math.abs(((Number)n1).floatValue()); break;
	      default:     throw new IllegalArgumentException("" + operator);
	    }
	    return new Float(d);
	}
	if (n1 instanceof Long ||
	    n1 instanceof Integer ||
	    n1 instanceof Short ||
	    n1 instanceof Byte){
	    long d = 0;
	    switch (operator){
	      case NEGATE: d = -((Number)n1).longValue(); break;
	      case ABS:    d = Math.abs(((Number)n1).longValue()); break;
	      case NOT:    d = ~((Number)n1).longValue(); break;
	      default:     throw new IllegalArgumentException("" + operator);
	    }
	    return longOrInt(d);
	}
	if (n1 instanceof Numeric){
	    switch (operator){
	    case NEGATE:
		if (((Numeric)n1).compareTo(zero) == RIGHT_IS_BIGGER){
		    return ((Numeric)n1).negate();
		} else {
		    return n1;
		}
	    }
	}
	throw new IllegalArgumentException(operator + "(" + n1 + ")");
    }

    static Object binary(int operator, Object n1, Object n2){
	boolean i1 = false;
	boolean i2 = false;
	boolean l1 = false;
	boolean l2 = false;
	boolean bd1 = false;
	boolean bd2 = false;
	boolean bi1 = false;
	boolean bi2 = false;
	boolean f1 = false;
	boolean f2 = false;
	boolean d1 = false;
	boolean d2 = false;

	if (n1 instanceof Integer || n1 instanceof Short || n1 instanceof Byte){
	    i1 = true;
	} else if (n1 instanceof Long){
	    l1 = true;
	} else if (n1 instanceof BigDecimal){
	    bd1 = true;
	} else if (n1 instanceof BigInteger){
	    bi1 = true;
	} else if (n1 instanceof Float){
	    f1 = true;
	} else if (n1 instanceof Double){
	    d1 = true;
	} else if (n1 instanceof String){
	    if (operator == ADD){
		return concat(n1, n2);
	    }
	} else if (n1 instanceof Numeric){
	    return numeric_object(operator, (Numeric)n1, n2);
	}
	if (n2 instanceof Integer || n2 instanceof Short || n2 instanceof Byte){
	    i2 = true;
	} else if (n2 instanceof Long){
	    l2 = true;
	} else if (n2 instanceof BigDecimal){
	    bd2 = true;
	} else if (n2 instanceof BigInteger){
	    bi2 = true;
	} else if (n2 instanceof Float){
	    f2 = true;
	} else if (n2 instanceof Double){
	    d2 = true;
	} else if (n2 instanceof String){
	    if (operator == ADD){
		return concat(n1, n2);
	    }
	} else if (n2 instanceof Numeric){
	    return object_numeric(operator, n1, (Numeric)n2);
	}
	if (i1 && i2){
	    return op2_int(operator, ((Number)n1).intValue(), ((Number)n2).intValue());
	}
	if ((i1 || l1) && (i2 || l2)){
	    return op2_long(operator, ((Number)n1).longValue(), ((Number)n2).longValue());
	}
	if (bd1 || bd2){
	    BigDecimal _d1, _d2;
	    if (bi1){
		_d1 = new BigDecimal((BigInteger)n1);
	    } else if (bd1){
		_d1 = (BigDecimal)n1;
	    } else if (f1 || d1){
		double dd = ((Number)n1).doubleValue();
	        if (Double.isInfinite(dd) || Double.isNaN(dd)){
	            return n1;
                } else {
		    String s = n1.toString();
		    int idx = s.indexOf('E');
		    int p = s.indexOf('.');
		    if (idx > 0){
			int e = Integer.parseInt(s.substring(idx));
		    }
		    _d1 = doubleToDecimal(dd);
	        }
	    } else {
		_d1 = new BigDecimal(((Number)n1).longValue());
	    }
	    if (bi2){
		_d2 = new BigDecimal((BigInteger)n2);
	    } else if (bd2){
		_d2 = (BigDecimal)n2;
	    } else if (f2 || d2){
		double dd = ((Number)n2).doubleValue();
	        if (Double.isInfinite(dd) || Double.isNaN(dd)){
		    return n2;
		} else {
		    _d2 = doubleToDecimal(dd);
		}
	    } else {
		_d2 = new BigDecimal(((Number)n2).longValue());
	    }
	    return op2_bdec(operator, _d1, _d2);
	} else if (bi1 || bi2){
	    if (d1 || d2 || f1 || f2){
		BigDecimal _d1, _d2;
		if (d1 || f1){
		    double dd = ((Number)n1).doubleValue();
		    if (Double.isInfinite(dd) || Double.isNaN(dd)){
		        return n1;
		    }
		    _d1 = doubleToDecimal(dd);
		} else {
		    _d1 = new BigDecimal((BigInteger)n1);
		}
		if (d2 || f2){
		    double dd = ((Number)n2).doubleValue();
		    if (Double.isInfinite(dd) || Double.isNaN(dd)){
		        return n2;
		    }
		    _d2 = doubleToDecimal(dd);
	        } else {
		    _d2 = new BigDecimal((BigInteger)n2);
		}
		return op2_bdec(operator, _d1, _d2);
	    } else {
		BigInteger _i1 = (l1 || i1) ? BigInteger.valueOf(((Number)n1).longValue()) : (BigInteger)n1;
		BigInteger _i2 = (l2 || i2) ? BigInteger.valueOf(((Number)n2).longValue()) : (BigInteger)n2;
		return op2_bint(operator, _i1, _i2);
	    }
	} else if (d1 || d2){
 	    return op2_double(operator, ((Number)n1).doubleValue(), ((Number)n2).doubleValue());
	} else if (f1 || f2){
	    return op2_float(operator, ((Number)n1).floatValue(), ((Number)n2).floatValue());
	} else if (n1.getClass().isArray() && n2.getClass().isArray()){
	    if (operator == ADD){
		return append(n1, n2);
	    }
	}
	throw new IllegalArgumentException(operator + "(" + n1 + "," + n2 + ")");
    }

    static Object op2_bdec(int operator, BigDecimal d1, BigDecimal d2){
	switch (operator){
	case ADD:
	    return d1.add(d2);
	case SUBTRACT:
	    return d1.subtract(d2);
	case MULTIPLY:
	    return d1.multiply(d2);
	case DIVIDE:
	    return d1.divide(d2, BigDecimal.ROUND_HALF_DOWN);
	case COMPARE:
	    int i = d1.compareTo(d2);
	    if (i > 0){
		return LEFT;
	    } else if (i < 0){
		return RIGHT;
	    } else {
		return EQ;
	    }
	case POW:
	    return new Double(Math.pow(d1.doubleValue(), d2.doubleValue()));
	default:
	    throw new IllegalArgumentException("" + operator);
        }
    }

    static Object op2_bint(int operator, BigInteger i1, BigInteger i2){
	switch (operator){
	case ADD:
	    return compress(i1.add(i2));
	case SUBTRACT:
	    return compress(i1.subtract(i2));
	case MULTIPLY:
	    return compress(i1.multiply(i2));
	case DIVIDE:
	    return compress(i1.divide(i2));
	case COMPARE:
	    return new Integer(i1.compareTo(i2));
	case REMAINDER:
	    return compress(i1.remainder(i2));
	case GCD:
	    return compress(i1.gcd(i2));
	case MOD:
	    return compress(i1.mod(i2));
	case AND:
	    return compress(i1.and(i2));
	case OR:
	    return compress(i1.or(i2));
	case XOR:
	    return compress(i1.xor(i2));
	case SHIFT_L:
	    return compress(i1.shiftLeft(i2.intValue()));
	case SHIFT_R:
	    return compress(i1.shiftRight(i2.intValue()));
	case POW:
	    Number n = compress(i2);
	    if (n instanceof Integer){
		return i1.pow(n.intValue());
	    } else {
		return new Double(Math.pow(i1.doubleValue(), i2.doubleValue()));
	    }
	default:
	    throw new IllegalArgumentException("" + operator);
	}
    }

    static Object op2_double(int operator, double d1, double d2){
	switch (operator){
	case ADD:
	    return new Double(d1 + d2);
	case SUBTRACT:
	    return new Double(d1 - d2);
	case MULTIPLY:
	    return new Double(d1 * d2);
	case DIVIDE:
	    return new Double(d1 / d2);
	case COMPARE:
	    if (d1 > d2){
		return LEFT;
	    } else if (d1 < d2){
		return RIGHT;
	    } else if (d1 == d2){
		return EQ;
	    } else {
		return NE;
	    }
	case POW:
	    return new Double(Math.pow(d1, d2));
	default:
	    throw new IllegalArgumentException("" + operator);
	}
    }

    static Object op2_float(int operator, float d1, float d2){
	switch (operator){
	case ADD:
	    return new Float(d1 + d2);
	case SUBTRACT:
	    return new Float(d1 - d2);
	case MULTIPLY:
	    return new Float(d1 * d2);
	case DIVIDE:
	    return new Float(d1 / d2);
	case COMPARE:
	    if (d1 > d2){
		return LEFT;
	    } else if (d1 < d2){
		return RIGHT;
	    } else if (d1 == d2){
		return EQ;
	    } else {
		return NE;
	    }
	case POW:
	    return new Float((float)Math.pow(d1, d2));
	default:
	    throw new IllegalArgumentException("" + operator);
	}
    }

    static Object op2_long(int operator, long d1, long d2){
	long d3;
	switch (operator){
	case ADD:
	    d3 = d1 + d2;
	    if (d3 > d1 && d2 < 0 || d3 < d1 && d2 > 0){
		return op2_bint(ADD, BigInteger.valueOf(d1), BigInteger.valueOf(d2));
	    }
	    return longOrInt(d3);
	case SUBTRACT:
	    d3 = d1 - d2;
	    if (d3 > d1 && d2 > 0 || d3 < d1 && d2 < 0){
		return op2_bint(SUBTRACT, BigInteger.valueOf(d1), BigInteger.valueOf(d2));
	    }
	    return longOrInt(d3);
	case MULTIPLY:
	    d3 = d1 * d2;
	    if (d2 == 0){
		return zero;
	    } else if (d3 / d2 == d1){
		return longOrInt(d3);
	    } else {
		return op2_bint(MULTIPLY, BigInteger.valueOf(d1), BigInteger.valueOf(d2));
	    }
	case DIVIDE:
	    d3 = d1 / d2;
	    return longOrInt(d3);
	case MOD:
	    d3 = d1 % d2;
	    return longOrInt(d3);
	case COMPARE:
	    if (d1 > d2){
		return LEFT;
	    } else if (d1 < d2){
		return RIGHT;
	    } else if (d1 == d2){
	        return EQ;
	    } else {
		return NE;
	    }
	case AND:
	    return longOrInt(d1 & d2);
	case OR:
	    return longOrInt(d1 | d2);
	case XOR:
	    return longOrInt(d1 ^ d2);
	case SHIFT_L:
	    int m = 0;
	    long dd = d1;
	    while (dd > 0){
		dd >>= 1;
		m++;
	    }
	    if (d2 + m > 63){
		return op2_bint(SHIFT_L, BigInteger.valueOf(d1), BigInteger.valueOf(d2));
	    }
	    return longOrInt(d1 << d2);
	case SHIFT_R:
	    return longOrInt(d1 >> d2);
	case SHIFT_A:
	    return longOrInt(d1 >>> d2);
	case GCD:
	    return longOrInt(gcd(d1, d2));
	case POW:
	      return op2_bint(POW, BigInteger.valueOf(d1), BigInteger.valueOf(d2));
	default:
	    throw new IllegalArgumentException("" + operator);
	}
    }

    final static Object op2_int(int operator, int d1, int d2){
	switch (operator){
	  case ADD:
	    return longOrInt((long)d1 + d2);
	  case SUBTRACT:
	    return longOrInt((long)d1 - d2);
	  case MULTIPLY:
	    return longOrInt((long)d1 * d2);
	  case DIVIDE:
	    return longOrInt(d1 / d2);
	  case MOD:
	    return longOrInt(d1 % d2);
	  case COMPARE:
	    if (d1 > d2){
		return LEFT;
	    } else if (d1 < d2){
		return RIGHT;
	    } else if (d1 == d2){
	        return EQ;
	    } else {
	        return NE;
	    }
	  case AND:
	    return longOrInt(d1 & d2);
	  case OR:
	    return longOrInt(d1 | d2);
	  case XOR:
	    return longOrInt(d1 ^ d2);
	  case SHIFT_L:
	    int m = 0;
	    long dd = (long)d1;
	    while (dd > 0){
		dd >>= 1;
		m++;
	    }
	    if (d2 + m > 63){
		return op2_bint(SHIFT_L, BigInteger.valueOf(d1), BigInteger.valueOf(d2));
	    }
	    return longOrInt((long)d1 << d2);
	  case SHIFT_R:
	    return longOrInt(d1 >> d2);
	  case SHIFT_A:
	    return longOrInt(d1 >>> d2);
	  case GCD:
	    return new Integer((int)gcd(d1, d2));
	  case POW:
	      return op2_bint(POW, BigInteger.valueOf(d1), BigInteger.valueOf(d2));
	  default:
	    throw new IllegalArgumentException("" + operator);
	}
    }

    static Number compress(Number n){
	if (n instanceof BigInteger){
	    BigInteger b = (BigInteger)n;
	    if (b.compareTo(maxLong) > 0 ||
		b.compareTo(minLong) < 0){
		return n;
	    } else {
		long l = b.longValue();
		return longOrInt(l);
	    }
	} else if (n instanceof BigDecimal){
	    int j = 0;
	    String f = n.toString();
	    int p = f.indexOf((int)'.');
	    if (p > 0){
	        char ca[] = f.toCharArray();
	        for (int i = 0; i < ca.length - 2 && i < ca.length - p - 2; i++){
	            if (ca[ca.length - i - 1] == '0'){
	                j++;
	            } else {
	                break;
	            }
                }
	        f = f.substring(0, ca.length - j);
	    }
	    if (f.length() < 16){
	        return new Double(n.doubleValue());
	    }
	    if (j > 0){
	        return new BigDecimal(f);
	    } else {
	        return n;
	    }
	} else {
	    return n;
	}
    }


    final static Number longOrInt(long l){
	if (l <= Integer.MAX_VALUE && l >= Integer.MIN_VALUE){
	    return new Integer((int)l);
	} else {
	    return new Long(l);
    	}
    }

    public final static Object add(Object n1, Object n2){
	return binary(ADD, n1, n2);
    }

    public final static Object subtract(Object n1, Object n2){
	return binary(SUBTRACT, n1, n2);
    }

    public final static Object multiply(Object n1, Object n2){
	return binary(MULTIPLY, n1, n2);
    }

    public final static Object divide(Object n1, Object n2){
	return binary(DIVIDE, n1, n2);
    }

    public final static int compareTo(Object n1, Object n2){
	if ((n1 instanceof Number || n1 instanceof Numeric) &&
	    (n2 instanceof Number || n2 instanceof Numeric)){
	    return ((Number)binary(COMPARE, n1, n2)).intValue();
	} else if (n1 instanceof Character && n2 instanceof Character){
	    int c = ((Character)n1).charValue() - ((Character)n2).charValue();
	    if (c < 0){
		return RIGHT_IS_BIGGER;
	    } else if (c > 0){
		return LEFT_IS_BIGGER;
	    } else {
		return EQUAL;
	    }
	} else if (n1 instanceof String && n2 instanceof String){
	    int c = ((String)n1).compareTo((String)n2);
	    if (c < 0){
		return RIGHT_IS_BIGGER;
	    } else if (c > 0){
		return LEFT_IS_BIGGER;
	    } else {
		return EQUAL;
	    }
	} else if (n1 == null){
	    return n2 == null ? EQUAL : NOT_EQUAL;
	} else if (n2 == null){
	    return NOT_EQUAL;
	} else if (isComparable(n1)){
	    return compareComparable(n1, n2);
	} else if (isComparable(n2)){
	    return - compareComparable(n2, n1);
	} else if (n1.getClass().isArray() && n2.getClass().isArray()){
	    int len = 0;
	    if ((len = Array.getLength(n2)) == Array.getLength(n1)){
		for (int i = 0; i < len; i++){
		    if (compareTo(Array.get(n2, i), Array.get(n1, i)) != EQUAL){
			return NOT_EQUAL;
		    }
		}
	    } else {
		return NOT_EQUAL;
	    }
	    return EQUAL;
	} else {
	    return n2.equals(n1) ? EQUAL : NOT_EQUAL;
	}
    }

    private static Class comparable = null;
    private static Method compare_to = null;
    private static int tryFlag = 0;

    final static int compareComparable(Object n1, Object n2){
	try {
	    return ((Integer)compare_to.invoke(n1, new Object[]{n2})).intValue();
	} catch (Throwable t){
	}
	return NOT_EQUAL;
    }

    final static boolean isComparable(Object n1){
	try {
	    switch (tryFlag){
	    case 0:
		comparable = Pnuts.loadClass("java.lang.Comparable");
		compare_to = comparable.getMethod("compareTo", new Class[]{Object.class});
		tryFlag = 1;
	    case 1:
		return comparable.isInstance(n1);
	    case -1:
		return false;
	    }
	} catch (Throwable t){
	    tryFlag = -1;
	}
	return false;
    }

    public final static Object remainder(Object n1, Object n2){
	return binary(REMAINDER, n1, n2);
    }

    public final static Object gcd(Object n1, Object n2){
	return binary(GCD, n1, n2);
    }

    public final static Object pow(Object n1, Object n2){
	return binary(POW, n1, n2);
    }

    public final static Object mod(Object n1, Object n2){
	return binary(MOD, n1, n2);
    }

    public final static Object and(Object n1, Object n2){
	return binary(AND, n1, n2);
    }

    public final static Object or(Object n1, Object n2){
	return binary(OR, n1, n2);
    }

    public final static Object xor(Object n1, Object n2){
	return binary(XOR, n1, n2);
    }

    public final static Object shiftLeft(Object n1, Object n2){
	return binary(SHIFT_L, n1, n2);
    }

    public final static Object shiftRight(Object n1, Object n2){
	return binary(SHIFT_R, n1, n2);
    }

    public final static Object shiftArithmetic(Object n1, Object n2){
	return binary(SHIFT_A, n1, n2);
    }

    public final static Object abs(Object n1){
	return unary(ABS, n1);
    }

    public static Object negate(Object n1){
	if (n1 instanceof Numeric){
	    return ((Numeric)n1).negate();
	} else {
	    return unary(NEGATE, n1);
	}
    }

    public static Object inverse(Object n1){
	if (n1 instanceof Numeric){
	    return ((Numeric)n1).inverse();
	} else {
	    return divide(new Double(1.0), (Number)n1);
	}
    }

    public final static Object not(Object n1){
	return unary(NOT, n1);
    }

    static BigDecimal doubleToDecimal(double d){
	return (BigDecimal)decimal("" + d, 10);
    }

    static Number decimalNumber(String s, int radix){
	return decimal(s, radix);
    }

    static Number decimal(String s, int radix){
	if (radix == 16){
	    return new BigInteger(s, radix);
	} else {
	    return new BigDecimal(deNormalize(s));
	}
    }

    private static String deNormalize(String fmt){
	int pt = fmt.indexOf((int)'.');
	int exp = fmt.indexOf((int)'E');
	if (exp < 0){
	    return fmt;
	}
	int e = Integer.parseInt(fmt.substring(exp + 1));
	int i = 0;
	int k = 0;
	char b[];
	if (e < 0){
	    b = new char[exp - e];
	    i = 0;
	    if (fmt.charAt(k) == '-'){
	        b[i++] = '-';
		k++;
	    }
	    b[i++] = '0';
	    b[i++] = '.';
	    for (int j = 0; j < -e - 1; j++){
	        b[i++] = '0';
	    }
	    b[i++] = fmt.charAt(k++);
	    k++;
	    for (int j = 0; j < exp - pt - 1; j++){
	        b[i++] = fmt.charAt(k++);
	    }
	} else {
	    if (exp - pt - 1 > e){
		b = new char[exp];
	        i = 0;
	        if (fmt.charAt(k) == '-'){
	            b[i++] = '-';
		    k++;
	        }
	        b[i++] = fmt.charAt(k++);
	        k++;
		for (int j = 0; j < e; j++){
	            b[i++] = fmt.charAt(k++);
	        }
		b[i++] = '.';
		for (int j = 0; j < exp - pt - 1 - e; j++){
	            b[i++] = fmt.charAt(k++);
	        }
	    } else {
		b = new char[exp -1 + (e - (exp - pt - 1))];
	        i = 0;
	        if (fmt.charAt(k) == '-'){
	            b[i++] = '-';
		    k++;
	        }
	        b[i++] = fmt.charAt(k++);
	        k++;
		for (int j = 0; j < exp - pt - 1; j++){
	            b[i++] = fmt.charAt(k++);
	        }
		for (int j = 0; j < e - (exp - pt - 1); j++){
	            b[i++] = '0';
	        }
	    }
	}
	return new String(b);
    }

    static long gcd(long a, long b){ // Euclid's algorithm
	if (b > a){
	    long tmp = a;
	    a = b;
	    b = tmp;
	}
	for(;;){
	    if (b == 0){
		return a;
	    } else if (b == 1){
		return b;
	    } else {
		long tmp = b;
		b = a % b;
		a = tmp;
	    }
	}
    }

    final static String NULL = "null";

    static String concat(Object o1, Object o2){
	String s2;
	if (o2 == null){
	    s2 = NULL;
	} else if (o2.getClass().isArray()){
	    s2 = Pnuts.format(o2);
	} else {
	    s2 = o2.toString();
	}
	String s1;
	if (o1 == null){
	    s1 = NULL;
	} else if (o1.getClass().isArray()){
	    s1 = Pnuts.format(o1);
	} else {
	    s1 = o1.toString();
	}
	return s1 + s2;
    }

    static Object append(Object p1, Object p2){
	int l1 = Array.getLength(p1);
	int l2 = Array.getLength(p2);
	Object p3 = Array.newInstance(p1.getClass().getComponentType(), l1 + l2);
	try {
	    for (int i = 0; i < l1; i++){
		Array.set(p3, i, Array.get(p1, i));
	    }
	    for (int i = 0; i < l2; i++){
		Array.set(p3, i + l1, Array.get(p2, i));
	    }
	    return p3;
	} catch (Exception e){
	    throw new IllegalArgumentException(e.getMessage());
	}
    }
}
