/*
 * @(#) Context.java	1.26 99/06/03
 *
 * Copyright (c) 1997-1999 Sun Microsystems, Inc. All Rights Reserved.
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package pnuts.lang;

import java.io.OutputStream;
import java.io.Writer;
import java.io.PrintWriter;
import java.util.Hashtable;

/**
 * <em>Context</em> represents an internal state in Pnuts runtime environment.
 *  A <a href="/pnuts/apidoc/pnuts.lang.Context.html">Context</a>  object contains
 *  the following information.
 * <ul>
 * <li>Stack frame
 * <li>Which Pnuts-package being used
 * <li>Imported Java-package list
 * <li>Writer to which print() write data
 * <li>Writer to which error() write message
 * </ul>
 * <p>
 * Note that the name space is shared by all contexts.
 * </p>
 *
 * @see		ManualContext
 * @version	1.26
 */
public class Context implements Cloneable {

    public final static PrintWriter defaultOutputStream = new PrintWriter(System.out, false);
    public final static PrintWriter defaultTerminalStream = defaultOutputStream;
    public final static PrintWriter defaultErrorStream = new PrintWriter(System.err, true);

    static Object UNDEF = new Object[0];
    static String exceptionHandlerTableSymbol = "!exception";

    static ImportEnv defaultImports = new ImportEnv();
    static {
	defaultImports.add("");
	defaultImports.add("java.lang");
    }

    /*
     * A flag to tell if this context is created with eval() builtin function
     */
    boolean eval = false;
    Context parent;

    Function frame;
    StackFrame env = new StackFrame();

    protected Stack loadingResource = new Stack();
    protected int beginLine = -1;
    protected ImportEnv importEnv;

    PrintWriter outputWriter = null;
    PrintWriter errorWriter = null;
    PrintWriter terminalWriter = null;

    String name = null;

    Package currentPackage = Package.getGlobalPackage();
	{
	    currentPackage.init(this);
	}

    /**
     * Create a new context
     */
    public Context(){
	this(Package.getGlobalPackage());
    }

    /**
     * Creates a context.
     *
     * @param pkg the name of the package.
     */
    public Context(String pkg){
	this(Package.getPackage(pkg));
    }

    /**
     * Creates a context.
     *
     * @param pkg the initial package of the context.
     */
    public Context(Package pkg){
	currentPackage = pkg;
	importEnv = defaultImports;

	this.outputWriter = defaultOutputStream;
	this.terminalWriter = defaultTerminalStream;
	this.errorWriter = defaultErrorStream;
    }

    /**
     * Make a clone of the context
     */
    public Object clone(){
	try {
	    if (importEnv == defaultImports){
		 return super.clone();
	    } else {
		Context ret = (Context)super.clone();
		ret.importEnv = (ImportEnv)importEnv.clone();
		return ret;
	    }
	} catch (Exception e){
	    throw new InternalError("clone error");
	}
    }

    public void setName(String name){
	this.name = name;
    }

    public String getName(){
	return name;
    }

    /**
     * set output stream of the context
     */
    public void setOutputStream(Object out, boolean autoFlush){
	if (out == null){
	    this.outputWriter = null;
	} else if (out instanceof OutputStream){
	    this.outputWriter = new PrintWriter((OutputStream)out, autoFlush);
	} else if (out instanceof Writer){
	    this.outputWriter = new PrintWriter((Writer)out, autoFlush);
	} else {
	    throw new IllegalArgumentException(PnutsException.getMessage("illegal.streamType", PnutsException.NO_PARAM));
	}
    }

    public void setOutputStream(Object outputStream){
	if (outputStream instanceof PrintWriter){
	    this.outputWriter = (PrintWriter)outputStream;
	} else {
	    setOutputStream(outputStream, false);
	}
    }

    /**
     * get output stream of the context
     */
    public PrintWriter getOutputStream(){
	return outputWriter;
    }

    /**
     * set terminal-output-stream of the context
     */
    public void setTerminalStream(Object str, boolean autoFlush){
	if (str == null){
	    this.terminalWriter = null;
	} else if (str instanceof OutputStream){
	    this.terminalWriter = new PrintWriter((OutputStream)str, autoFlush);
	} else if (str instanceof Writer){
	    this.terminalWriter = new PrintWriter((Writer)str, autoFlush);
	} else {
	    throw new IllegalArgumentException(PnutsException.getMessage("illegal.streamType", PnutsException.NO_PARAM));
	}
    }

    public void setTerminalStream(Object stream){
	if (stream instanceof PrintWriter){
	    this.terminalWriter = (PrintWriter)stream;
	} else {
	    setTerminalStream(stream, false);
	}
    }

    /**
     * get terminal-output-stream of the context
     */
    public PrintWriter getTerminalStream(){
	return terminalWriter;
    }

    /**
     * Set an OutputStream or a Writer to which error() write message
     * If errorStream is null, exception is thrown out of eval loop.
     */
    public void setErrorStream(Object errorStream, boolean autoFlush){
	if (errorStream == null){
	    this.errorWriter = null;
 	} else if (errorStream instanceof OutputStream){
	    this.errorWriter = new PrintWriter((OutputStream)errorStream, autoFlush);
 	} else if (errorStream instanceof Writer){
	    this.errorWriter = new PrintWriter((Writer)errorStream, autoFlush);
	} else {
	    throw new IllegalArgumentException(PnutsException.getMessage("illegal.streamType", PnutsException.NO_PARAM));
	}
    }

    /**
     * set an OutputStream or a Writer to which error() write message
     * If errorStream is null, exception is thrown out of eval loop.
     */
    public void setErrorStream(Object errorStream){
	if (errorStream instanceof PrintWriter){
	    this.errorWriter = (PrintWriter)errorStream;
	} else {
	    setErrorStream(errorStream, false);
	}
    }

    /**
     * Get an OutputStream or a Writer to which error() write message
     */
    public PrintWriter getErrorStream(){
	return errorWriter;
    }

    /**
     * get the current package
     */
    public Package getCurrentPackage(){
	return currentPackage;
    }

    void setCurrentPackage(Package pkg){
	pkg.init(this);
	currentPackage = pkg;
    }

    void clear(){
	reset();
	eval = false;
	frame = null;
	importEnv = defaultImports;
	currentPackage = Package.globalPackage;
	beginLine = -1;
    }

    /*
     * open new scope with the naive interpreter
     */
    void open(String locals[], Function func){
	env = new StackFrame(locals, func, env);
    }

    /*
     * open for/foreach scope with the naive interpreter
     */
    void openLocal(String locals[]){
	env.openLocal(locals);
    }

    /*
     * close the scope with the naive interpreter
     */
    void close(){
	env = env.parent;
	if (env != null){
	    frame = env.func;
	} else {
	    frame = null;
	}
    }

    /*
     * close for/foreach scope with the naive interpreter
     */
    void closeLocal(){
	env.closeLocal();
    }

    /*
     * reset scope with the naive interpreter
     */
    void reset(){
	env = new StackFrame();
    }

    protected void set(String symbol, Object obj){
	Binding b = env.lookup(symbol);
	if (b != null){
	    b.value = obj;
	    return;
	}

	Function ff = env.func;
	while (ff != null){
	    SymbolTable ls = ff.lexicalScope;
	    if (ls != null){
		Binding bb = ls.lookup0(symbol);
		if (bb != null){
		    ((Binding)bb.value).value = obj;
		    return;
		}
	    }
	    ff = ff.outer;
	}

	if (env.parent != null){
	    env.put1(symbol, obj);
	} else {
	    currentPackage.set(symbol, obj, this);
	}
    }

    final void bind(String symbol, Object obj){
	env.bind(symbol, obj);
    }

    void addImport(Class clazz){
	if (importEnv == defaultImports){
	    importEnv = (ImportEnv)importEnv.clone();
	}
	importEnv.add(clazz);
    }

    void addImport(String pkgName){
	if (importEnv == defaultImports){
	    importEnv = (ImportEnv)importEnv.clone();
	}
	importEnv.add(pkgName);
    }

    void clearImport(){
	importEnv = new ImportEnv();
    }

    Object undefined(String sym){
	throw new PnutsException("not.defined", new Object[]{sym}, this);
    }

    protected Object get(String symbol){
	Object val = _get(symbol);
	if (val == UNDEF){
	    return undefined(symbol);
	} else {
	    return val;
	}
    }

    public boolean defined(String name){
	return _get(name.intern()) != UNDEF;
    }

    Object _get(String symbol){
	Binding b = env.lookup(symbol);
	if (b != null){
	    return b.value;
	}

	Function ff = env.func;
	while (ff != null){
	    SymbolTable ls = ff.lexicalScope;
	    if (ls != null){
		Binding bb = ls.lookup0(symbol);
		if (bb != null){
		    return ((Binding)bb.value).value;
		}
	    }
	    ff = ff.outer;
	}
	return _getId(symbol);
    }

    public Object getId(String symbol){
	Object v = _getId(symbol.intern());
	if (v == UNDEF){
	    return undefined(symbol);
	} else {
	    return v;
	}
    }

    public Object _getId(String symbol){
	Package pkg = currentPackage;

	Value b = pkg.lookup(symbol, this);
	if (b != null){
	    return b.get();
	}

	Package global = Package.globalPackage;
	if (pkg != global){
	    b = global.lookup(symbol, this);
	    if (b != null){
		return b.get();
	    }
	}

	if (symbol.charAt(0) == '!'){
	    return null;
	}

	Object c = importEnv.get(symbol);
	if (c != null){
	    return c;
	}

	SymbolTable tab = global.autoloadTable;

	if (tab != null){
	    String fname = (String)tab.get(symbol);
	    if (fname != null){
		try {
		    Context ct = (Context)clone();
		    ct.clear();
		    Pnuts.internalLoad(fname, ct);
		    tab.remove(symbol);
		} catch (PnutsException e){
		    throw new PnutsException("autoload.failed", new Object[]{fname}, this);
		}
		if (global.defined(symbol, this)){
		    return Package.globalPackage.get(symbol, this);
		}
	    }
	}
	return UNDEF;
    }

    final void pushFile(String s){
	loadingResource.push(s);
    }

    final void popFile(){
	if (loadingResource.size() > 0){
	    loadingResource.pop();
	}
    }

    final boolean isLoading(){
	return loadingResource.size() > 0;
    }
    
    final String peekFile(){
	return (String)loadingResource.peek();
    }

    final void clearFile(){
	loadingResource.removeAllElements();
    }

    String getFile(){
	if (frame != null && frame.file != null){
	    return frame.file;
	} else if (isLoading()){
	    return peekFile();
	} else {
	    return null;
	}
    }

    void catchException(Class t, PnutsFunction f){
	if (env.parent != null){
	    Hashtable table = null;
	    Value b = env.lookup(exceptionHandlerTableSymbol);
	    if (b != null){
		table = (Hashtable)b.get();
	    }
	    if (table == null){
		table = new Hashtable(10);
		env.put1(exceptionHandlerTableSymbol, table);
	    }
	    if (f == null){
		table.remove(t);
	    } else {
		table.put(t, f);
	    }
	} else {
	    Runtime.catchException(t, f, this);
	}
    }
}
