/*
 * @(#) Function.java	1.17 99/06/02
 *
 * Copyright (c) 1997-1999 Sun Microsystems, Inc. All Rights Reserved.
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package pnuts.lang;

import java.util.Vector;
import java.io.Serializable;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import pnuts.lang.Runtime;
import pnuts.ext.UnparseVisitor;

/**
 * This class represents a function with a certain number of parameters.
 * In Pnuts, functions should be accessed through PnutsFunction.
 * This class is used mainly by compiler implementors.  Note that
 * there is no way to create a Function object through public API.
 *
 * @version 	1.17
 */
public class Function extends Runtime implements Serializable {

    static final long serialVersionUID = -3237941930735690285L;

    /**
     * name of the function including the scope information
     *
     * @serial
     */
    protected String funcName;

    /**
     * name of the function without the scope information
     *
     * @serial
     */
    protected String name;

    /**
     * the number of arguments
     *
     * @serial
     */
    protected int nargs;

    /**
     * local parameters
     *
     * @serial
     */
    protected String locals[];

    /**
     * file name in which this function is defined
     *
     * @serial
     */
    protected String file;

    /**
     * reference to the definition
     *
     * @serial
     */
    protected SimpleNode node;

    /**
     * @serial
     */
    SymbolTable lexicalScope;

    /**
     * "import" environment
     *
     * @serial
     */
    protected ImportEnv importEnv;

    /**
     * reference to the outer function
     *
     * @serial
     */
    protected Function outer;

    /**
     * the package name in which this function is defined
     *
     * @serial
     */
    protected String pkgName;

    /**
     * reference to PnutsFunction
     *
     * @serial
     */
    protected PnutsFunction function;

    transient protected Package pkg;

    protected void init(){
    }

    protected Function(){
    }

    protected Function(String func, String[] locals, int nargs, SimpleNode node, Package pkg, Context context){
	if (context.frame != null){
	    outer = context.frame;
	    this.file = outer.file;
	    this.importEnv = outer.importEnv;
	} else {
	    if (context.isLoading()){
	        this.file = context.peekFile();
	    }
	    this.importEnv = context.importEnv;
	}

	Function of = outer;

	String scopeName = null;

	if (of != null){
	    scopeName = of.getName() + "$";
	}

	if (func == null){
	    func = "";
	}

	this.name = func;
	this.funcName = func;
	if (scopeName != null){
	    this.funcName = scopeName + func;
	}
	this.funcName = funcName.intern();
	this.nargs = nargs;
	this.locals = locals;
	this.node = node;
	this.pkg = pkg;
	this.pkgName = pkg.getName();

	StackFrame lex = context.env;
	if (lex != null){
	    SymbolTable tab = new SymbolTable();
	    lex.makeLexicalScope(tab);
	    this.lexicalScope = tab;
	}
    }

    public String getName(){
	return funcName;
    }

    protected String getFile(){
	return file;
    }

    protected Object call(Object[] args, Context context){
	ImportEnv saved = context.importEnv;
	Package outerPackage = context.currentPackage;
	boolean eval = context.eval;
	context.importEnv = importEnv;
	context.eval = false;
	context.currentPackage = this.pkg;
	context.frame = this;

	try {
	    return exec(args, context);
	} catch (Escape e){
	    throw e;
	} catch (Throwable t){
	   String msg = null;
	   try {
		String param = Format.listToString(args);
		msg = name + "(" + param.substring(1, param.length() - 1) + ")";
	   } catch (Throwable tt){
		msg = name + "(?)";
	   }

	   PnutsException p = null;
	   if (t instanceof PnutsException){
		p = (PnutsException)t;
	   } else {
		p = new PnutsException(t, context);
	   }
	   if (p.trace == null){
		p.trace = new Vector();
	   } 
	   p.trace.addElement(msg);

	   throw p;

	} finally {
	    context.importEnv = saved;
	    context.eval = eval;
	    context.currentPackage = outerPackage;
 	}
    }

    protected Object exec(Object[] args, Context context){
	try {
	    context.open(locals, this);
	   
	    for (int i = 0; i < args.length; i++){
		context.bind(locals[i], args[i]);
	    }
	    if (outer != null && name != null){
		context.bind(name, function);
	    }
	    return node.accept(Pnuts.interpreter, context);
	} catch (Throwable t){
	   Runtime.checkException(context, t);
	   return null;
	} finally {
	    context.close();
	}
    }

    String paramString(){
        String s = "";
	s += "(";
	if (nargs != 0){
	    s += locals[0];
	}
	if (nargs == -1){
	    s += "[]";
	} else {
	    for (int i = 1; i < nargs; i++){
		s += ",";
		s += locals[i];
	    }
	}
	s += ")";
	return s;
    }

    protected PnutsFunction register(PnutsFunction pf){
	if (pf == null){
	    if (name != null){
		pf = new PnutsFunction(name);
	    } else {
		pf = new PnutsFunction();
	    }
	}
	pf.put(nargs, this);
	return pf;
    }

    protected Object accept(Visitor visitor, Context context){
	return node.jjtGetParent().accept(visitor, context);
    }

    protected String unparse(Context context){
	return (String)accept(new UnparseVisitor(), context);
    }

    public String toString(){
	String s = "function ";
	s += funcName;
	s += paramString();
	return s;
    }

    private synchronized void readObject(ObjectInputStream s) throws IOException, ClassNotFoundException {
	s.defaultReadObject();
	if (pkgName != null){
	    this.pkg = Package.getPackage(pkgName);
	} else {
	    this.pkg = new Package();
	}
	for (int i = 0; i < locals.length; i++){
	    locals[i] = locals[i].intern();
	}
    }
}
