/*
 * @(#) Package.java	1.15 99/06/03
 *
 * Copyright (c) 1997-1999 Sun Microsystems, Inc. All Rights Reserved.
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package pnuts.lang;

import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;
import java.io.Serializable;
import java.io.OutputStream;
import java.io.InputStream;
import java.io.ByteArrayOutputStream;
import java.io.ObjectOutputStream;
import java.io.ObjectInputStream;
import java.io.IOException;
import java.io.FileNotFoundException;

/**
 * This class represents a Pnuts package (not Java's).
 *
 * @version	1.15
 * @author	Toyokazu Tomatsu
 */
public class Package implements Property, Serializable, Cloneable {

    static final long serialVersionUID = 5749562856803991817L;

    /**
     * All packages with a non-null name
     */
    protected static Hashtable packages = new Hashtable(10);

    /**
     * The package with name "".
     */
    public static Package globalPackage = null;

    final static String delimiter = "::";

    transient protected SymbolTable table;
    transient protected Vector children;
    transient protected Package parent;

    /**
     * The name of the package.
     * @serial
     */
    protected String name;

    /**
     * This represents a set of names which are registered as autoloaded.
     * This is only for global package.
     * @serial
     */
    public SymbolTable autoloadTable; // only for global package

    /**
     * This represents a set of registered QuantityFactory.
     * This is only for global package.
     * @serial
     */
    protected SymbolTable unitTable; // only for global package

    /**
     * This represents a set of loaded script names.
     * This is only for global package.
     * @serial
     */
    protected SymbolTable requireTable; // only for global package

    static void initGlobal(){
	packages.put("", getGlobalPackage());
    }

    static PackageFactory factory;

    static {
	try {
	    SecuritySensitive.init();
	    String prop = SecuritySensitive.getProperty("pnuts.package.factory");
	    if (prop != null){
		Object obj = Class.forName(prop).newInstance();
		if (obj instanceof PackageFactory){
		    factory = (PackageFactory)obj;
		}
	    }
	} catch(Throwable e){/* skip */}
    }

    static {
	initGlobal();
    }

    /**
     * @return the global package
     */
    public static Package getGlobalPackage(){
	if (globalPackage == null){
	    globalPackage = getInstance("");
	}
	return globalPackage;
    }

    /**
     * When an instance of PackageFactory is registered by this
     * method, the package() builtin function calls its
     * PackageFactory.createPackage() method.
     *
     * @since	Pnuts 1.0beta3
     */
    public static void setPackageFactory(PackageFactory factory){
	Package.factory = factory;
    }

    static Package getInstance(String name){
	if (name == null){
	    if (factory != null){
		return factory.createPackage(null);
	    } else {
		return new Package();
	    }
	}
	Package p = (Package)packages.get(name);
	if (p == null){
	    if (factory != null){
		p = factory.createPackage(name);
	    } else {
		p = new Package(name);
	    }
	    packages.put(name, p);
	}
	return p;
    }

    /**
     * If package "pkg" exists returns that, otherwise
     * creates and returns one.
     */
    public static Package getPackage(String pkg){
        Package g = globalPackage;
	Package p = g;
	int index = pkg.indexOf(delimiter);
	String rest = pkg;
	String s = "";
	while (index > 0){
	    String t = rest.substring(0, index);
	    if (t.indexOf(".") >= 0 ||
	        t.indexOf(":") >= 0){
	        new PnutsException("illegal package name " + pkg);
            }
	    s += t;

	    if (globalPackage.get(s) instanceof Class){
		new PnutsException("package name and Class conflicted: " + pkg);
	    }
	    Package c = getInstance(s);
	    if (!p.children.contains(c)){
		p.children.addElement(c);
		c.parent = p;
	    }
	    p = c;

	    rest = rest.substring(index + 2);
	    index = rest.indexOf(delimiter);
	    s += delimiter;
        }
	if (rest.indexOf(".") >= 0 ||
	    rest.indexOf(":") >= 0){
	    new PnutsException("illegal package name " + pkg);
        }
	s += rest;

	Package c = getInstance(s);
	if (p != c && !p.children.contains(c)){
	    p.children.addElement(c);
	    c.parent = p;
	}
	return c;
    }

    /**
     * @return true if name is defined in the package.
     */
    public boolean defined(String name){
	return defined(name, null);
    }

    /**
     * @return true if name is defined in the package.
     * @since	Pnuts 1.0beta3
     */
    public boolean defined(String name, Context context){
	return lookup(name.intern(), context) != null;
    }

    /**
     * @return the value of specified variable in the package.
     */
    public Object get(String symbol){
	return get(symbol, null);
    }

    /**
     * Get the value of a symbol in the package.
     * @param symbol a name in the package
     * @param context the context in which the symbol is referenced.  null means "not specified".
     * @return the value of specified variable in the package.
     * @since	Pnuts 1.0beta3
     */
    public Object get(String name, Context context){
	Value b = lookup(name = name.intern(), context);
	if (b == null){
	    return null;
	} else {
	    return b.get();
	}
    }

    /**
     * Set a value of a symbol in the package.
     * @param sym a name of variable
     * @param obj the value of the variable
     */
    public void set(String symbol, Object obj){
	set(symbol, obj, null);
    }

    /**
     * Set a value of a symbol in the package.
     * @param	symbol a name of variable
     * @param	obj the value of the variable
     * @since	Pnuts 1.0beta3
     */
    public void set(String symbol, Object obj, Context context){
	table.put(symbol.intern(), obj);
    }

    /**
     * Delete a symbol from the package
     * @param symbol a name of variable to be deleted
     */
    public void clear(String symbol){
	clear(symbol, null);
    }

    /**
     * Delete a symbol from the package
     * @param symbol a name of variable to be deleted
     */
    public void clear(String symbol, Context context){
	if (table != null){
	    table.remove(symbol);
	}
    }

    /**
     * Remove the specified package
     */
    public static void remove(String name){
	Package p = (Package)packages.get(name);
	if (p == null){
	    return;
	}
	Package p2 = p.parent;
	if (p2 == null){
	    return;
	}
	p2.children.removeElement(p);
        packages.remove(name);
    }

    /**
     * @param pkg a name of package to look
     * @return a package with name "pkg" if it exits.
     */
    public static Package find(String pkg){
        return (Package)packages.get(pkg);
    }

    /**
     * Remove all packages
     */
    public static void reset(){
	packages = new Hashtable(10);
	initGlobal();
    }

    /**
     * creates a package that is not visible from other packages
     */
    public Package(){
	this(null);
    }

    /**
     * creates a package and register it in a static hashtable.
     */
    protected Package(String name){
	if (name != null){
	    this.name = name.intern();
	}
	init();
    }

    /**
     * This method is called by the constructors.
     */
    protected void init(){
	this.table = new SymbolTable(64);
	if (name != null){
	    this.children = new Vector();
	    if ("".equals(name)){
		autoloadTable = new SymbolTable(10);
		unitTable = new SymbolTable(10);
		requireTable = new SymbolTable(10);

		table.put("pnuts_version", Pnuts.pnuts_version);
		table.put("int", Integer.TYPE);
		table.put("short", Short.TYPE);
		table.put("long", Long.TYPE);
		table.put("byte", Byte.TYPE);
		table.put("char", Character.TYPE);
		table.put("float", Float.TYPE);
		table.put("double", Double.TYPE);
		table.put("boolean", Boolean.TYPE);

		PnutsFunction[] prim = PnutsFunction.primitives;
		for (int i = 0; i < prim.length; i++){
		    table.put(prim[i].getName(), prim[i]);
		}
	    }
	}
	/*
	notToSerialize = new Hashtable(10);
	bout = new ByteArrayOutputStream();
	try {
	    out = new ObjectOutputStream(bout);
	} catch (IOException e){
	    throw new PnutsException(e);
	}
	*/
    }

    /**
     * This method is called when the package become the current package with package() function.
     * This method in a subclass must call super.init(context) first.
     *
     * @since	Pnuts 1.0beta3
     */
    protected void init(Context context){
	if ("".equals(name)){
	    globalPackage = this;
	}
	if (name != null){
	    packages.put(name, this);
	}
	context.currentPackage = this;
    }

    /**
     * @return the name of the package
     */
    public String getName(){
	return name;
    }

    /**
     * Returns an enumeration of the names in the package.
     */
    public Enumeration keys(){
	return table.keys();
    }

    /**
     * the number of symbols
     */
    public int size(){
	return table.size();
    }

    /**
     * lookup the symbol in the package
     *
     * @param symbol intern'ed string
     */
    protected Value lookup(String symbol, Context context){
	return table.lookup0(symbol);
    }

    transient SymbolTable notToSerialize;
    transient ByteArrayOutputStream bout;
    transient ObjectOutputStream out;

    /**
     * save the package into specified output stream.
     */
    void prepareToSave(){
	int size = table.size();
	for (int i = 0; i < size; i++){
	    try {
		bout = new ByteArrayOutputStream();
		out = new ObjectOutputStream(bout);
		out.writeObject(this);
		break;
	    } catch (IOException e){
		/* ignore */
	    }
	}
    }

    /**
     * save the package to an outputStream.
     * If a value of a symbol can not be serialized, the symbol is
     * ignored.
     */
    public void save(OutputStream s){
	prepareToSave();
	try {
	    bout.writeTo(s);
	} catch (IOException e){
	    throw new PnutsException(e);
	} finally {
	    out = null;
	    bout = null;
	}
    }

    private synchronized void writeObject(ObjectOutputStream s)
        throws IOException
    {
	s.defaultWriteObject();
	if (notToSerialize == null){
	    notToSerialize = new SymbolTable(10);
	}
	Runtime.writeSymbolTable(table, notToSerialize, s);
    }

    private synchronized void readObject(ObjectInputStream s)
         throws IOException, ClassNotFoundException
    {

	s.defaultReadObject();
	table = new SymbolTable();
	children = new Vector();
	Runtime.readSymbolTable(table, s);
	notToSerialize = null;
    }

    /**
     * enumerates sub-packages
     */
    public Enumeration elements(){
	return children.elements();
    }

    static void provide(String file) {
	if (file.endsWith(".pnut")){
	    file = file.substring(0, file.length() - 5);
	}
	globalPackage.requireTable.put(file, file);
    }

    static void require(String file) throws FileNotFoundException {
	if (file.endsWith(".pnut")){
	    file = file.substring(0, file.length() - 5);
	}
	if (globalPackage.requireTable.get(file) == null){
	    Pnuts.load(file);
	}
    }

    public static void registerQuantityFactory(String unit, QuantityFactory fac){
	if (fac != null){
	    globalPackage.unitTable.put(unit, fac);
	} else {
	    globalPackage.unitTable.remove(unit);
	}
    }

    public Object clone(){
	try {
	    Package p = (Package)super.clone();
	    for (Enumeration e = keys(); e.hasMoreElements(); ){
		String key = (String)e.nextElement();
		p.set(key, get(key));
	    }
	    return p;
	} catch (Throwable t){
	    throw new InternalError();
	}
    }

    public String toString(){
	if (name != null){
	    return "package \"" + name + "\"";
	} else {
	    return "package null";
	}
    }
}
