/*
 * %Z% %M%	%I% %E%
 *
 * Copyright (c) 1997-1999 Sun Microsystems, Inc. All Rights Reserved.
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package pnuts.lang;

import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.PrintWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.BufferedInputStream;
import java.io.StringReader;
import java.io.IOException;
import java.io.FileNotFoundException;
import java.io.Serializable;
import java.io.ObjectInputStream;
import java.net.URL;
import java.util.StringTokenizer;
import java.util.Hashtable;
import java.applet.Applet;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;

/**
 * "Manager" class for Pnuts interpreter.<br>
 * "pnuts" command starts from main method of this class.<br>
 * Annotated reference manual is <a href="/pnuts/doc/api.html">here</a>.
 *
 * @version	%I%
 * @author	Toyokazu Tomatsu
 */
public class Pnuts implements Serializable {

    static final long serialVersionUID = 3374216268395731960L;

    /**
     * If Pnuts starts in applet this variable should be set.
     */
    public static Applet applet = null;

    static String java_version;
    static String pnuts_version = "1.0beta4";

    static String greetingMessage = "Pnuts interpreter Version " + pnuts_version;
    static String copyrightMessage= "Copyright (c) 1997-1999 Sun Microsystems, Inc. All rights reserved.";

    /**
     * "prompt" string for the command shell
     */
    public static String prompt = "> ";

    static String java_vendor;

    static Hashtable nestCount = new Hashtable(10);
    static Hashtable callCount = new Hashtable(10);

    final static String load_path_sym = "pnuts_load_path";

    protected static Visitor interpreter = new PnutsInterpreter();

    static ClassLoader classLoader;

    static boolean verbose = false;

    private static Stack freeParsers = new Stack();
    static {
	freeParsers.push(new PnutsParser((Reader)null));
    }

    static Context contextTemplate;

    static {
	Package.getGlobalPackage().set(load_path_sym, new Object[]{"/"});
	contextTemplate = new Context();
    }

    private static int java2 = 0;

    final public static boolean isJava2(){
	if (java2 == 0){
	    try {
		if (java_version == null){
		    java_version = System.getProperty("java.version");
		}
		String version = java_version;
		int idx = 0;
		for (; idx < version.length(); idx++){
		    if (Character.isDigit(version.charAt(idx))){
			break;
		    }
		}
		if (idx > 0 && idx < version.length()){
		    version = version.substring(idx);
		}
		if (version.startsWith("1.0") ||
		    version.startsWith("1.1.") ||
		    "1.1".equals(version)){
		    java2 = -1;
		} else {
		    java2 = 1;
		}
	    } catch (Throwable t){
		java2 = -1;
	    }
	}
	if (java2 == 1){
	    return true;
	} else {
	    return false;
	}
    }

    public static void registerQuantityFactory(String unit, QuantityFactory factory){
	Package.registerQuantityFactory(unit, factory);
    }

    /**
     * set ClassLoader for class access
     * @param	loader the ClassLoader with which classes are accessed
     */
    public static void setClassLoader(ClassLoader loader){
	classLoader = loader;
    }

    /**
     * get ClassLoader for class access
     */
    public static ClassLoader getClassLoader(){
	return classLoader;
    }

    public final static Class loadClass(String name) throws ClassNotFoundException {
	if (classLoader != null){
	    if (isJava2()){
		return Class.forName(name, true, classLoader);
	    } else {
		return classLoader.loadClass(name);
	    }
	} else {
	    if (isJava2()){
		ClassLoader cl = null;
		try {
		    cl = ClassLoader.getSystemClassLoader();
		} catch (SecurityException t){/* skip */}
		if (cl != null){
		    return Class.forName(name, true, cl);
		}
	    }
	    return Class.forName(name);
	}
    }

    /**
     * set a "prompt" string for the command shell
     */
    public static void setPrompt(String str){
	prompt = str;
    }

    /**
     * set verbose mode
     */
    public static void setVerbose(boolean b) {
	verbose = b;
    }

    public static boolean isVerbose() {
	return verbose;
    }

    /**
     * set a search path for scripts as an object array
     * Elements of the array should one of the followings.
     *    String:  base name of resource
     *    URL:     base name of URL
     *    File:    base directory
     */
    public static void setLoadPath(Object paths[]){
	set(load_path_sym, paths);
    }

    /**
     * set a search path for scripts as a colon-separated String
     * This method remains for compatibility reason. It will be deleted in the future.
     */
    public static void setLoadPath(String path){
	String sep = ":";
	StringTokenizer st = new StringTokenizer(path, sep);
	int n = st.countTokens();
	Object loadPath[] = new Object[n];
	for (int i = 0; i < n; i++){
	    loadPath[i] = st.nextToken();
	}
	set(load_path_sym, loadPath);
    }

    private static int debug = 0;

    static boolean debug(){
	try {
	    if (debug == 0){
		if (applet == null && Boolean.getBoolean("pnuts.debug") ||
		    applet != null && applet.getParameter("pnuts.debug") != null){
		    debug = 1;
		} else {
		    debug = -1;
		}
	    }
	    return (debug == 1);
	} catch (Throwable t){
	    debug = -1;
	    return false;
	}
    }

    static InputStream findFile(String name, Context context) {
	Object o = Package.getGlobalPackage().get(load_path_sym, context);

	if (!(o instanceof Object[])){
	    throw new PnutsException("loadpath.notArray", new Object[]{o}, context);
	}
	Object loadPath[] = (Object[])o;
	if (java_vendor == null){
	    java_vendor = getProperty("java.vendor");
	}

	for (int i = 0; i < loadPath.length; i++){
	    InputStream in = null;
	    Object obj = loadPath[i];
	    if (obj instanceof String){
		String s = (String)obj;
		if (s != "" && !s.endsWith("/")){
		    s += "/";
		}
		s += name;
		if (applet != null &&
		    java_vendor != null && java_vendor.startsWith("Netscape")){
		    /*
		     * workaround for resource bug
		     */
		    try {
			URL u = new URL(applet.getCodeBase(), s.substring(1));
			in = u.openStream();
		    } catch (IOException e){}
		} else {
		    in = getResourceAsStream(s);
		}
		if (in != null){
		    if (verbose){
			PrintWriter out = context.getOutputStream();
			out.println("[loading " + s + "]");
			out.flush();
		    }
		    return in;
		}
	    } else if (obj instanceof File){
		File dir = (File)obj;
		if (!dir.isDirectory()){
		    return null;
		}
		File f = new File(dir, name);
		if (f.exists()){
		    try {
			in = new FileInputStream(f);
		    } catch (FileNotFoundException e){ /* ignore */ }
		} else {
		    continue;
		}
		if (in != null){
		    return in;
		}
	    } else if (obj instanceof URL){
		try {
		    URL url = new URL((URL)obj, name);
		    in = url.openStream();
		} catch (IOException e){ /* ignore */ }
		if (in != null){
		    return in;
		}
	    }
	}
	return null;
    }

    private static String getProperty(String s){
	return SecuritySensitive.getProperty(s);
    }

    static InputStream getResourceAsStream(String s){
	if (classLoader != null){
	    if (isJava2()){
		InputStream in = classLoader.getResourceAsStream(s);
		if (in == null){
		    return SecuritySensitive.getResourceAsStream(s);
		} else {
		    return in;
		}
	    } else {
		return classLoader.getResourceAsStream(s);
	    }
	} else {
	    return SecuritySensitive.getResourceAsStream(s);
	}
    }

    /**
     * returns a formatted string for obj
     */
    public static String format(Object obj){
	return Format.listToString(obj);
    }

    static void greeting(Context context){
	PrintWriter w = context.getTerminalStream();
	w.println(copyrightMessage);
	if (java_vendor == null){
	    java_vendor = getProperty("java.vendor");
	}
	w.println(greetingMessage + ", " + java_version + " (" + java_vendor + ")");
	if (verbose){
	    String vm_name = getProperty("java.vm.name");
	    String vm_info = getProperty("java.vm.info");
	    if (vm_name != null){
		w.println(vm_name + " (" + vm_info + ")");
	    }
	}
    }

    public static Object get(String str){
	return get(str, "");
    }

    public static Object get(String str, String pkg){
	Package p = Package.find(pkg);
	if (p != null){
	    return p.get(str);
	} else {
	    return null;
	}
    }

    /**
     * set a value "val" to a global variable "str"
     */
    public static void set(String str, Object val){
	set(str, val, "");
    }

    /**
     * set a value "val" to a variable "str" in package "pkg"
     */
    public static void set(String str, Object val, String pkg){
	if (str.length() > 0 && Character.isJavaIdentifierStart(str.charAt(0))){
	    Package p = Package.find(pkg);
	    if (p != null){
		p.set(str, val);
	    } else {
		throw new IllegalArgumentException(PnutsException.getMessage("package.notFound", new Object[]{pkg}));
	    }
	} else {
	    throw new IllegalArgumentException(PnutsException.getMessage("illegal.symbolForId", PnutsException.NO_PARAM));
	}
    }

    /**
     * evaluate "str" in package "pkg"
     */
    public static Object eval(String str, String pkg){
	return eval(str, new Context(pkg));
    }

    /**
     * evaluate "str" in the global package
     */
    public static Object eval(String str){
	return eval(str, (Context)contextTemplate.clone());
    }

    /**
     * evaluate "str" in "context"
     */
    public static Object eval(String str, Context context){
	return new Pnuts(new StringReader(str), false, context).value;
    }

    /**
     * load a local script "file"
     */
    public static Object loadFile(String file) throws FileNotFoundException {
	return loadFile(file, (Context)contextTemplate.clone());
    }

    /**
     * load a local script "file" in "context"
     */
    public static Object loadFile(String file, Context context) throws FileNotFoundException {
	String path = file;
	try {
	    path = new File(file).getCanonicalPath();
	} catch (IOException e){ /* ignore */ }
	FileInputStream fin = new FileInputStream(path);
	context.pushFile(file);
	try {
	    return load(fin, context);
	} finally {
	    context.popFile();
	}
    }

    static Object internalLoad(String file, Context context){
	try {
	    return load(file, context);
	} catch (Throwable fnf){
	    throw new PnutsException(fnf, context);
	}
    }

    static Object internalLoadFile(String file, Context context){
	try {
	    return loadFile(file, context);
	} catch (Throwable e){
	    throw new PnutsException(e, context);
	}
    }

    /**
     * load a script "file"
     */
    public static Object load(String file) throws FileNotFoundException {
	return load(file, (Context)contextTemplate.clone());
    }

    /**
     * load a script "file" in "context"
     */
    public static Object load(String file, Context context) throws FileNotFoundException {
	if (!file.endsWith(".pnut") && !file.endsWith(".pnuts")){
	    file = file + ".pnut";
	}
	final String f = file;
	final Context c = context;
	InputStream rsrc = findFile(file, context);

	if (rsrc != null){
	    context.pushFile(file);
	    try {
		Object r = load(new BufferedInputStream(rsrc), context);
		provide(file);
		return r;
	    } finally {
		context.popFile();
	    }
	} else {
	    throw new FileNotFoundException(file);
	}
    }

    /**
     * load a script from InputStream "in"
     */
    public static Object load(InputStream in) {
	return load(in, (Context)contextTemplate.clone());
    }

    /**
     * load a script from InputStream "in" in "context"
     */
    public static Object load(InputStream in, Context context) {
	return load(in, false, context);
    }

    /**
     * Load a script from an InputStream in the specified Context.
     *
     * @param in an InputStream from which the interpreter reads an input
     * @param interactive
     * <ul>
     * <li>When "interactive" is true, the greeting message, the prompt, and the results of
     * evaluations are displayed.  When an exception is thrown and not caught by any exception
     * handler, it is caught at the top level of the interpreter, display an error message, 
     * and resume the interactive session.  If the exception is caught by a handler that is
     * registered at the top level, the result of the handler becomes the return value of the
     * last expression.
     *
     * <li>When "interactive" is false, exceptions are caught at the top level of the interpreter
     * and exits this function.  If the exception thrown is caught by a handler that
     * is registered at the top level,  the result of the handler becomes the return value of
     * this method.
     * </ul>
     * @param context a Context in which the interpretation is taken place.
     * @return the result of the last expression
     */
    public static Object load(InputStream in,
			      boolean interactive,
			      Context context) {
	return load(new InputStreamReader(in), interactive, context);
    }

    /**
     * load a script from Reader "reader"
     */
    public static Object load(Reader reader) {
	return load(reader, (Context)contextTemplate.clone());
    }

    /**
     * load a script from Reader "reader" in context
     */
    public static Object load(Reader reader, Context context) {
	return load(reader, false, context);
    }

    public static Object load(Reader reader, boolean interactive){
	return load(reader, interactive, (Context)contextTemplate.clone());
    }

    /**
     * Load a script from an Reader in the specified Context.
     *
     * @param reader an Reader from which the interpreter reads an input
     * @param interactive
     * <ul>
     * <li>When "interactive" is true, the greeting message, the prompt, and the results of
     * evaluations are displayed.  When an exception is thrown and not caught by any exception
     * handler, it is caught at the top level of the interpreter, display an error message, 
     * and resume the interactive session.  If the exception is caught by a handler that is
     * registered at the top level, the result of the handler becomes the return value of the
     * last expression.
     *
     * <li>When "interactive" is false, exceptions are caught at the top level of the interpreter
     * and exits this function.  If the exception thrown is caught by a handler that
     * is registered at the top level,  the result of the handler becomes the return value of
     * this method.
     * </ul>
     * @param context a Context in which the interpretation is taken place.
     * @return the result of the last expression
     */
    public static Object load(Reader reader,
			      boolean interactive,
			      Context context) {
	return new Pnuts(reader, interactive, context).value;
    }

    /**
     * parse a script from InputStream and return a Pnuts object
     *
     * @return the Pnuts object including a parsed syntax tree
     * @param in the InputStream
     * @since Pnuts 1.0beta3
     */
    public static Pnuts parse(InputStream in) throws ParseException {
	return parse(new InputStreamReader(in));
    }

    /**
     * parse a script from Reader and return a Pnuts object
     *
     * @return the Pnuts object including a parsed syntax tree
     * @param reader the Reader
     * @since Pnuts 1.0beta3
     */
    public static Pnuts parse(Reader reader) throws ParseException {
	return new Pnuts(reader);
    }

    /**
     * parse a script and return a Pnuts object
     *
     * @return the Pnuts object including a parsed syntax tree
     * @param expr the script
     * @since Pnuts 1.0beta3
     */
    public static Pnuts parse(String expr) throws ParseException {
	return new Pnuts(new StringReader(expr));
    }

    /**
     * load a script "file" only when the script has not been read.
     */
    public static void require(String file) throws FileNotFoundException {
	Package.require(file);
    }

    static void provide(String file){
	Package.provide(file);
    }

    void loop(PnutsParser parser, boolean interactive, Context context){
	while (true) {
	    try {
		if (interactive){
		    PrintWriter out = context.getOutputStream();
		    if (out != null){
			out.flush();
		    }
		    PrintWriter term = context.getTerminalStream();
		    if (term != null){
			term.print(prompt);
			term.flush();
		    }
		}
		
		SimpleNode start = parser.Start();

		switch (getCode(start)){
		case -1:
		    ((int[])start.info)[0] = 0;
		    return;
		case 0:
		    reset(parser);
		    break;
		case 1:
		    if (interactive){
			try { 
			    value = start.accept(interpreter, context);
			} catch (Jump jump){
			    value = jump.getValue();
			}
		    } else {
			value = start.accept(interpreter, context);
		    }
		    if (getCode(start) == -1){
			value = start.value;
			return;
		    }
		    PrintWriter term = context.getTerminalStream();
		    if (interactive && term != null){
			try {
			    term.println(format(value));
			    term.flush();
			} catch (Throwable t){
			    Runtime.checkException(context, t);
			}
		    }
		    reset(parser);
		    break;
		}
		    
	    } catch (ParseException p){
		error_skipto(parser, PnutsParserConstants.EOL);

		if ((context instanceof ManualContext) ||
		    !interactive ||
		    getI(Thread.currentThread(), nestCount)[0] > 1){

 		    Runtime.checkException(context, p);
		} else {
		    PrintWriter err = context.getErrorStream();
		    try {
			context.beginLine = p.currentToken.next.beginLine;
			Runtime.checkException(context, p);
		    } catch (Jump j){
			value = j.getValue();
			PrintWriter term = context.getTerminalStream();
			if (interactive && term != null){
			    term.println(value);
			    term.flush();
			}
		    } catch (PnutsException t){
			if (err != null){
			    if (verbose){
				t.printStackTrace(err);
			    } else {
				err.println(t);
			    }
			    err.flush();
			}
		    }
		    reset(parser);
		    recover(context);
		    value = null;
		}
	    } catch (Escape esc){
		reset(parser);
		throw esc;
	    } catch (Throwable t){
		if ((context instanceof ManualContext) ||
		    !interactive ||
		    getI(Thread.currentThread(), nestCount)[0] > 1){

		    if (t instanceof PnutsException){
			throw (PnutsException)t;
		    } else {
			throw new PnutsException(t, context);
		    }
		}
		PrintWriter err = context.getErrorStream();
		if (err != null){
		    if (verbose){
			t.printStackTrace(err);
		    } else {
			err.println(t);
		    }
		    err.flush();
		}
		reset(parser);
		recover(context);
		value = null;
	    }
	}
    }


    static void error_skipto(PnutsParser parser, int kind){
	Token t;
	do {
	    t = parser.getNextToken();
	} while (t != null && t.image.length() > 0 && t.kind != kind);
    }


    /**
     * parsed scripts
     *
     * @serial
     */
    protected SimpleNode startNodes = null;

    transient protected Object value;

    protected Pnuts(){
    }

    protected Pnuts(Reader reader) throws ParseException {
	PnutsParser parser;
	if (freeParsers.size() > 0){
	    parser = (PnutsParser)freeParsers.pop();
	    parser.ReInit(reader);
	} else {
	    parser = new PnutsParser(reader);
	}
	try {
	    startNodes = parser.StartSet();
	} finally {
	    if (freeParsers.size() < 3){
		freeParsers.push(parser);
	    }
	}
    }

    /**
     * executes a Pnuts object with a new Context
     *
     * @since Pnuts 1.0beta3
     */
    public Object run(){
	return run((Context)contextTemplate.clone());
    }

    /**
     * executes a Pnuts object with the specified Context
     *
     * @param context the Context
     * @since Pnuts 1.0beta3
     */
    public Object run(Context context){
	return accept(interpreter, context);
    }

    /**
     * traverse the start nodes with the specified Visitor and Context
     *
     * @param context the Context
     * @since Pnuts 1.0beta3
     */
    public Object accept(Visitor visitor, Context context) {
	try {
	    return startNodes.accept(visitor, context);
	} catch (Jump jump){
	    return jump.getValue();
	}
    }

    /**
     * Starts Pnuts interpreter. If <em>interactive</em> is true
     * prompt message and the results of evaluation are printed
     * on the screen.
     */
    protected Pnuts(Reader reader, boolean interactive, Context context){
	PnutsParser parser;
	if (freeParsers.size() > 0){
	    parser = (PnutsParser)freeParsers.pop();
	    parser.ReInit(reader);
	} else {
	    parser = new PnutsParser(reader);
	}

	Thread th = Thread.currentThread();
	int min[] = getI(th, nestCount);
	if (min[0] == Integer.MAX_VALUE){
	    throw new PnutsException("evalDepth overflow");
	} else {
	    ++min[0];
	}

	try {
	    loop(parser, interactive, context);
	} catch (Escape esc){
	    value = esc.getValue();
	} catch (PnutsException t){
	    if ((context instanceof ManualContext) || min[0] > 1){
		throw t;
	    }
	    PrintWriter err = context.getErrorStream();
	    if (err != null){
		if (verbose){
		    t.printStackTrace(err);
		} else {
		    err.println(t);
		}
		err.flush();
	    }
	} finally {
	    if (min[0] == 0){
		throw new PnutsException("evalDepth overflow");
	    } else {
		--min[0];
	    }
	}

	if (freeParsers.size() < 3){
	    freeParsers.push(parser);
	}
    }

    /**
     * returns the depth of evaluations
     */
    public static synchronized int evalDepth(){
	Thread th = Thread.currentThread();
	long l = (long)getI(th, nestCount)[0] +
	         (long)getI(th, callCount)[0];
	if (l > Integer.MAX_VALUE){
	    return Integer.MAX_VALUE;
	} else {
	    return (int)l;
	}
    }

    final static int[] getI(Thread th, Hashtable tab){
	int[] min = (int[])tab.get(th);
	if (min == null){
	    tab.put(th, min = new int[]{0});
	}
	return min;
    }

    /**
     * starts the command shell interpreter
     */
    public static void main(String args[]) throws Throwable {
	String param = null;

	param = getProperty("pnuts.loadpath");
	if (param != null){
	    setLoadPath(param);
	}

	Context context = new Context();

	boolean evaluated = false;
	boolean interactive = false;
	boolean quiet = false;

	String iparam = getProperty("pnuts.interactive");
	if (iparam != null){
	    if ("false".equals(iparam)){
		quiet = true;
	    } else if ("true".equals(iparam)){
		interactive = true;
	    }
	}

	for (int i = 0; i < args.length; i++) {
	    if ("-r".equals(args[i])){
		if (++i < args.length){
		    load(args[i]);
		    continue;
		} else {
		    break;
		}
	    } else if ("-e".equals(args[i])){
		if (++i < args.length){
		    eval(args[i], context);
		    evaluated = true;
		    continue;
		} else {
		    break;
		}
	    } else if ("-v".equals(args[i])){
		verbose = true;
	    } else {
		evaluated = true;
		loadFile(args[i]);
	    }
	}
	if (evaluated){
	    context.getOutputStream().flush();
	    return;
	}

	if (quiet){
	    interactive = false;
	} else if (!interactive){
	    int av = 0;
	    try {
		av = System.in.available();
	    } catch (IOException e){
	    }
	    if (av < 1){
		interactive = true;
	    } else if (av == 1){ // Some implementations of JDK1.1-win32 has a bug in System.in.available()
		if (!isJava2()){
		    String osname = getProperty("os.name");
		    if (osname != null && osname.startsWith("Win") &&
			java_vendor != null && 
			(java_vendor.startsWith("Sun") || java_vendor.startsWith("IBM"))){
			interactive = true;
		    }
		}
	    }
	}
	    
	if (interactive){
	    greeting(context);
	    load(new TerminalInputStream(System.in), true, context);
	} else {
	    load(System.in, false, context);
	}
	context.getOutputStream().flush();
    }

    static void reset(PnutsParser parser){
	parser.jjtree.reset();
    }

    static void recover(Context context){
	context.reset();
	context.clearFile();
    }

    final static int getCode(SimpleNode node){
	return ((int[])node.info)[0];
    }
}
