/*
 * @(#) PnutsFunction.java	1.20 99/06/03
 *
 * Copyright (c) 1997-1999 Sun Microsystems, Inc. All Rights Reserved.
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package pnuts.lang;

import java.io.StringReader;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.IOException;
import java.io.Serializable;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;

/**
 * A PnutsFunction represents a group of Pnuts functions with a same name.
 *
 * @version	1.20
 * @author	Toyokazu Tomatsu
 */
public class PnutsFunction implements Serializable {

    static final long serialVersionUID = 7844128309015973149L;

    /**
     * @serial
     */
    Vector vec = new Vector(5);

    /**
     * the name
     * @serial
     */
    String name;

    /**
     * true if the function is built-in.
     * @serial
     */
    boolean builtin;

    final static String _getContext = "getContext".intern();
    final static String _package = "package".intern();
    final static String _import = "import".intern();
    final static String _throw = "throw".intern();
    final static String _catch = "catch".intern();
    final static String _eval = "eval".intern();
    final static String _loadFile = "loadFile".intern();
    final static String _load = "load".intern();
    final static String _quit = "quit".intern();
    final static String _defined = "defined".intern();
    public final static PnutsFunction GET_CONTEXT = new PnutsFunction(_getContext);
    public final static PnutsFunction CATCH = new PnutsFunction(_catch);
    public final static PnutsFunction PACKAGE= new PnutsFunction(_package);
    public final static PnutsFunction IMPORT = new PnutsFunction(_import);
    public final static PnutsFunction THROW = new PnutsFunction(_throw);
    public final static PnutsFunction EVAL = new PnutsFunction(_eval);
    public final static PnutsFunction LOAD_FILE = new PnutsFunction(_loadFile);
    public final static PnutsFunction LOAD = new PnutsFunction(_load);
    public final static PnutsFunction QUIT = new PnutsFunction(_quit);
    public final static PnutsFunction DEFINED =  new PnutsFunction(_defined);
    public final static PnutsFunction  primitives[] = {GET_CONTEXT, PACKAGE, IMPORT, THROW,
						       CATCH, EVAL, LOAD_FILE, LOAD, QUIT, DEFINED};
    public final static String  primitive_names[] = {_getContext, _package, _import, _throw,
					      _catch, _eval, _loadFile, _load, _quit, _defined};
    static {
	for (int i = 0; i < primitives.length; i++){
	    primitives[i].builtin = true;
	}
    }

    protected PnutsFunction(){
    }

    protected PnutsFunction(String name){
	this.name = name.intern();
    }

    /**
     * register "name" as an autoloaded function.
     * Later "file" is automatically loaded if the function is not defined 
     */
    public static void autoload(String name, String file){
	if (file == null){
	    Package.globalPackage.autoloadTable.remove(name.intern());
	} else {
	    Package.globalPackage.autoloadTable.put(name.intern(), file);
	}
    }

    protected void put(int narg, Function f){
	int n = narg + 1;
	if (vec.size() < n + 1){
	    vec.setSize(n + 1);
	}
	vec.setElementAt(f, n);
	f.function = this;
    }

    protected Function get(int narg){
	int n = narg + 1;
	if (n >= vec.size()){
	    return null;
	}
	return (Function)vec.elementAt(n);
    }

    /**
     * Check if the function with narg parameter is defined
     *
     * @param narg the number of paramters.  -1 means a arbitrary length parameter.
     * @return true if a function with narg, otherwise false
     */
    public boolean defined(int narg){
	return get(narg) instanceof Function;
    }

    /**
     * @return the unique name
     */
    public String getId(){
	return "function_" + hashCode();
    }

    /**
     * @return the name of functions
     */
    public String getName(){
	return name;
    }

    /**
     * Call a function in "context" with arguments "args".
     * Increments the counter for Pnuts.evalDepth() during
     * the execution.
     *
     * @return the result of the call
     */
    public final Object call(Object[] args, Context context){
	int min[] = Pnuts.getI(Thread.currentThread(), Pnuts.callCount);
	if (min[0] == Integer.MAX_VALUE){
	    throw new PnutsException("evalDepth overflow");
	} else {
	    ++min[0];
	}

	Context c = null;
	try {
	    Function f = get(args.length);
	    c = context;
	    if (f != null){
		c = (Context)context.clone();
	    }
	    return exec(args, c);
	} finally {
	    if (min[0] == 0){
		throw new PnutsException("evalDepth overflow");
	    } else {
		--min[0];
	    }
	}
    }

    /**
     * Call a function in "context" with arguments "args".
     * Subclasses of this class should override this method.
     *
     * @return the result of the call
     */
    protected Object exec(Object[] args, Context context){
	try {
	    if (builtin){
		return call_builtin(args, context);
	    }
	    Function f = (Function)get(args.length);
	    if (f != null){
		return f.call(args, context);
	    } else {
		f = (Function)get(-1); // arbitrary length paramter
		if (f != null){
		    return f.call(new Object[]{args}, context);
		}
		throw new PnutsException("function.notDefined",
					 new Object[]{getName(), new Integer(args.length)},
					 context);
	    }
	} catch (Jump jump){
	    if (_throw == name && args.length == 1){
		throw jump;
	    }
	    return jump.getValue();
	}
    }

    Object call_builtin(Object[] args, Context context){
	if (name == _getContext){
	    if (args.length == 0){
		return context;
	    }
	} else if (name == _package){
	    if (args.length == 1){
		Package pkg;
		if (args[0] instanceof Package){
		    pkg = (Package)args[0];
		} else {
		    pkg = Package.getInstance((String)args[0]);
		}
		pkg.init(context);
		return context.currentPackage;
	    } else if (args.length == 0){
		return context.currentPackage;
	    }
	} else if (name == _import){
	    if (args.length == 1){
		if (args[0] == null){
		    context.clearImport();
		    return null;
		} else {
		    String s = (String)args[0];
		    if (!s.endsWith("*")){
			try {
			    context.addImport(Pnuts.loadClass(s));
			} catch (ClassNotFoundException e){ }
		    } else {
			int idx = s.lastIndexOf('.');
			if (idx > 0){
			    context.addImport(s.substring(0, idx));
			} else {
			    context.addImport("");
			}
		    }
		    return null;
		}
	    } else if (args.length == 0){
		return context.importEnv.list();
	    }
	} else if (name == _throw){
	    if (args.length == 1){
		if (args[0] instanceof Throwable){
		    throw new PnutsException((Throwable)args[0], context);
		} else {
		    throw new PnutsException(args[0].toString(), context);
		}
	    }
	} else if (name == _catch){
	    if (args.length == 2){
		context.catchException((Class)args[0], (PnutsFunction)args[1]);
		return null;
	    }
	} else if (name == _defined){
	    if (args.length == 1){
		return new Boolean(context.defined((String)args[0]));
	    }
	} else if (name == _quit){
	    if (args.length == 0){
		throw new Escape();
	    } else {
		throw new Escape(args[0]);
	    }
	} else if (name == _eval){
	    Context c = null;
	    if (args.length == 1){
		c = (Context)((Context)context).clone();
		c.parent = context;
		c.eval = true;
		return Pnuts.eval((String)args[0], c);
	    }
	    if (args.length == 2){
		if (args[1] instanceof Context){
		    c = (Context)((Context)args[1]).clone();
		} else {
		    c = new Context((String)args[1]);
		}
		c.parent = context;
		c.eval = true;
		return Pnuts.eval((String)args[0], c);
	    }
	} else if (name == _loadFile){
	    if (args.length == 1){
		Context c = (Context)context.clone();
		c.clear();
		return Pnuts.internalLoadFile((String)args[0], c);
	    }
	} else if (name == _load){
	    if (args.length == 1){
		Context c = (Context)context.clone();
		c.clear();
		if (args[0] instanceof InputStream){
		    return Pnuts.load((InputStream)args[0], c);
		} else {
		    return Pnuts.internalLoad((String)args[0], c);
		}
	    }
	}
	throw new PnutsException("function.notDefined",
				 new Object[]{name, new Integer(args.length)},
				 context);
    }

    /**
     * call a function with arguments "args"
     *
     * @return the result of the call
     */
    public final Object call(Object[] args){
	Context context = new Context();
	Object ret = call(args, context);
	context.getOutputStream().flush();
	return ret;
    }

    public String toString(){
	if (builtin){
	    return "<builtin " + name + ">";
	}
	StringBuffer buf = new StringBuffer();
	if (vec == null){
	    return "function " + getName();
	}
	Enumeration e = vec.elements();

	while (e.hasMoreElements()){
	    Object f = e.nextElement();
	    if (f != null){
		buf.append(f.toString());
		break;
	    }
	}
	while (e.hasMoreElements()){
	    Function f = (Function)e.nextElement();
	    if (f != null){
		buf.append(",");
		buf.append(f.paramString());
	    }
	}
	return buf.toString();
    }


    /**
     * call a function "pkg::name" with arguments "args"
     */
    public static Object call(String name, Object args[], String pkg){
	return call(name, args, new Context(pkg));
    }

    /**
     * call a function "name" in "context" with arguments "args"
     */
    public static Object call(String name, Object args[], Context context){
	Object o = context.getId(name);
	if (o instanceof PnutsFunction){
	    PnutsFunction func = (PnutsFunction)o;
	    Object ret = func.call(args, context);
	    context.getOutputStream().flush();
	    return ret;
	} else {
	    throw new PnutsException("function.notDefined", new Object[]{name, new Integer(args.length)}, context);
	}
    }

    /**
     * call a function "name" with arguments "args"
     */
    public static Object call(String name, Object args[]){
	return call(name, args, "");
    }

    /**
     * Retrieve the symbolic definition of a function.
     *
     * @param narg  the number of paramters.  -1 means a arbitrary length parameter.
     * @return the function definition
     */
    public String unparse(int narg){
	if (!builtin){
	    Function f = (Function)get(narg);
	    if (f != null){
		return f.unparse(null);
	    }
	}
	return null;
    }

    /**
     * @param narg  the number of paramters.  -1 means a arbitrary length parameter.
     * @return Package in which the function is defined
     */
    public Package getPackage(int narg){
	if (builtin){
	    return Package.globalPackage;
	}
	Function f = (Function)get(narg);
	if (f != null){
	    return f.pkg;
	}
	return null;
    }

    /**
     * @param narg  the number of paramters.  -1 means a arbitrary length parameter.
     * @return imports of the function (array of Class or String)
     */
    public Object[] getImportEnv(int narg){
	if (builtin){
	    return null;
	}
	Function f = (Function)get(narg);
	if (f != null){
	    return f.importEnv.list();
	}
	return null;
    }

    public boolean isBuiltin(){
	return builtin;
    }

    /**
     * @param narg  the number of paramters.  -1 means a arbitrary length parameter.
     */
    public Object accept(int narg, Visitor visitor, Context context){
	if (!builtin){
	    Function f = (Function)get(narg);
	    if (f != null){
		return f.accept(visitor, context);
	    }
	}
	return null;
    }

    private synchronized void readObject(ObjectInputStream s)
	throws IOException, ClassNotFoundException {

	s.defaultReadObject();
	if (name != null){
	    name = name.intern();
	}
	if (vec == null){
	    vec = new Vector(5);
	}
    }
}
