/*
 * @(#) PnutsInterpreter.java	1.24 99/06/03
 *
 * Copyright (c) 1997-1999 Sun Microsystems, Inc. All Rights Reserved.
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package pnuts.lang;

import java.lang.reflect.Array;
import java.lang.reflect.Method;
import java.util.Hashtable;
import java.util.Enumeration;
import pnuts.lang.Runtime;

class PnutsInterpreter extends Runtime implements Visitor, PnutsParserTreeConstants {

    private final static Integer one = new Integer(1);
    private final static Object[] NO_PARAM = PnutsException.NO_PARAM;

    private final static Boolean TRUE = Boolean.TRUE;
    private final static Boolean FALSE = Boolean.FALSE;

    private final static int CODE = 0;
    private final static int VALUE = 1;
    private final static int BROKEN = 0;
    private final static int CONT = 1;
    private final static int PASSED = 2;

    // node.info = int[]{0 or 1 or -1}; 0:EOL, 1:EXPR, -1:EOF

    public Object start(SimpleNode node, Context context){
	if (node.jjtGetNumChildren() > 0){
	    try {
		return accept(node, 0, context);
	    } catch (Throwable t){
		checkException(context, t);
	    }
	}
	return null;
    }

    public Object startSet(SimpleNode node, Context context){
	int n = node.jjtGetNumChildren();
	Object ret = null;
	try {
	    for (int i = 0; i < n; i++) {
		ret = accept(node, i, context);
	    }
	} catch (Throwable t){
	    checkException(context, t);
	}
	return ret;
    }

    public Object expressionList(SimpleNode node, Context context){
	Object ret = null;
	int n = node.jjtGetNumChildren();
	for (int i = 0; i < n; i++) {
	    ret = accept(node, i, context);
	}
	return ret;
    }

    public Object global(SimpleNode node, Context context){
	String symbol = node.str;
	if (Package.globalPackage.defined(symbol, context)){
	    return Package.globalPackage.get(symbol, context);
	} else {
	    return context.undefined(symbol);
	}
    }

    public Object idNode(SimpleNode node, Context context){
	context.beginLine = node.beginLine;
	return context.get(node.str);
    }

    public Object arrayType(SimpleNode node, Context context){
	SimpleNode parent = node.jjtGetParent();
	int count = 1;
	while (parent.id == JJTARRAYTYPE){
	    count++;
	    parent = parent.jjtGetParent();
	}
	node.info = new int[]{count}; // dimension
	Object target = accept(node, 0, context);
	SimpleNode child = node.children[0];
	if (target instanceof Class){
	    return Runtime.arrayType((Class)target, 1);
	} else if (Runtime.isArray(target)){
	    return target;
	} else {
	    throw new PnutsException("classOrArray.expected", NO_PARAM, context);
	}
    }

    public Object listElements(SimpleNode node, Context context){
	int n = node.jjtGetNumChildren();
	Object ret[] = new Object[n];

	for (int i = 0; i < n; i++){
	    ret[i] = accept(node, i, context);
	}
	return ret;
    }

    public Object castExpression(SimpleNode node, Context context){
	Class type = (Class)accept(node, 0, context);
	node.info = type;
	Object obj = accept(node, 1, context);
	return Runtime.cast(context, type, obj, true);
    }

    public Object classNode(SimpleNode node, Context context){
	context.beginLine = node.beginLine;

	String name = node.str;
	int n = node.jjtGetNumChildren();

	for (int i = 0; i < n; i++){
	    SimpleNode ch = node.jjtGetChild(i);
	    if (ch.id == JJTPACKAGE){
		name += ".";
		name += ch.str;
	    }
	}
    
	try {
	    return Pnuts.loadClass(name);
	} catch (ClassNotFoundException e){
	    throw new PnutsException(e, context);
	}
    }

    public Object rangeNode(SimpleNode node, Context context){
	context.beginLine = node.beginLine;
	Object target = accept(node, 0, context);
	Object idx1 = accept(node, 1, context);
	Object idx2 = null;
	if (node.jjtGetNumChildren() >= 3){
	    idx2 = accept(node, 2, context);
	}
	if (idx1 instanceof Number && (idx2 == null || idx2 instanceof Number)){
	    int from = ((Number)idx1).intValue();
	    int to = 0;
	    if (idx2 != null){
		to = ((Number)idx2).intValue();
		if (to < 0){
		    throw new ArrayIndexOutOfBoundsException("" + to);
		}
	    }
	    if (target instanceof String){
		String s = (String)target;
		if (idx2 != null){
		    return s.substring(from, to + 1);
		} else {
		    return s.substring(from);
		}
	    } else {
		int len = Array.getLength(target);
		if (idx2 == null){
		    to = len - 1;
		} else if (to > len - 1){
		    throw new ArrayIndexOutOfBoundsException("" + to);
		}
		int size = to - from + 1;
		if (from + size > len){
		    size = len - from;
		}
		Class c = target.getClass().getComponentType();
		Object ret = Array.newInstance(c, size);
		if (size > 0){
		    System.arraycopy(target, from, ret, 0, size);
		}
		return ret;
	    }
	} else {
	    throw new PnutsException("number.expected", NO_PARAM, context);
	}
    }

    public Object indexNode(SimpleNode node, Context context){
	context.beginLine = node.beginLine;
	Object top = accept(node, 0, context);
	Object idx = accept(node, 1, context);

	// node.info =  [Class componentType, size_of_the_dimension, total_dimension]
	node.info = null;

	if (top == null){
	    throw new PnutsException(new NullPointerException(), context);
	}
	if (idx instanceof Number){
	    int ix = ((Number)idx).intValue();
	    SimpleNode c = node.children[0];
	    int dim = 0;
	    if (node.parent.id == JJTARRAYTYPE){
		dim = ((int[])node.parent.info)[0];
	    }
	    if (c.id == JJTINDEXNODE && c.info != null){
		Class type = (Class)((Object[])c.info)[0];
		int count = ((Integer)((Object[])c.info)[2]).intValue() + 1;
		node.info = new Object[]{type, idx, new Integer(count)};
		int sizes[] = new int[count];
		SimpleNode n = node;
		int i = sizes.length;
		while (n.info != null && i > 0){
		    sizes[--i] = ((Integer)((Object[])n.info)[1]).intValue();
		    n = n.children[0];
		}
		return Array.newInstance(Runtime.arrayType(type, dim), sizes);
	    } else if (top instanceof Indexed){
		return ((Indexed)top).get(ix);
	    } else if (top instanceof String){
		return new Character(((String)top).charAt(ix));
	    } else if (Runtime.isArray(top)){
		if (ix >= 0 && ix < Array.getLength(top)){
		    return Array.get(top, ix);
		} else {
		    throw new PnutsException(new IndexOutOfBoundsException("" + ix), context);
		}
	    } else if (c.id == JJTCLASS || c.id == JJTIDNODE && top instanceof Class){
		node.info = new Object[]{top, idx, one};
		return Array.newInstance(Runtime.arrayType((Class)top, dim), ix);
	    } else {
		throw new PnutsException("illegal.subscript", NO_PARAM, context);
	    }
	} else {
	    throw new PnutsException("number.expected", NO_PARAM, context);
	}
    }

    public Object methodNode(SimpleNode node, Context context){
	context.beginLine = node.beginLine;
	Object target = accept(node, 0, context);
	Object[] arg = (Object[])accept(node, 1, context);

	SimpleNode argNode = node.children[1];
	Class types[] = null;
	for (int i = 0; i < arg.length; i++){
	    SimpleNode n = argNode.children[i];
	    if (n.id == JJTCASTEXPRESSION){
		if (types == null){
		    types = new Class[arg.length];
		}
		types[i] = (Class)n.info;
	    }
	}

	if (target == null){
	    PnutsException pe = new PnutsException(new NullPointerException(), context);
	    pe.line = node.beginLine;
	    throw pe;
	}
	
	/** deleted as of 1.0beta3 **
	if (node.str.equals("sublist") && target.getClass().isArray()){
	    return sublist(target, arg, context);
	} else {
	    return Runtime.callMethod(context, target.getClass(), node.str, arg, types, target);
	}
	**/
	return Runtime.callMethod(context, target.getClass(), node.str, arg, types, target);
    }

    /* deleted as of 1.0beta3 **
    final static Object sublist(Object target, Object arg[], Context context){
	if (arg[0] instanceof Number){
	    int off = ((Number)arg[0]).intValue();
	    int len = Array.getLength(target);
	    if (off < 0 || off > len){
		throw new PnutsException("array offset is out of range", context);
	    }
	    int size;
	    if (arg.length > 1){
		size = ((Number)arg[1]).intValue();
	    } else {
		size = len - off;
	    }
	    if (size < 0){
		throw new PnutsException(new NegativeArraySizeException(""+size), context);
	    }
	    
	    if (size + off > len){
		size = len - off;
	    }
	    Class c = target.getClass().getComponentType();
	    Object ret = Array.newInstance(c, size);
	    System.arraycopy(target, off, ret, 0, size);
	    return ret;
	} else {
	    throw new PnutsException("number.expected", NO_PARAM, context);
	}
    }
    **/

    public Object staticMethodNode(SimpleNode node, Context context){
	context.beginLine = node.beginLine;
	SimpleNode c1 = node.jjtGetChild(0);
	String pkgName = getPackageName(c1);
			     
	SimpleNode argNode = node.children[1];
	Class types[] = null;
	Object[] arg = null;

	Package pkg = Package.find(pkgName);
	if (pkg != null){
	    Object fun = pkg.get(node.str, context);

	    if (fun instanceof PnutsFunction){
		arg = (Object[])accept(node, 1, context);
		PnutsFunction ft = (PnutsFunction)fun;
		return ft.exec(arg, context);
	    } else if (fun instanceof Class){
		arg = (Object[])accept(node, 1, context);
		for (int i = 0; i < arg.length; i++){
		    SimpleNode n = argNode.children[i];
		    if (n.id == JJTCASTEXPRESSION){
			if (types == null){
			    types = new Class[arg.length];
			}
			types[i] = (Class)n.info;
		    }
		}
		return Runtime.callConstructor(context, (Class)fun, arg, types);
	    } else {
		throw new PnutsException("illegal.staticCall", 
					new Object[]{
						pkgName,
						node.str,
						new Integer(argNode.jjtGetNumChildren())},
					context);
	    }
	}
	Object target = accept(node, 0, context);
	arg = (Object[])accept(node, 1, context);
	if (target instanceof Class){
	    for (int i = 0; i < arg.length; i++){
		SimpleNode n = argNode.children[i];
		if (n.id == JJTCASTEXPRESSION){
		    if (types == null){
			types = new Class[arg.length];
		    }
		    types[i] = (Class)n.info;
		}
	    }
	    return Runtime.callMethod(context, (Class)target, node.str, arg, types, null);
	} else {
	    throw new PnutsException("illegal.staticCall", 
					 new Object[]{pkgName, node.str, new Integer(arg.length)},
					 context);
	}
    }

    public Object memberNode(SimpleNode node, Context context){
	context.beginLine = node.beginLine;
	Object target = accept(node, 0, context);

	if (target == null){
	    throw new PnutsException(new NullPointerException(), context);
	} else if (Runtime.isArray(target) && node.str.equals("length")){
	    return new Integer(Array.getLength(target));
	}
	return Runtime.getField(context, target, node.str);
    }

    static String getPackageName(SimpleNode node){
	if (node.jjtGetNumChildren() > 0){
	    SimpleNode c1 = node.jjtGetChild(0);
	    return getPackageName(c1) + "::" + node.str;
	} else {
	    return node.str;
	}
    }

    public Object staticMemberNode(SimpleNode node, Context context){
	context.beginLine = node.beginLine;
	SimpleNode c1 = node.jjtGetChild(0);
	String pkgName = getPackageName(c1);

	Package pkg = Package.find(pkgName);
	if (pkg != null){
	    if (pkg.defined(node.str, context)){
		return pkg.get(node.str, context);
	    } else {
		return context.undefined(node.str);
	    }
	}
	Object target = c1.accept(this, context);
	if (target instanceof Class){
	    return Runtime.getStaticField(context, (Class)target, node.str);
	} else {
	    throw new PnutsException("packageOrClass.expected", NO_PARAM, context);
	}
    }

    public Object applicationNode(SimpleNode node, Context context){
	context.beginLine = node.beginLine;

	Object target = accept(node, 0, context);
	
	if (target instanceof PnutsFunction){
	    PnutsFunction ftable = (PnutsFunction)target;
	    Object[] arg = (Object[])accept(node, 1, context);
	    return ftable.exec(arg, context);

	} else if (target instanceof Class){  // constructor
	    Object[] arg = (Object[])accept(node, 1, context);
	    if (((Class)target).isPrimitive()){
		return Runtime.primitive(context, (Class)target, arg[0], true);
	    } else if (((Class)target).isArray()){
		return Runtime.cast(context, (Class)target, arg[0], true);
	    } else {
		SimpleNode argNode = node.children[1];
		Class types[] = null;
		for (int i = 0; i < arg.length; i++){
		    SimpleNode n = argNode.children[i];
		    if (n.id == JJTCASTEXPRESSION){
			if (types == null){
			    types = new Class[arg.length];
			}
			types[i] = (Class)n.info;
		    }
		}
		return Runtime.callConstructor(context, (Class)target, arg, types);
	    }
	}
	throw new PnutsException("funcOrType.expected", new Object[]{Pnuts.format(target)}, context);
    }

    // node.info = [int[] unitOffset]

    public Object integerNode(SimpleNode node, Context context){
	String str = node.str;
	if (node.value != null){
	    if (node.info == null){
		return node.value;
	    } else {
		int offset = ((int[])node.info)[0];
		return Runtime.quantity((Number)node.value,
				      str.substring(0, offset),
				      str.substring(offset),
				      context);
	    }
	}
	Object[] p = Runtime.parseInt(str); // [Number value, int[] unitOffset]
	Number n = (Number)p[0];
	node.value = n;
	if (p[1] != null){
	    node.info = p[1];
	    int offset = ((int[])p[1])[0];
	    return Runtime.quantity(n, str.substring(0, offset), str.substring(offset), context);
	} else {
	    return n;
	}
    }

    public Object floatingNode(SimpleNode node, Context context){
	context.beginLine = node.beginLine;
	
	String str = node.str;
	if (node.value != null){
	    if (node.info == null){ // no unit
		return node.value;
	    } else {
		int offset = ((int[])node.info)[0];
		return Runtime.quantity((Number)node.value,
				      str.substring(0, offset),
				      str.substring(offset), context);
	    }
	}
	Object[] p = Runtime.parseFloat(str); // [Number value, int[] unitOffset]
	Number n = (Number)p[0];
	node.value = n;
	if (p[1] != null){
	    node.info = p[1];
	    int offset = ((int[])p[1])[0];
	    return Runtime.quantity(n, str.substring(0, offset), str.substring(offset), context);
	} else {
	    return n;
	}
    }

    public Object characterNode(SimpleNode node, Context context){
	if (node.value != null){
	    return node.value;
	}
	try {
	    return node.value = Runtime.parseChar(node.str);
	} catch (PnutsException pe){
	    throw new PnutsException(pe.getThrowable(), context);
	}
    }

    public Object stringNode(SimpleNode node, Context context){
	if (node.value != null){
	    return node.value;
	}
	try {
	    return node.value = Runtime.parseString(node.str);
	} catch (PnutsException pe){
	    throw new PnutsException(pe.getThrowable(), context);
	}
    }

    public Object trueNode(SimpleNode node, Context context){
	return TRUE;
    }

    public Object falseNode(SimpleNode node, Context context){
	return FALSE;
    }

    public Object nullNode(SimpleNode node, Context context){
	return null;
    }

    static Object compute(Object o1, Object o2, Context context, int id){
	try {
	    switch (id){
	    case JJTASSIGNMENTTA:
		return Arithmetic.multiply(o1, o2);
	    case JJTASSIGNMENTMA:
		return Arithmetic.mod((Number)o1, (Number)o2);
	    case JJTASSIGNMENTDA:
		return Arithmetic.divide(o1, o2);
	    case JJTASSIGNMENTPA:
		return Arithmetic.add(o1, o2);
	    case JJTASSIGNMENTSA:
		return Arithmetic.subtract(o1, o2);
	    case JJTASSIGNMENTLA:
		return Arithmetic.shiftLeft((Number)o1, (Number)o2);
	    case JJTASSIGNMENTRA:
		return Arithmetic.shiftRight((Number)o1, (Number)o2);
	    case JJTASSIGNMENTRAA:
		return Arithmetic.shiftArithmetic((Number)o1, (Number)o2);
	    case JJTASSIGNMENTAA:
		return Arithmetic.and((Number)o1, (Number)o2);
	    case JJTASSIGNMENTEA:
		return Arithmetic.xor((Number)o1, (Number)o2);
	    case JJTASSIGNMENTOA:
		return Arithmetic.or((Number)o1, (Number)o2);
	    default:
		return null; // never happen
	    }
	} catch (IllegalArgumentException e){
	    throw new PnutsException("illegal.operandType", NO_PARAM, context);
	}
    }

    private static Object putArray(Object array, int idx, Object expr, Context context, int id){
	if (id != JJTASSIGNMENT){
	    Object obj = Array.get(array, idx);
	    expr = compute(obj, expr, context, id);
	}
	Array.set(array, idx, expr);
	return expr;
    }

    private Object doAssign(SimpleNode node, Object expr, Context context, int id){
	if (node.id == JJTIDNODE){
	    if (id != JJTASSIGNMENT){
		expr = compute(context.get(node.str), expr, context, id);
	    }
	    context.set(node.str, expr);
	    return expr;
	} else if (node.id == JJTGLOBAL){
	    if (id != JJTASSIGNMENT){
		expr = compute(Package.globalPackage.get(node.str, context), expr, context, id);
	    }
	    Package.globalPackage.set(node.str, expr, context);
	    return expr;
	} else if (node.id == JJTINDEXNODE){
	    Object o = accept(node, 0, context);
	    Object idx = accept(node, 1, context);
	    if (!(idx instanceof Number)){
		throw new PnutsException("number.expected", NO_PARAM, context);
	    }
	    int i = ((Number)idx).intValue();
	    if (o instanceof Indexed){
		Indexed indexed = (Indexed)o;
		if (id != JJTASSIGNMENT){
		    expr = compute(indexed.get(i), expr, context, id);
		}
		indexed.set(i, expr);
		return expr;
	    } else if (o instanceof String){
		if (id != JJTASSIGNMENT || !(expr instanceof Character)){
	            throw new PnutsException("illegal.assign", NO_PARAM, context);
		}
		Object value = assignRange(o, (Number)idx, (Number)idx, expr);
		SimpleNode cld = node.jjtGetChild(0);

		if (cld.id == JJTIDNODE ||
		    cld.id == JJTGLOBAL ||
		    cld.id == JJTINDEXNODE ||
		    cld.id == JJTSTATICMEMBERNODE ||
		    cld.id == JJTMEMBERNODE){
		    return doAssign(cld, value, context, id);
		} else {
		    return value;
		}
	    } else if (Runtime.isArray(o)){
		if (id != JJTASSIGNMENT){
		    expr = compute(Array.get(o, i), expr, context, id);
		}
		Array.set(o, i, expr);
		return expr;
	    }
	    throw new PnutsException("illegal.assign", NO_PARAM, context);
	} else if (node.id == JJTSTATICMEMBERNODE){
	    SimpleNode c1 = node.jjtGetChild(0);
	    String pkgName = getPackageName(c1);
	    Package pkg = Package.find(pkgName);
	    if (pkg != null){
		pkg.set(node.str, expr, context);
		return expr;
	    }
	    Object target = c1.accept(this, context);
	    if (target instanceof Class){
		Class clazz = (Class)target;
		if (id != JJTASSIGNMENT){
		    Object f = Runtime.getField(context, clazz, node.str);
		    expr = compute(f, expr, context, id);
		}
		Runtime.putStaticField(context, clazz, node.str, expr);
		return expr;
	    } else {
		throw new PnutsException("package.notFound", NO_PARAM, context);
	    }
	} else if (node.id == JJTMEMBERNODE){
	    Object target = accept(node, 0, context);

	    if (id != JJTASSIGNMENT){
		Object f = Runtime.getField(context, target, node.str);
		expr = compute(f, expr, context, id);
	    }
	    Runtime.putField(context, target, node.str, expr);
	    return expr;
	} else if (node.id == JJTRANGENODE){
	    if (id != JJTASSIGNMENT){
	        throw new PnutsException("illegal.assign", NO_PARAM, context);
	    }
	    Object target = accept(node, 0, context);
	    Object idx1 = accept(node, 1, context);
	    Object idx2 = null;
	    if (node.jjtGetNumChildren() >= 3){
		idx2 = accept(node, 2, context);
	    }
	    if (!(idx1 instanceof Number) || (idx2 != null && !(idx2 instanceof Number))){
		throw new PnutsException("number.expected", NO_PARAM, context);
	    }
	    Object value = assignRange(target, (Number)idx1, (Number)idx2, expr);
	    if (target instanceof String){
		SimpleNode cld = node.jjtGetChild(0);

		if (cld.id == JJTIDNODE ||
		    cld.id == JJTGLOBAL ||
		    cld.id == JJTINDEXNODE ||
		    cld.id == JJTSTATICMEMBERNODE ||
		    cld.id == JJTMEMBERNODE){
		    return doAssign(cld, value, context, id);
		} else {
		    return value;
		}
	    }
	    return value;
	} else {
	    throw new PnutsException("illegal.assign", NO_PARAM, context);
	}
    }

    private Object assign(SimpleNode node, Context context, int id){
	context.beginLine = node.beginLine;

	Object expr = accept(node, 1, context);
	try {
	    return doAssign(node.children[0], expr, context, id);
	} catch (PnutsException p){
	    return p;
	} catch (Throwable t){
	    throw new PnutsException(t, context);
	}
    }


    public Object assignment(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENT);
    }

    public Object assignmentTA(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENTTA);
    }

    public Object assignmentMA(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENTMA);
    }

    public Object assignmentDA(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENTDA);
    }

    public Object assignmentPA(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENTPA);
    }

    public Object assignmentSA(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENTSA);
    }

    public Object assignmentLA(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENTLA);
    }

    public Object assignmentRA(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENTRA);
    }

    public Object assignmentRAA(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENTRAA);
    }

    public Object assignmentAA(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENTAA);
    }

    public Object assignmentEA(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENTEA);
    }

    public Object assignmentOA(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENTOA);
    }

    public Object orNode(SimpleNode node, Context context){
	Object o1 = accept(node, 0, context);
	if (!(o1 instanceof Boolean)){
	    throw new PnutsException("boolean.expected", NO_PARAM, context);
	}
	if (((Boolean)o1).booleanValue()){
	    return TRUE;
	} else {
	    Object o2 = accept(node, 1, context);
	    if (!(o2 instanceof Boolean)){
		throw new PnutsException("boolean.expected", NO_PARAM, context);
	    }
	    return o2;
	}
    }

    public Object andNode(SimpleNode node, Context context){
	Object o1 = accept(node, 0, context);
	if (!(o1 instanceof Boolean)){
	    throw new PnutsException("boolean.expected", NO_PARAM, context);
	}
	if (((Boolean)o1).booleanValue()){
	    Object o2 = accept(node, 1, context);
	    if (!(o2 instanceof Boolean)){
		throw new PnutsException("boolean.expected", NO_PARAM, context);
	    }
	    return o2;
	} else {
	    return FALSE;
	}
    }

    public Object logOrNode(SimpleNode node, Context context){
	try {
	    return Arithmetic.or(accept(node, 0, context), accept(node, 1, context));
	} catch (IllegalArgumentException e){
	    throw new PnutsException("illegal.operandType", NO_PARAM, context);
	}
    }

    public Object logExorNode(SimpleNode node, Context context){
	try {
	    return Arithmetic.xor(accept(node, 0, context), accept(node, 1, context));
	} catch (IllegalArgumentException e){
	    throw new PnutsException("illegal.operandType", NO_PARAM, context);
	}
    }

    public Object logAndNode(SimpleNode node, Context context){
	try {
	    return Arithmetic.and(accept(node, 0, context), accept(node, 1, context));
	} catch (IllegalArgumentException e){
	    throw new PnutsException("illegal.operandType", NO_PARAM, context);
	}
    }

    public Object equalNode(SimpleNode node, Context context){
	return eq(node, context) ? TRUE : FALSE;
    }

    public Object notEqNode(SimpleNode node, Context context){
	return eq(node, context) ? FALSE : TRUE;
    }

    boolean eq(SimpleNode node, Context context){
	try {
	    return eq(accept(node, 0, context), accept(node, 1, context));
	} catch (IllegalArgumentException e){
	    throw new PnutsException("illegal.operandType", NO_PARAM, context);
	}
    }

    static boolean eq(Object o1, Object o2){
	return Arithmetic.compareTo(o1, o2) == Numeric.EQUAL;
    }

    public Object instanceofExpression(SimpleNode node, Context context){
	Object o1 = accept(node, 0, context);
	Object o2 = accept(node, 1, context);

	if (o2 instanceof Class){
	    Class c = (Class)o2;
	    return (c.isInstance(o1) ? TRUE : FALSE);
	} else if (o2 == null){
	    return FALSE;
	} else {
	    throw new PnutsException("class.expected", NO_PARAM, context);
	}
    }

    public Object ltNode(SimpleNode node, Context context){
	try {
	    int i = Arithmetic.compareTo(accept(node, 0, context),
					 accept(node, 1, context));
	    if (i == Numeric.NOT_EQUAL){
		throw new PnutsException("illegal.operandType", NO_PARAM, context);
	    }
	    return (i == Numeric.RIGHT_IS_BIGGER) ? TRUE : FALSE;
	} catch (IllegalArgumentException e){
	    throw new PnutsException("illegal.operandType", NO_PARAM, context);
	}	
    }

    public Object gtNode(SimpleNode node, Context context){
	try {
	    int i = Arithmetic.compareTo(accept(node, 0, context),
					 accept(node, 1, context));
	    if (i == Numeric.NOT_EQUAL){
		throw new PnutsException("illegal.operandType", NO_PARAM, context);
	    }
	    return (i == Numeric.LEFT_IS_BIGGER) ? TRUE : FALSE;
	} catch (IllegalArgumentException e){
	    throw new PnutsException("illegal.operandType", NO_PARAM, context);
	}
    }

    public Object leNode(SimpleNode node, Context context){
	try {
	    int i = Arithmetic.compareTo(accept(node, 0, context),
					 accept(node, 1, context));
	    if (i == Numeric.NOT_EQUAL){
		throw new PnutsException("illegal.operandType", NO_PARAM, context);
	    }
	    if (i == Numeric.EQUAL || i == Numeric.RIGHT_IS_BIGGER){
		return TRUE;
	    } else {
		return FALSE;
	    }
	} catch (IllegalArgumentException e){
	    throw new PnutsException("illegal.operandType", NO_PARAM, context);
	}
    }

    public Object geNode(SimpleNode node, Context context){
	try {
	    int i = Arithmetic.compareTo(accept(node, 0, context),
					 accept(node, 1, context));
	    if (i == Numeric.NOT_EQUAL){
		throw new PnutsException("illegal.operandType", NO_PARAM, context);
	    }
	    if (i == Numeric.EQUAL || i == Numeric.LEFT_IS_BIGGER){
		return TRUE;
	    } else {
		return FALSE;
	    }
	} catch (IllegalArgumentException e){
	    throw new PnutsException("illegal.operandType", NO_PARAM, context);
	}
    }

    public Object shiftLeftNode(SimpleNode node, Context context){
	try {
	    return Arithmetic.shiftLeft(accept(node, 0, context), accept(node, 1, context));
	} catch (IllegalArgumentException e){
	    throw new PnutsException("illegal.operandType", NO_PARAM, context);
	}
    }

    public Object shiftRightNode(SimpleNode node, Context context){
	try {
	    return Arithmetic.shiftRight(accept(node, 0, context), accept(node, 1, context));
	} catch (IllegalArgumentException e){
	    throw new PnutsException("illegal.operandType", NO_PARAM, context);
	}
    }

    public Object shiftArithmeticNode(SimpleNode node, Context context){
	try {
	    return Arithmetic.shiftArithmetic(accept(node, 0, context), accept(node, 1, context));
	} catch (IllegalArgumentException e){
	    throw new PnutsException("illegal.operandType", NO_PARAM, context);
	}
    }

    public Object addNode(SimpleNode node, Context context){
	try {
	    return Arithmetic.add(accept(node, 0, context), accept(node, 1, context));
	} catch (IllegalArgumentException e){
	    throw new PnutsException("illegal.operandType", NO_PARAM, context);
	}
    }

    public Object subtractNode(SimpleNode node, Context context){
	try {
	    return Arithmetic.subtract(accept(node, 0, context), accept(node, 1, context));
	} catch (IllegalArgumentException e){
	    throw new PnutsException("illegal.operandType", NO_PARAM, context);
	}
    }

    public Object multNode(SimpleNode node, Context context){
	try {
	    return Arithmetic.multiply(accept(node, 0, context),
				       accept(node, 1, context));
	} catch (IllegalArgumentException e){
	    throw new PnutsException("illegal.operandType", NO_PARAM, context);
	}
    }

    public Object divideNode(SimpleNode node, Context context){
	try {
	    return Arithmetic.divide(accept(node, 0, context),
				     accept(node, 1, context));
	} catch (IllegalArgumentException e){
	    throw new PnutsException("illegal.operandType", NO_PARAM, context);
	}
    }

    public Object modNode(SimpleNode node, Context context){
	try {
	    return Arithmetic.mod(accept(node, 0, context), accept(node, 1, context));
	} catch (IllegalArgumentException e){
	    throw new PnutsException("illegal.operandType", NO_PARAM, context);
	}
    }

    public Object negativeNode(SimpleNode node, Context context){
	Object o = accept(node, 0, context);
	try {
	    if (o instanceof Numeric || o instanceof Number){
		return Arithmetic.negate(o);
	    } else {
		throw new PnutsException("illegal.operandType", NO_PARAM, context);
	    }
	} catch (IllegalArgumentException e){
	    throw new PnutsException("illegal.operandType", NO_PARAM, context);
	}
    }

    public Object preIncrNode(SimpleNode node, Context context){
	return doAssign(node.jjtGetChild(0), one, context, JJTASSIGNMENTPA);
    }

    public Object preDecrNode(SimpleNode node, Context context){
	return doAssign(node.jjtGetChild(0), one, context, JJTASSIGNMENTSA);
    }

    public Object postIncrNode(SimpleNode node, Context context){
	SimpleNode n = node.jjtGetChild(0);
	Object ret = n.accept(this, context);
	Object dummy = doAssign(n, one, context, JJTASSIGNMENTPA);
	return ret;
    }

    public Object postDecrNode(SimpleNode node, Context context){
	SimpleNode n = node.jjtGetChild(0);
	Object ret = n.accept(this, context);
	Object dummy = doAssign(n, one, context, JJTASSIGNMENTSA);
	return ret;
    }

    public Object logNotNode(SimpleNode node, Context context){
	try {
	    return Arithmetic.not(accept(node, 0, context));
	} catch (IllegalArgumentException e){
	    throw new PnutsException("illegal.operandType", NO_PARAM, context);
	}
    }

    public Object notNode(SimpleNode node, Context context){
	Object o = accept(node, 0, context);
	try {
	    if (o instanceof Boolean){
		return ((Boolean)o).booleanValue() ? FALSE : TRUE;
	    } else {
		throw new PnutsException("boolean.expected", NO_PARAM, context);
	    }
	} catch (IllegalArgumentException e){
	    throw new PnutsException("illegal.operandType", NO_PARAM, context);
	}
    }

    public Object continueNode(SimpleNode node, Context context){
	SimpleNode parent = node.jjtGetParent();
	while (parent != null){
	    if (parent.id == JJTFOREACHSTATEMENT){
 		setContinued(parent.jjtGetChild(1));
		break;
	    } else if (parent.id == JJTWHILESTATEMENT){
 		setContinued(parent.jjtGetChild(1));
		break;
	    } else if (parent.id == JJTBLOCK){
 		setContinued(parent);
	    }
	    parent = parent.jjtGetParent();
	}
	return null;
    }

    public Object breakNode(SimpleNode node, Context context){
	Object o = null;
	if (node.jjtGetNumChildren() > 0){
	    o = accept(node, 0, context);
	}

	SimpleNode parent = node.jjtGetParent();
	while (parent != null){
	    if (parent.id == JJTFORSTATEMENT ||
		parent.id == JJTWHILESTATEMENT ||
		parent.id == JJTFOREACHSTATEMENT ||
		parent.id == JJTSWITCHSTATEMENT){
		setBroken(parent, o);
		break;
	    } else if (parent.id == JJTBLOCK ||
		       parent.id == JJTSWITCHBLOCK){
		setBroken(parent, o);
	    }
	    parent = parent.jjtGetParent();
	}
	return o;
    }

    public Object returnNode(SimpleNode node, Context context){
	Object o = null;
	if (node.jjtGetNumChildren() > 0){
	    o = accept(node, 0, context);
	}
	if (context.frame == null){
	    throw new Escape(o);
	} else { 
	    throw new Jump(o);
	}
    }

    public Object blockNode(SimpleNode node, Context context){
	int n = node.jjtGetNumChildren();
	Object last = null;
	Object attr[] = (Object[])node.info;
	attr[CONT] = FALSE;
	for (int i = 0; i < n; i++){
	    try {
		last = accept(node, i, context);
	    } catch (PnutsException e){
		Throwable t = e.getThrowable();
		if (t instanceof ThreadDeath){
		    setBroken(node, last);
		}
		throw e;
	    }
	    if (attr[BROKEN] == TRUE){
		attr[BROKEN] = FALSE;
		last = attr[PASSED];
		break;
	    }
	    if (attr[CONT] == TRUE){
		break;
	    }
	}
	return last;
    }

    public Object ifStatement(SimpleNode node, Context context){
	context.beginLine = node.beginLine;

	Object con = accept(node, 0, context);
	if (condition(con, context)){
	    return accept(node, 1, context);
	}
	int n = node.jjtGetNumChildren();
	for (int i = 2; i < n; i++){
	    SimpleNode _node = node.jjtGetChild(i);
	    if (_node.id == JJTELSEIFNODE){
		if (condition(accept(_node, 0, context), context)){
		    return accept(_node, 1, context);
		}
	    } else if (_node.id == JJTELSENODE){
		return accept(_node, 0, context);
	    }
	}
	return null;
    }

    static boolean condition(Object cond, Context context){
	if (cond instanceof Boolean){
	    return ((Boolean)cond).booleanValue();
	} else {
	    throw new PnutsException("boolean.expected", NO_PARAM, context);
	}
    }

    public Object whileStatement(SimpleNode node, Context context){
	context.beginLine = node.beginLine;

	Object[] attr = (Object[])node.info;
	while (true){
	    Object cond = accept(node, 0, context);
	    if (cond instanceof Boolean){
		Object last = null;
		boolean b = ((Boolean)cond).booleanValue();
		if (b){
		    last = accept(node, 1, context);

		    if (attr[BROKEN] == TRUE){
			attr[BROKEN] = FALSE;
			return attr[PASSED];
		    }
		} else {
		    attr[BROKEN] = FALSE;
		    return last;
		}
	    } else {
		throw new PnutsException("boolean.expected", NO_PARAM, context);
	    }
	}
    }

    public Object forStatement(SimpleNode node, Context context){
	SimpleNode initNode = null;
	SimpleNode condNode = null;
	SimpleNode updateNode = null;
	SimpleNode blockNode = null;

	int j = 0;
	SimpleNode n = node.jjtGetChild(j);
	if (n.id == JJTFORINIT){
	    initNode = n;
	    j++;
	}
	n = node.jjtGetChild(j);
	if (n.id != JJTFORUPDATE && n.id != JJTBLOCK){
	    condNode = n;
	    j++;
	}
	n = node.jjtGetChild(j);
	if (n.id == JJTFORUPDATE){
	    updateNode = n;
	    j++;
	}

	blockNode = node.jjtGetChild(j);

	Object last = null;

	if (initNode != null){
	    int num = initNode.jjtGetNumChildren();
	    String[] env = new String[num];
	    for (int i = 0; i < env.length; i++){
		SimpleNode sn = initNode.jjtGetChild(i);
		env[i] = sn.str;
	    }
	    context.openLocal(env);
	    for (int i = 0; i < env.length; i++){
		SimpleNode sn = initNode.jjtGetChild(i);
		context.bind(env[i], accept(sn, 0, context));
	    }
	} else {
	    context.openLocal(new String[]{});
	}

	Object c = null;
	if (condNode != null){
	    c = condNode.accept(this, context);
	} else {
	    c = Boolean.TRUE;
	}
	Object[] attr = (Object[])node.info;
	try {
	    while (((Boolean)c).booleanValue()){
		last = blockNode.accept(this, context);
		if (attr[BROKEN] == TRUE){
		    attr[BROKEN] = FALSE;
		    last = attr[PASSED];
		    break;
		}
		if (updateNode != null){
		    expressionList(updateNode, context);
		}
		if (condNode != null){
		    c = condNode.accept(this, context);
		}
	    }
	} catch (ClassCastException e){
	    throw new PnutsException("boolean.expected", NO_PARAM, context);
	}
	context.closeLocal();
	return last;
    }

    public Object foreachStatement(SimpleNode node, Context context){
	SimpleNode n2 = node.jjtGetChild(1);

	Object o = accept(node, 0, context);
	if (o == null){
	    return null;
	}
	Object[] attr = (Object[])node.info;
	if (Runtime.isArray(o)){
	    Object last = null;
	    context.openLocal(new String[]{node.str});
	    for (int i = 0; i < Array.getLength(o); i++){
		context.bind(node.str, Array.get(o, i));
		last = n2.accept(this, context);
		if (attr[BROKEN] == TRUE){
		    attr[BROKEN] = FALSE;
		    last = attr[PASSED];
		    break;
		}
	    }
	    context.closeLocal();
	    return last;
	} else if (o instanceof Enumeration){
	    Enumeration e = (Enumeration)o;
	    Object last = null;
	    context.openLocal(new String[]{node.str});
	    while (e.hasMoreElements()){
		context.bind(node.str, e.nextElement());
		last = n2.accept(this, context);
		if (attr[BROKEN] == TRUE){
		    attr[BROKEN] = FALSE;
		    last = attr[PASSED];
		    break;
		}
	    }
	    context.closeLocal();
	    return last;
	} else {
	    if (iterator == null){
		try {
		    iterator = Pnuts.loadClass("java.util.Iterator");
		} catch (ClassNotFoundException e){/* ignore */}
	    }
	    if (iterator == null){
		try {
		    iterator = Pnuts.loadClass("com.sun.java.util.collections.Iterator");
		} catch (ClassNotFoundException e){
		    /* ignore */
		}
	    }
	    if (iterator != null && iterator.isInstance(o)){
		return doIterator(o, node, n2, iterator, context);
	    }
	    throw new PnutsException("illegal.type.foreach", NO_PARAM, context);
	}
    }

    private static Class iterator = null;
    private static Method hasNext = null;
    private static Method next = null;

    Object doIterator(Object o, SimpleNode node, SimpleNode n2, Class it, Context context){
	Class t_noarg[] = new Class[]{};
	Object noarg[] = new Object[]{};
	if (hasNext == null){
	    try {
		hasNext = it.getMethod("hasNext", t_noarg);
	    } catch (NoSuchMethodException e){/* ignore */}
	}
	if (next == null){
	    try {
		next = it.getMethod("next", t_noarg);
	    } catch (NoSuchMethodException e){/* ignore */}
	}
	Object last = null;
	context.openLocal(new String[]{node.str});
	Object[] attr = (Object[])node.info;
	while (true){
	    try {
		if (!(((Boolean)hasNext.invoke(o, noarg)).booleanValue())){
		    break;
		}
		context.bind(node.str, next.invoke(o, noarg));
		last = n2.accept(this, context);
		if (attr[BROKEN] == TRUE){
		    attr[BROKEN] = FALSE;
		    last = attr[PASSED];
		    break;
		}
	    } catch (PnutsException e0){
		throw e0;
	    } catch (Throwable e){
		throw new PnutsException(e, context);
	    }
	}
	context.closeLocal();
	return last;
    }

    public Object switchStatement(SimpleNode node, Context context){
	Object[] attr = (Object[])node.info;
	int n = node.jjtGetNumChildren();
	Object target = accept(node, 0, context);
	Object last = null;
	boolean matched = false;
	for (int i = 1; i < n; i++){
	    if (attr[BROKEN] == TRUE){
		last = attr[PASSED];
		break;
	    }
	    SimpleNode _node = node.jjtGetChild(i);
	    if (_node.jjtGetNumChildren() == 1){ // case
		Object o = accept(_node, 0, context);
		i++;
		if (matched || eq(target, o)){
		    last = accept(node, i, context);
		    matched = true;
		}
	    } else { // default
		attr[BROKEN] = FALSE;
		matched = true;
		last = accept(node, ++i, context);
	    }
	}
	attr[BROKEN] = FALSE;
	return last;
    }

    public Object switchBlock(SimpleNode node, Context context){
	Object[] attr = (Object[])node.info;
	int n = node.jjtGetNumChildren();
	Object last = null;
	for (int i = 0; i < n; i++){
	    try {
		last = accept(node, i, context);
	    } catch (PnutsException e){
		Throwable t = e.getThrowable();
		if (t instanceof ThreadDeath){
		    setBroken(node, last);
		}
		throw e;
	    }
	    if (attr[BROKEN] == TRUE){
		attr[BROKEN] = FALSE;
		last = attr[PASSED];
		break;
	    }
	}
	return last;
    }

    public Object functionStatement(SimpleNode node, Context context){
	context.beginLine = node.beginLine;
	int n = node.jjtGetNumChildren();

	SimpleNode block = null;
	int nargs = 0;

	String locals[] = null;

	for (int i = 0; i < n; i++){
	    SimpleNode _node = node.jjtGetChild(i);
	    if (_node.id == JJTPARAMLIST){
		nargs = _node.jjtGetNumChildren();
		locals = new String[nargs];

		SimpleNode n0 = null;
		if (nargs == 1 &&
		    (n0 = _node.jjtGetChild(0)).id == JJTINDEXNODE){
		    nargs = -1;
		    locals[0] = n0.jjtGetChild(0).str;
		} else {
		    for (int j = 0; j < nargs; j++){
			locals[j] = _node.jjtGetChild(j).str;
		    }
		}
	    } else if (_node.id == JJTBLOCK){
		block = _node;
	    }
	}

	Package pkg = context.getCurrentPackage();
	Function f = new Function(node.str, locals, nargs, block, pkg, context);

	Object o = null;
	if (node.str != null && context.env != null){
	    o = context.env.get(node.str);
	}
	String name = f.getName();
	if (o == null){
	    if (pkg.defined(name, context)){
		o = pkg.get(name, context);
	    }
	}
	if (!(o instanceof PnutsFunction)){
	    o = null;
	}
	PnutsFunction ht = f.register((PnutsFunction)o);

	if (node.str != null){
	    context.set(node.str, ht);
	}
	return ht;
    }

    protected static void setBroken(SimpleNode node, Object value){
	Object[] attr = (Object[])node.info;
	attr[BROKEN] = TRUE;
	attr[PASSED] = value;
    }

    protected static void setContinued(SimpleNode node){
	Object[] attr = (Object[])node.info;
	attr[CONT] = TRUE;
    }

    protected Object accept(SimpleNode node, int idx, Context context){
	return node.jjtGetChild(idx).accept(this, context);
    }
}
