/*
 * @(#) Runtime.java	1.3 99/06/04
 *
 * Copyright (c) 1997-1999 Sun Microsystems, Inc. All Rights Reserved.
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package pnuts.lang;

import java.util.Hashtable;
import java.util.Enumeration;
import java.io.ObjectOutputStream;
import java.io.ObjectInputStream;
import java.io.CharArrayWriter;
import java.io.Serializable;
import java.io.IOException;
import java.lang.reflect.Method;
import java.lang.reflect.Array;
import java.lang.reflect.Constructor;
import java.lang.reflect.Modifier;
import java.lang.reflect.InvocationTargetException;

/**
 * This class provides runtime supports for Pnuts compiler/interpreter.
 * All mehtods are protected static, so that only subclasses can access them.
 *
 * @version	1.3
 */
public class Runtime implements PnutsParserTreeConstants {

    protected Runtime(){}

    /**
     * call a method
     */
    public static Object callMethod(Context context, Class c, String name, Object args[], Class types[], Object target){
	if (target instanceof AbstractData){
	    return ((AbstractData)target).invoke(name, args);
	}
	try {
	    return callMethod(c, name, args, types, target);
	} catch (InvocationTargetException e1){
	    Throwable t = e1.getTargetException();
	    if (t instanceof PnutsException){
		throw (PnutsException)t;
	    } else {
	        throw new PnutsException(t, context);
	    }
	} catch (IllegalArgumentException e2){
	    throw new PnutsException("method.notFound",
				     new Object[]{name, ""+(target==null?c:target), ""+Pnuts.format(args)}, context);
	} catch (PnutsException e3){
	    throw e3;
	} catch (Throwable e4){
	    throw new PnutsException(e4, context);
	}
    }

    /**
     * Call a method
     *
     * @param c the class of method
     * @param name the method name
     * @param args the paramters
     * @param target the target object
     * @return the return value
     */
    public static Object callMethod(Class c, String name, Object args[], Object target)
	throws IllegalAccessException, IllegalArgumentException, InvocationTargetException {
	    return callMethod(c, name, args, null, target);
    }

    /**
     * Call a method
     *
     * @param c the class of method
     * @param name the method name
     * @param args the paramters
     * @param types the types of the formal arguments
     * @param target the target object
     * @return the return value
     */
    public static Object callMethod(Class c, String name, Object args[], Class types[], Object target)
	throws IllegalAccessException, IllegalArgumentException, InvocationTargetException {
					
        Method method = null;
	boolean _static = (target == null);
	try {
	    if (c.isArray()){
		c = Object.class;  // This line is needed for JavaStation
	    }
	    Method m[] = getMethods(c, name);

	    int count = 0;
	    int min = Integer.MAX_VALUE;
	    Stack methods = new Stack();
	cand:
	    for (int i = 0; i < m.length; i++){
		Class p[] = m[i].getParameterTypes();
		if (p.length != args.length){
		    continue;
		}
		
		count = 0;
		for (int j = 0; j < p.length; j++){
		    
		    if (types != null &&
			types[j] != null &&
			p[j] != types[j] &&
			!p[j].isAssignableFrom(types[j])){
			    
			continue cand;
		    }
		    int t = matchType(p[j], args[j]);
		    if (t < 0){
			continue cand;
		    }
		    count += t;
		}
		if (count > min){
		    continue;
		}
		
		boolean st = Modifier.isStatic(m[i].getModifiers());
		if (st && !_static || !st && _static){
		    continue;
		}
		    
		if (count < min){
		    methods.removeAllElements();
		    methods.push(m[i]);
		    min = count;
		} else if (count == min){
		    methods.push(m[i]);
		}
	    }
	    Class clazz = c;
	out:
	    while (clazz != null){
		for (int i = 0; i < methods.size(); ){
		    method = (Method)methods.pop();
		    if (method.getDeclaringClass() == clazz){
			break out;
		    }
		}
		clazz = clazz.getSuperclass();
	    }

	    if (method != null){
		Class p[] = method.getParameterTypes();
		for (int j = 0; j < p.length; j++){
		    if (p[j].isArray() && args[j] != null && isArray(args[j])){
			if (!p[j].isAssignableFrom(args[j].getClass())){
			    Class compType = getBottomType(p[j]);
			    args[j] = transform(compType, args[j]);
			}
		    }
		}
		return method.invoke(target, args);
	    } else {
		throw new IllegalArgumentException();
	    }
	} catch (IllegalAccessException pe){
	    if (!Modifier.isPublic(c.getModifiers())){
		if (accessible != null){
		    accessible.invoke(method, new Object[]{Boolean.TRUE});
		    return method.invoke(target, args);
		} else {
		    Class p[] = method.getParameterTypes();
		    Method m = findCallableMethod(c, name, p);
		    if (m != null){
			return m.invoke(target, args);
		    }
		}
	    }
	    throw pe;
	}
    }

    static Method accessible;
    static {
	try {
	    accessible = Method.class.getMethod("setAccessible", new Class[]{Boolean.TYPE});
	} catch (Exception e){/* ignore */}
    }

    public static Object callConstructor(Context context, Class c, Object args[], Class types[]){
	try {
	    return callConstructor(c, args, types);
	} catch (InvocationTargetException e1){
	    Throwable t = ((InvocationTargetException)e1).getTargetException();
	    if (t instanceof PnutsException){
		throw (PnutsException)t;
	    } else {
		throw new PnutsException(t, context);
	    }
	} catch (IllegalArgumentException e2){
	    throw new PnutsException("constructor.notFound", new Object[]{c, Pnuts.format(args)}, context);
	} catch (PnutsException e3){
	    throw e3;
	} catch (Throwable e4){
	    throw new PnutsException(e4, context);
	}
    }

    /**
     * Call a constructor
     *
     * @param c the class of constructor
     * @param args the paramters
     * @return the created instance
     */
    public static Object callConstructor(Class c, Object args[]) throws IllegalAccessException, IllegalArgumentException, InvocationTargetException, InstantiationException {
	return callConstructor(c, args, null);
    }

    /**
     * Call a constructor
     *
     * @param c the class of constructor
     * @param args the paramters
     * @param types the types of the formal arguments
     * @return the created instance
     */
    public static Object callConstructor(Class c, Object args[], Class types[])
	throws IllegalAccessException, IllegalArgumentException, InvocationTargetException, InstantiationException {

//	Constructor cs[] = c.getConstructors();
	Constructor cs[] = getConstructors(c);

	Constructor cons = null;
	int count = 0;
	int min = Integer.MAX_VALUE;
    cand:
	for (int i = 0; i < cs.length; i++){
	    Class p[] = cs[i].getParameterTypes();
	    if (p.length != args.length){
		continue;
	    }
	    count = 0;
	    for (int j = 0; j < p.length; j++){

		if (types != null &&
		    types[j] != null &&
		    p[j] != types[j] &&
		    !p[j].isAssignableFrom(types[j])){

		    continue cand;
		}
		int t = matchType(p[j], args[j]);
		if (t < 0){
		    continue cand;
		}
		count += t;
	    }
	    if (count < min){
		min = count;
		cons = cs[i];
	    }
	}
	if (cons != null){
	    Class p[] = cons.getParameterTypes();
	    for (int j = 0; j < p.length; j++){
		if (p[j].isArray() && args[j] != null && isArray(args[j])){
		    if (!p[j].isAssignableFrom(args[j].getClass())){
			Class compType = getBottomType(p[j]);
			args[j] = transform(compType, args[j]);
		    }
		}
	    }
	    return cons.newInstance(args);
	} else {
	    throw new IllegalArgumentException();
	}
    }

    public static void putStaticField(Context context, Class clazz, String name, Object expr) {
	try {
	    clazz.getField(name).set(null, expr);
	} catch (PnutsException e0){
	    throw e0;
	} catch (Throwable e){
	    throw new PnutsException(e, context);
	}
    }

    public static Object getStaticField(Context context, Class clazz, String name){
	try {
	    return getStaticField(clazz, name);
	} catch (PnutsException p){
	    throw p;
	} catch (Throwable t){
	    throw new PnutsException(t, context);
	}
    }

    static Object getStaticField(Class clazz, String name) throws IllegalArgumentException, IllegalAccessException, NoSuchFieldException {
	return clazz.getField(name).get(null);
    }

    public static void putField(Context context, Object target, String name, Object expr) {
	try {
	    if (target instanceof Property){
		((Property)target).set(name, expr);
	    } else {
		target.getClass().getField(name).set(target, expr);
	    }
	} catch (PnutsException p){
	    throw p;
	} catch (Throwable t){
	    throw new PnutsException(t, context);
	}
    }

    public static Object getField(Context context, Object target, String name){
	try {
	    if (target instanceof Property){
		return ((Property)target).get(name);
	    } else {
		return target.getClass().getField(name).get(target);
	    }
//	    return getField(target, name);
	} catch (PnutsException p){
	    throw p;
	} catch (Throwable t){
	    throw new PnutsException(t, context);
	}
    }

    /**
     * Get true component type from an array type.
     *  e.g. int[][] ==> int,  String[] ==> String
     *
     * @param clazz	An array type to be examined
     * @return		The component type of the array type.
     */
    static Class getBottomType(Class clazz){
	while (clazz.isArray()){
	    clazz = clazz.getComponentType();
	}
	return clazz;
    }

    public static Class arrayType(Class c, int dim){
	if (dim == 0){
	    return c;
	} else {
	    return arrayType(Array.newInstance(c, 0).getClass(), dim - 1);
	}
    }

    /**
     * Transform an array to a different array type.
     *
     * @param baseType	The component type of which the array to create
     * @param array	The source array of the transformation
     * @return		An array of the <code>clazz</code> type that has same values as <code>obj</code>
     */
    static Object transform(Class baseType, Object array){
	Stack dim = new Stack();
	Object b = array;
	while (isArray(b)){
	    int len = Array.getLength(b);
	    dim.push(new Integer(len));
	    if (len == 0){
		break;
	    }
	    b = Array.get(b, 0);
	}
	int dimensions[] = new int[dim.size()];
	for (int i = dimensions.length - 1; i >= 0; i--){
	    dimensions[i] = ((Integer)dim.pop()).intValue();
	}
	Object ret = Array.newInstance(baseType, dimensions);
	copy(array, ret);
	return ret;
    }

    /*
     * @param src should be an array
     * @param dst should be an array
     */
    private static void copy(Object src, Object dst){
	for (int i = 0; i < Array.getLength(src); i++){
	    Object elem = Array.get(src, i);
	    if (isArray(elem)){
		copy(elem, Array.get(dst, i));
	    } else {
		Array.set(dst, i, Array.get(src, i));
	    }
	}
    }

    /**
     * Write a hashtable to the objectOutputStream.
     * If a value of the hashtable is a
     * Class object, it is encapsulated to avoid NotSerializableException.
     * But if a value contains a Class object in its structure
     * e.g. "new Class[]{ Object.class }",  NotSerializableException
     * is thrown, even though the Class object is Serializable.
     * If notToSerialize is not null, falied keys are accumulated in
     * notToSerialize.
     *
     * This is a workaround for bug:4075221.
     */
    public static int writeSymbolTable(SymbolTable tab, SymbolTable notToSerialize, ObjectOutputStream s) throws IOException {
	int count = 0;
	for (Enumeration e = tab.keys(); e.hasMoreElements(); ){
	    String key = (String)e.nextElement();
	    if (notToSerialize != null && notToSerialize.get(key) != null){
		continue;
	    }
	    Object value = tab.get(key);
	    if (Pnuts.debug()){
		System.out.println(key + "=" + value);
	    }
	    try {
		if (!(value instanceof Serializable)){
		    continue;
		}
		if (value instanceof Class){
		    value = new ClassHolder((Class)value);
		}
		s.writeObject(key);
		s.writeObject(value);
		count++;
	    } catch (IOException ex){
		if (notToSerialize != null){
		    notToSerialize.put(key, key);
		}
		throw ex;
	    }
	}
	s.writeObject(null);
	
	return count;
    }

    /**
     * Read a hashtable from objectInputStream.
     * If Class objects are encapsulated,
     * it retrieves the Class objects from them.
     */
    public static void readSymbolTable(SymbolTable table, ObjectInputStream s)
	throws IOException, ClassNotFoundException {
	    for (;;){
		String k = (String)s.readObject();
		if (k == null){
		    break;
		}
		Object v = s.readObject();
		if (v instanceof ClassHolder){
		    v = ((ClassHolder)v).get();
		}
		if (Pnuts.debug()){
		    System.out.println(k + "=" + v);
		}
		table.put(k.intern(), v);
	    }
    }

    private static int matchType(Class f, Object a){
	if (a == null){
	    if (f.isPrimitive()){
		return -1;
	    }
	    return 0;
	}
	Class clazz = a.getClass();
	if (clazz == f){
	    return 0;
	}
	if (f == Boolean.TYPE){
	    return (clazz == Boolean.class) ? 0 : -1;
	}
	if (f == Byte.TYPE){
	    return distance(0, clazz);
	}
	if (f == Character.TYPE){
	    return distance(1, clazz);
	}
	if (f == Short.TYPE){
	    return distance(1, clazz);
	}
	if (f == Integer.TYPE){
	    return distance(2, clazz);
	}
	if (f == Long.TYPE){
	    return distance(3, clazz);
	}
	if (f == Float.TYPE){
	    return distance(4, clazz);
	}
	if (f == Double.TYPE){
	    return distance(5, clazz);
	}
	if (f.isAssignableFrom(clazz)){
	    return 1;
	} else if (f.isArray() && clazz.isArray()){
	    int j = 1;
	    for (int i = 0; i < Array.getLength(a); i++){
		int s = matchType(f.getComponentType(), Array.get(a, i));
		if (s < 0){
		    return -1;
		} else {
		    j += s;
		}
	    }
	    return j;
	} else {
	    return -1;
	}
    }

    private static int distance(int pos, Class clazz){
	if (clazz == Double.class){
	    return pos - 5;
	} else if (clazz == Float.class){
	    return pos - 4;
	} else if (clazz == Long.class){
	    return pos - 3;
	} else if (clazz == Integer.class){
	    return pos - 2;
	} else if (clazz == Short.class){
	    return pos - 1;
	} else if (clazz == Character.class){
	    return pos - 1;
	} else if (clazz == Byte.class){
	    return pos;
	} else {
	    return -1;
	}
    }

    private static Method findCallableMethod(Class clazz, String name, Class args[]){
      if (clazz == null){
	return null;
      }
      Method method;
      if (Modifier.isPublic(clazz.getModifiers())){
	try {
	  method = clazz.getMethod(name, args);
	  if (method != null){
	    return method;
	  }
	} catch (NoSuchMethodException nme){
	}
      }
      if (!clazz.isInterface()){
	Class it[] = clazz.getInterfaces();
	for (int i = 0; i < it.length; i++){
	  method = findCallableMethod(it[i], name, args);
	  if (method != null){
	    return method;
	  }
	}
	method = findCallableMethod(clazz.getSuperclass(), name, args);
	if (method != null){
	  return method;
	}
      }
      return null;
    }

    private static Hashtable mtab = new Hashtable(32);

    private static Method[] getMethods(Class c, String name){
	String key = c.getName() + "/" + name;
	Object v = mtab.get(key);
	if (v instanceof Method[]){
	    return (Method[])v;
	} else {
	    Method m[] = getMethods(c);
	    if (m == null){ // for Bug:4137722
		throw new NoClassDefFoundError("" + c);
	    }
	    int j = 0;
	    for (int i = 0; i < m.length; i++){
		if (m[i].getName().equals(name) && i >= j){
		    m[j] = m[i];
		    j++;
		}
	    }
	    Method m2[] = new Method[j];
	    System.arraycopy(m, 0, m2, 0, j);
	    if (mtab.size() > 0xFFF){
		mtab.clear();
	    }
	    mtab.put(key, m2);
	    return m2;
	}
    }

    private static Method[] getMethods(Class c){
	return SecuritySensitive.getMethods(c);
    }

    private static Hashtable ctab = new Hashtable(4);

    private static Constructor[] getConstructors(Class c){
	String key = c.getName();
	Object v = ctab.get(key);
	if (v instanceof Constructor[]){
	    return (Constructor[])v;
	} else {
	    Constructor con[] = c.getConstructors();
	    if (ctab.size() > 0xFFF){
		ctab.clear();
	    }
	    ctab.put(key, con);
	    return con;
	}
    }

    private static Hashtable ptab = new Hashtable(32);

    private static Class[] getParamTypes(Method m){
	Object o = ptab.get(m);
	if (o instanceof Class[]){
	    return (Class[])o;
	} else {
	    Class[] p = m.getParameterTypes();
	    if (ptab.size() > 0xFFF){
		ptab.clear();
	    }
	    ptab.put(m, p);
	    return p;
	}
    }

    /**
     * [Number number, int offset_of_unit_symbol]
     */
    public static Object[] parseInt(String str){
	char c1 = str.charAt(0);
	if (c1 == '#'){
	    return parseInt(str.substring(1), 16, true);
	} else if (c1 == '0'){
	    if (str.length() > 1){
		char c2 = str.charAt(1);
		if (c2 == 'x' || c2 == 'X'){
		    return parseInt(str.substring(2), 16, false);
		} else {
		    return parseInt(str, 10, false);
		}
	    } else {
		return new Object[]{new Integer(0), null};
	    }
	} else {
	    return parseInt(str, 10, false);
	}
    }

    static Object[] parseInt(String str, int radix, boolean shrink){
	boolean overflow = false;
	long value = 0;
	char chars[] = str.toCharArray();
	int i = 0;
    out:
	for (i = 0; i < chars.length; i++) {
	    int ch = chars[i];
	    switch (ch) {
	    case '0': case '1': case '2': case '3': case '4':
	    case '5': case '6': case '7': case '8': case '9':
		if (radix == 10){
		    int c = ch - '0';
		    value = (value * 10) + c;
		    if (value < 0){ 
			overflow = true;
			break out;
		    }
		    break;
		} else if (radix == 16){
		    value = (value << 4) + 
			Character.toLowerCase((char)ch) - '0';
		    if (value < 0){ 
			overflow = true;
			break out;
		    }
		    break;
		}
	    case 'a': case 'A': case 'b': case 'B': case 'c': case 'C':
	    case 'd': case 'D': case 'e': case 'E': case 'f': case 'F':
		if (radix == 10){
		    break out;
		}
		value = (value << 4) + 10 + 
		    Character.toLowerCase((char)ch) - 'a';
		if (value < 0){ 
		    overflow = true;
		    break out;
		}
		break;
	    default:
		break out;
	    }
	}

    out2:
	for (; i < chars.length; i++) {
	    int ch = chars[i];
	    switch (ch) {
	    case '0': case '1': case '2': case '3': case '4':
	    case '5': case '6': case '7': case '8': case '9':
		break;
	    case 'a': case 'A': case 'b': case 'B': case 'c': case 'C':
	    case 'd': case 'D': case 'e': case 'E': case 'f': case 'F':
		if (radix == 10){
		    break out2;
		}
		break;
	    default:
		break out2;
	    }
	}

	Number number = null;

	if (overflow){
	    String s = str;
	    if (str.length() > i){
		s = str.substring(0, i);
	    }
	    number = Arithmetic.decimalNumber(s, radix);
	} else {
	    if (value >= 0 && value <= 255 && radix == 16){
		if (shrink){
		    number = new Byte((byte)value);
		} else {
		    number = new Integer((int)value);
		}
	    } else if (value <= Integer.MIN_VALUE ||
		       value >= Integer.MAX_VALUE){
		number = new Long(value);
	    } else {
		number = new Integer((int)value);
	    }
	}
	if (str.length() > i){
	    return new Object[]{number, new int[]{i}};
	} else {
	    return new Object[]{number, null};
	}
    }

    /**
     * [Number number, int offset_of_unit_symbol]
     */
    public static Object[] parseFloat(String str){
	char chars[] = str.toCharArray();
	int i = 0;
    out1:
	for (i = 0; i < chars.length; i++){
	    int ch = chars[i];
	    switch (ch){
	    case '0': case '1': case '2': case '3': case '4':
	    case '5': case '6': case '7': case '8': case '9':
	    case '.':
		break;
	    default:
		break out1;
	    }
	}
	if (i < chars.length && (chars[i] == 'e' || chars[i] == 'E')){
	    i++;
	out2:
	    for (; i < chars.length; i++){
		int ch = chars[i];
		switch (ch){
		case '+': case '-':
		case '0': case '1': case '2': case '3': case '4':
		case '5': case '6': case '7': case '8': case '9':
		    break;
		default:
		    break out2;
		}
	    }
	}
	String s = str;
	if (i < chars.length){
	    s = str.substring(0, i);
	}
	Number n = null;
	if (s.length() > 16){
	    n = Arithmetic.decimalNumber(s, 10);
	} else {
	    n = Double.valueOf(s);
	}
	if (i < chars.length){
	    return new Object[]{n, new int[]{i}};
	} else {
	    return new Object[]{n, null};
	}
    }

    static CharArrayWriter cout = new CharArrayWriter();

    public static String parseString(String s){
	try {
	    char[] cc = s.toCharArray();

	    cout.reset();

	    for (int i = 1; i < cc.length - 1; i++){
		if (cc[i] == '\\'){
		    i++;
		    switch (cc[i]){
		    case '"':
			cout.write((int)'"');
			break;
		    case 'b':
			cout.write((int)'\b');
			break;
		    case 'f':
			cout.write((int)'\f');
			break;
		    case 't':
			cout.write((int)'\t');
			break;
		    case 'r':
			cout.write((int)'\r');
			break;
		    case 'n':
			cout.write((int)'\n');
			break;
		    case '0':
			cout.write((int)'\0');
			break;
		    case '\\':
			cout.write((int)'\\');
			break;
		    }
		} else {
		    cout.write((int)cc[i]);
		}
	    }
	    return new String(cout.toCharArray());
	} catch(Throwable e){
	    throw new PnutsException(e);
	}
    }

    public static Character parseChar(String s){
	char c[] = s.toCharArray();
	try {
	    if (c[1] == '\\'){
		switch (c[2]){
		case '\'':
		    return new Character('\'');
		case 'b':
		    return new Character('\b');
		case 'f':
		    return new Character('\f');
		case 't':
		    return new Character('\t');
		case 'r':
		    return new Character('\r');
		case 'n':
		    return new Character('\n');
		case '0':
		    return new Character('\0');
		case '\\':
		    return new Character('\\');
		default:
		    return null;
		}
	    } else {
		return new Character(c[1]);
	    }
	} catch(Throwable e){
	    throw new PnutsException(e);
	}
    }

    public static Object quantity(Number number, String numberString, String unit, Context context){
	SymbolTable units = Package.globalPackage.unitTable;

	QuantityFactory factory = (QuantityFactory)units.get(unit);
	if (factory != null){
	    return factory.make(number, unit);
	}
	if ("f".equals(unit) || "F".equals(unit)){
	    return new Float(number.floatValue());
	} else if ("d".equals(unit) || "D".equals(unit)){
	    return new Double(number.doubleValue());
	} else if ("b".equals(unit) || "B".equals(unit)){
	    return Arithmetic.decimalNumber(numberString, 10);
	}
	throw new PnutsException("unitName.notDefined", new Object[]{unit}, context);
    }

    /**
     * This method is called by the syntax
     * "primitiveType(object)" and "(primitiveType)object"
     *
     * @param context	the context
     * @param primitiveType	a primitive type
     * @param param	the parameter
     * @param flag	string<->number conversion
     */
    public static Object primitive(Context context, Class primitiveType, Object param, boolean flag){
	if (primitiveType == Integer.TYPE){
	    if (param instanceof Byte){
		byte b = ((Byte)param).byteValue();
//    return new Integer((b + 256) & 255);
		return new Integer((int)b);
	    } else if (param instanceof Character){
		return new Integer((int)((Character)param).charValue());
	    } else if (param instanceof Number){
		return new Integer(((Number)param).intValue());
	    } else if (flag && param instanceof String){
		return new Integer(Integer.parseInt(((String)param).trim()));
	    }
	} else if (primitiveType == Byte.TYPE){
	    if (param instanceof Character){
		return new Byte((byte)((Character)param).charValue());
	    } else if (param instanceof Number){
		return new Byte(((Number)param).byteValue());
	    } else if (flag && param instanceof String){
		return new Byte((byte)Integer.parseInt(((String)param).trim()));
	    }
	} else if (primitiveType == Character.TYPE){
	    if (param instanceof Number){
		int i = ((Number)param).intValue();
		if (i >= 0 && i <= 0xffff){
		    return new Character((char)i);
		} else {
		    throw new PnutsException("char.outOfBound", PnutsException.NO_PARAM, context);
		}
	    }
	} else if (primitiveType == Long.TYPE){
	    if (param instanceof Number){
		return new Long(((Number)param).longValue());
	    } else if (flag && param instanceof String){
		return new Long(Long.parseLong(((String)param).trim()));
	    }
	} else if (primitiveType == Short.TYPE){
	    if (param instanceof Character){
		return new Short((short)((Character)param).charValue());
	    } else if (param instanceof Number){
		return new Short(((Number)param).shortValue());
	    } else if (flag && param instanceof String){
		return new Short(Short.parseShort(((String)param).trim()));
	    }
	} else if (primitiveType == Double.TYPE){
	    if (param instanceof Number){
		return new Double(((Number)param).doubleValue());
	    } else if (flag && param instanceof String){
		return Double.valueOf((String)param);
	    }
	} else if (primitiveType == Float.TYPE){
	    if (param instanceof Number){
		return new Float(((Number)param).floatValue());
	    } else if (flag && param instanceof String){
		return Float.valueOf((String)param);
	    }
	} else if (primitiveType == Boolean.TYPE){
	    if (param instanceof Boolean){
		return param;
	    } else if (flag && param instanceof String){
		return Boolean.valueOf((String)param);
	    }
	}
	throw new ClassCastException("(" + primitiveType.getName() + ")" + Pnuts.format(param));
    }

    /**
     * This method is called by the syntax "(Class)object"
     *
     * @param context	the context
     * @param type	the type
     * @param flag	object_array<->primitive_array conversion
     */
    public static Object cast(Context context, Class type, Object object, boolean flag){
	if (type.isPrimitive()){
	    return primitive(context, type, object, false);
	}
	if (object == null){
	    return null;
	} else if (type.isInstance(object)){
	    return object;
	} else if (type.isArray() && isArray(object)){
	    for (int i = 0; i < Array.getLength(object); i++){
		int s = matchType(type.getComponentType(), Array.get(object, i));
		if (s < 0){
		    throw new ClassCastException("(" + Format.getClassName(type) + ")" + Pnuts.format(object));
		}
	    }
	    if (flag){
		return transform(getBottomType(type), object);
	    } else {
		return object;
	    }
	}
	throw new ClassCastException("(" + Format.getClassName(type) + ")" + Pnuts.format(object));
    }

    /**
     * Check if the parameter is an array
     */
    public final static boolean isArray(Object obj){
	return (obj instanceof Object[] ||
		obj instanceof int[] ||
		obj instanceof char[] ||
		obj instanceof boolean[] ||
		obj instanceof byte[] ||
		obj instanceof double[] ||
		obj instanceof long[] ||
		obj instanceof short[] ||
		obj instanceof float[]);
    }

    /**
     * This method is called by the syntax "id[from..to] = sth"
     */
    protected static Object assignRange(Object target, Number idx1, Number idx2, Object expr){
	int from = idx1.intValue();
	int to = 0;
	if (idx2 != null){
	    to = idx2.intValue();
	    if (to < 0){
		throw new ArrayIndexOutOfBoundsException("" + to);
	    }
	}
	if (target instanceof String){
	    StringBuffer s = new StringBuffer((String)target);
	    if (idx2 == null){
		to = s.length() - 1;
	    }
	    if (expr instanceof Character){
		char c = ((Character)expr).charValue();
		for (int i = from; i < to + 1; i++){
		    s.setCharAt(i, c);
		}
	    } else {
		s.replace(from, to + 1, String.valueOf(expr));
	    }
	    String val = s.toString();
	    return val;
	} else if (isArray(target)){
	    int len = Array.getLength(target);
	    if (idx2 == null){
		to = len - 1;
	    } else if (to > len - 1){
		throw new ArrayIndexOutOfBoundsException("" + to);
	    }
	    int size = to - from + 1;
	    if (from + size > len){
		size = len - from;
	    }
	    for (int i = from; i < to + 1; i++){
		Array.set(target, i, expr);
	    }
	    return target;
	} else {
	    throw new IllegalArgumentException();
	}
    }

    protected static void checkException(Context context, Throwable throwable){
	checkException(context, throwable, (Hashtable)context.get(Context.exceptionHandlerTableSymbol));
    }

    /**
     * @param context	the Context in which the exception is checked
     * @param throwable	the exception
     * @param tab	the exception handler table
     */
    protected static void checkException(Context context, Throwable throwable, Hashtable tab){
	Throwable e = throwable;
	if (e instanceof PnutsException){
	    throwable = ((PnutsException)e).getThrowable();
	}
	if (throwable instanceof Escape){
	    throw (Escape)throwable;
	}
	if (tab == null){
	    if (e instanceof PnutsException){
		throw (PnutsException)e;
	    } else {
		throw new PnutsException(throwable, context);
	    }
	} else {
	    Class clazz = throwable.getClass();
	    PnutsFunction f = null;
	    while (clazz != null){
		f = (PnutsFunction)tab.get(clazz);
		if (f != null){
		    break;
		}
		clazz = clazz.getSuperclass();
	    }
	    if (f != null){
		Object ret = f.exec(new Object[]{e}, context);
		context.getOutputStream().flush();
		throw new Jump(ret);
	    } else {
		if (e instanceof PnutsException){
		    throw (PnutsException)e;
		} else {
		    throw new PnutsException(throwable, context);
		}
	    }
	}
    }

    /**
     * This method is called when catch() function is called in a
     * package(non-local) scope 
     */
    protected static void catchException(Class t, PnutsFunction f, Context context){
	Package p = context.getCurrentPackage();
	Hashtable table = (Hashtable)p.get(Context.exceptionHandlerTableSymbol);
	if (table == null){
	    table = new Hashtable(10);
	    p.set(Context.exceptionHandlerTableSymbol, table, context);
	}
	if (f == null){
	    table.remove(t);
	} else {
	    table.put(t, f);
	}
    }

    protected final static Object getIndex(Object target, int index, Context context){
	if (target instanceof Object[]){
	    return ((Object[])target)[index];
	} else if (target instanceof int[]){
	    return new Integer(((int[])target)[index]);
	} else if (target instanceof char[]){
	    return new Character(((char[])target)[index]);
	} else if (target instanceof byte[]){
	    return new Byte(((byte[])target)[index]);
	} else if (target instanceof boolean[]){
	    return new Boolean(((boolean[])target)[index]);
	} else if (target instanceof double[]){
	    return new Double(((double[])target)[index]);
	} else if (target instanceof float[]){
	    return new Float(((float[])target)[index]);
	} else if (target instanceof long[]){
	    return new Long(((long[])target)[index]);
	} else if (target instanceof short[]){
	    return new Short(((short[])target)[index]);
	} else if (target instanceof Indexed){
	    return ((Indexed)target).get(index);
	} else if (target instanceof String){
	    return new Character(((String)target).charAt(index));
	} else {
	    throw new PnutsException("illegal.subscript", new Object[]{}, context);
	}
    }

    protected static Object newInstance(Context context, Class clazz, Object[] args, Class[] casts){
	if (clazz.isPrimitive()){
	    return primitive(context, clazz, args[0], true);
	} else if (clazz.isArray()){
	    return cast(context, clazz, args[0], true);
	} else {
	    return callConstructor(context, clazz, args, casts);
	}
    }

    protected static Object callFunction(Context context, PnutsFunction func, Object[] args){
	if (func == PnutsFunction.PACKAGE){
	    if (args.length > 0){
		if (args[0] instanceof Package){
		    ((Package)args[0]).init(context);
		} else {
		    Package.getPackage((String)args[0]).init(context);
		}
	    }
	    return context.getCurrentPackage();
	} else {
	    return func.exec(args, context);
	}
    }

    protected static Visitor getInterpreter(){
	return Pnuts.interpreter;
    }

    protected static void setInterpreter(Visitor visitor){
	Pnuts.interpreter = visitor;
    }

    protected static SimpleNode getNodes(Pnuts pnuts){
	return pnuts.startNodes;
    }

    protected static void doReturn(Object v){
	throw new Jump(v);
    }

    /**
     * Set line number information for error reporting
     */
    protected static void setLine(Context context, int beginLine){
	context.beginLine = beginLine;
    }
}
