/*
 * @(#) SymbolTable.java	1.2 99/06/03
 *
 * Copyright (c) 1997-1999 Sun Microsystems, Inc. All Rights Reserved.
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package pnuts.lang;

import java.util.Enumeration;
import java.util.NoSuchElementException;
import java.io.Serializable;

class SymbolTable implements Serializable {
    Binding[] table;
    SymbolTable parent;
    int count = 0;
    static final float THRESHOLD = 0.7f;

    public SymbolTable() {
	this(8);
    }

    public SymbolTable(int initialCapacity){
	table = new Binding[initialCapacity];
    }

    public SymbolTable(SymbolTable parent){
	this();
	this.parent = parent;
    }

    public Binding lookup(String name){
	int hash = System.identityHashCode(name);
	for (SymbolTable env = this; env != null; env = env.parent){
	    Binding[] env_tab = env.table;
	    int index = (hash & 0x7FFFFFFF) % env_tab.length;
	    for (Binding binding = env_tab[index]; binding != null; binding = binding.chain){
		if (binding.name == name){
		    return binding;
		}
	    }
	}
	return null;
    }

    public Binding lookup0(String name){
	int hash = System.identityHashCode(name);
	for (Binding binding = table[(hash & 0x7FFFFFFF) % table.length];
	     binding != null;
	     binding = binding.chain){

	    if (binding.name == name){
		return binding;
	    }
	}
	return null;
    }

    public Binding define(String name, Object value){
	if (count >= table.length * THRESHOLD){
	    rehash(2 * table.length);
	}

	count++;
	int hash = System.identityHashCode(name);
	int index = (hash & 0x7FFFFFFF) % table.length;

	Binding binding = new Binding(name, value);
	binding.chain = table[index];
	table[index] = binding;

	return binding;
    }

    void rehash(int new_capacity){
	Binding[] new_table = new Binding[new_capacity];
	for (int i = table.length;  --i >= 0;) {
	    Binding prev = null;
	    for (Binding cur = table[i];  cur != null; ){
		Binding next = cur.chain;
		cur.chain = prev;
		prev = cur;
		cur = next;
	    }
	    table[i] = prev;

	    for (Binding cur = table[i]; cur != null; ){
		int hash = System.identityHashCode(cur.name);
		int new_index = (hash & 0x7FFFFFFF) % new_capacity;
		Binding next = cur.chain;
		cur.chain = new_table[new_index];
		new_table[new_index] = cur;
		cur = next;
	    }
	}
	table = new_table;
    }

    public Object remove(String name){
	return _remove(name.intern());
    }

    public Object _remove(String name){
	int hash = System.identityHashCode(name);
	SymbolTable env = this;
	for ( ; ;  env = env.parent) {
	    if (env == null){
		return null;
	    }
	    Binding[] env_tab = env.table;
	    int index = (hash & 0x7FFFFFFF) % env_tab.length;
	    for (Binding binding = env_tab[index]; binding != null; binding = binding.chain){
		if (binding.name == name){
		    Object old = binding.get();
		    env.remove(binding);
		    return old;
		}
	    }
	}
    }

    void remove(Binding binding){
	int hash = System.identityHashCode(binding.name);
	int index = (hash & 0x7FFFFFFF) % table.length;
	Binding prev = null;
	for (Binding b = table[index];  b != null ; ){
	    Binding next = b.chain;
	    if (b == binding){
		if (prev == null){
		    table[index] = next;
		} else {
		    prev.chain = next;
		}
		count--;
		return;
	    }
	    prev = b;
	    b = next;
	}
    }

    public Object get(String interned) {
	Binding binding = lookup(interned);
	return binding == null ? null : binding.get();
    }

    public void put(String interned, Object value){
	Binding binding = lookup(interned);
	if (binding == null) {
	    define(interned, value);
	} else {
	    binding.set(value);
	}
    }

    public void put0(String interned, Object value){
	Binding binding = lookup0(interned);
	if (binding == null) {
	    define(interned, value);
	} else {
	    binding.set(value);
	}
    }

    public int size(){
	return count;
    }

    public synchronized Enumeration keys() {
	return new Enumerator();
    }

    private class Enumerator implements Enumeration {
	Binding bind = null;
	int index = table.length;

	public boolean hasMoreElements(){
	    while (bind == null && index > 0){
		bind = table[--index];
	    }
	    return bind != null;
	}

	public Object nextElement(){
	    while (bind == null && index > 0){
		bind = table[--index];
	    }
	    if (bind != null){
		Binding b = bind;
		bind = b.chain;
		return b.name;
	    }
	    throw new NoSuchElementException("SymbolTable Enumerator");
	}
    }
}
