/*
 * @(#) Translator.java	1.5 99/06/03
 *
 * Copyright (c) 1997-1999 Sun Microsystems, Inc. All Rights Reserved.
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package pnuts.trans;

import java.io.IOException;
import java.util.Enumeration;
import java.math.BigInteger;
import java.math.BigDecimal;
import pnuts.lang.Context;
import pnuts.lang.Runtime;
import pnuts.lang.Pnuts;
import pnuts.lang.SimpleNode;
import pnuts.lang.Visitor;
import pnuts.lang.Jump;
import pnuts.lang.PnutsParserTreeConstants;
import pnuts.lang.PnutsFunction;
import pnuts.lang.PnutsException;
import pnuts.lang.Function;
import pnuts.lang.Package;
import pnuts.ext.UnparseVisitor;

/**
 * Pnuts to Java translator
 * 
 * <pre>
 * Usage:
 *
 * function trans(expr_or_stream, className, output){
 *   p = Pnuts::parse(expr_or_stream)
 *   cc = CompileContext()
 *   cc.setName("translator")
 *   p.accept(Translator(className), cc)
 *   cc.writeTo(output)
 *   output.flush()
 * }
 *
 * trans("function fib(n) if (n < 2) n else fib(n - 1) + fib(n - 2)", "fib")
 * 
 * C:\> javac fib.java
 * C:\> pnuts
 * > fib::main(getContext())
 * > fib(25)
 * </pre>
 *
 * For more information, refer to the <a href="/pnuts/doc/trans.html">User's Guide</a>
 * 
 * @version	1.2
 * @see		pnuts.trans.CompileContext
 * @see		pnuts.lang.Visitor
 * @see		pnuts.lang.PnutsParserTreeConstants
 */
public class Translator implements Visitor, PnutsParserTreeConstants {

    String className;
    String packageName;
    boolean _includeLineNo = true;

    static Integer ONE = new Integer(1);

    /*
     * How many expressions are in a Java method.
     */
    static int NUMBER_OF_LINES_PER_METHOD = 100;

    public Translator(){
	this("Main");
    }

    public Translator(String className){
	int idx = className.lastIndexOf('.');
	if (idx > 0){
	    this.className = className.substring(idx + 1);
	    this.packageName = className.substring(0, idx);
	} else {
	    this.className = className;
	}
    }

    public void includeLineNo(boolean flag){
	_includeLineNo = flag;
    }

    public Object startSet(SimpleNode node, Context context){
	CompileContext cc = (CompileContext)context;
	if (packageName != null){
	    cc.emit("package " + packageName + ";\n");
	}
	cc.emit("import java.lang.reflect.*;\n");
	cc.emit("import java.util.*;\n");
	cc.emit("import java.io.*;\n");
	cc.emit("import pnuts.lang.*;\n");
	cc.emit("import pnuts.lang.Runtime;\n");
	cc.emit("import pnuts.lang.Package;\n");
	cc.emit("public class " + className + " extends pnuts.lang.Runtime {\n");
	cc.emit("static Object top = null;\n");
	cc.emit("static Object[] NO_PARAM = new Object[]{};\n");

	int n = node.jjtGetNumChildren();

	int m = (n + NUMBER_OF_LINES_PER_METHOD - 1) / NUMBER_OF_LINES_PER_METHOD;
	String names[] = new String[m];
	for (int i = 0, k = 0; i < m; i++, k += NUMBER_OF_LINES_PER_METHOD) {
	    String s1 = gensym(context);
	    names[i] = s1;
	    cc.emit("public static Object " + s1 + "(Context context){\n");

	    cc._declare("!exception");
	    Reference eref = cc.getReference("!exception");
	    cc.emit("Hashtable " + eref.symbol + " = null; // exception handler\n");

	    for (int j = 0; j < NUMBER_OF_LINES_PER_METHOD && k + j < n; j++){
		String s = (String)accept(node, k + j, context);
		if (s != null){
		    cc.emit(cc.top + " = " + s + ";\n");
		}
	    }
	    cc.emit("return " + cc.top + ";\n");
	    cc.emit("}\n");
	}

	cc.emit("public static void main(String a[]) throws Throwable {\n");
	cc.emit("Pnuts.load(\"init.pnut\");\n");
	cc.emit("main(new Context());\n");
	cc.emit("}\n");

	cc.emit("public static Object main(Context context) {\n");
	cc.emit("try {\n");

	for (int i = 0; i < m; i++) {
	    cc.emit(cc.top + " = " + names[i] + "(context);\n");
	}
	cc.emit("return " + cc.top + ";\n");
	cc.emit("} catch (Jump jump){\n");
	cc.emit("return jump.getValue();\n");
	cc.emit("} finally {\n");
	cc.emit(cc.top + " = null;\n");
	cc.emit("PrintWriter pw = context.getOutputStream();\n");
	cc.emit("if (pw != null) pw.flush();\n");
	cc.emit("}\n");
	cc.emit("}\n");

	for (Enumeration e = cc.constants.keys(); e.hasMoreElements();){
	    Object obj = e.nextElement();
	    String k = (String)cc.constants.get(obj);
	    if (obj instanceof Integer){
		cc.emit("static Integer " + k + " = new Integer((int)" + obj + ");\n");
	    } else if (obj instanceof Byte){
		cc.emit("static Byte " + k + " = new Byte((byte)" + obj + ");\n");
	    } else if (obj instanceof Character){
		cc.emit("static Character " + k + " = new Character((char)" + (int)((Character)obj).charValue() + ");\n");
	    } else if (obj instanceof Long){
		cc.emit("static Long " + k + " = new Long((long)" + obj + ");\n");
	    } else if (obj instanceof Float){
		cc.emit("static Float " + k + " = new Float((float)" + obj + ");\n");
	    } else if (obj instanceof Double){
		cc.emit("static Double " + k + " = new Double((double)" + obj + ");\n");
	    } else if (obj instanceof BigDecimal){
		cc.emit("static BigDecimal " + k + " = new BigDecimal(\"" + obj + "\");\n");
	    } else if (obj instanceof BigInteger){
		cc.emit("static BigInteger " + k + " = new BigInteger(\"" + obj + "\");\n");
	    }
	}
	cc.emit("}\n");
	return null;
    }

    static String gensym(Context context){
	return ((CompileContext)context).sym.gen();
    }

    public Object start(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	String s = (String)accept(node, 0, context);
	if (s != null){
	    cc.emit(cc.top + " = " + s + ";\n");
	}
	return null;
    }

    public Object expressionList(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	int n = node.jjtGetNumChildren();
	for (int i = 0; i < n; i++) {
	    SimpleNode _node = node.jjtGetChild(i);
	    String s = (String)accept(node, i, context);
	    if (s != null){
		cc.emit(cc.top + " = " + s + ";\n");
	    }
	}
	return null;
    }

    public Object integerNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;

	String str = node.str;
	Object p[] = Runtime.parseInt(str);
	Number n = (Number)p[0];

	String assoc = (String)cc.constants.get(n);
	if (assoc == null){
	    assoc = gensym(context);
	    cc.constants.put(n, assoc);
	}
	String top = assoc;
	if (p[1] != null){
	    int offset = ((int[])p[1])[0];
	    top = "Runtime.quantity((Number)" + top + ", " + str.substring(0, offset) + ", " + str.substring(offset) + ", context)";
	}
	return top;
    }

    public Object floatingNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;

	Object p[] = Runtime.parseFloat(node.str);
	Number n = (Number)p[0];

	String assoc = (String)cc.constants.get(n);
	if (assoc == null){
	    assoc = gensym(context);
	    cc.constants.put(n, assoc);
	}
	String top = assoc;
	if (p[1] != null){
	    int offset = ((int[])p[1])[0];
	    top = "Runtime.quantity((Number)" + top + ", " + node.str.substring(0, offset) + ", " + node.str.substring(offset) + ", context)";
	}
	return top;
    }

    public Object stringNode(SimpleNode node, Context context) {
	return node.str;
    }

    public Object characterNode(SimpleNode node, Context context){
	CompileContext cc = (CompileContext)context;
	Character ch = Runtime.parseChar(node.str);

	String assoc = (String)cc.constants.get(ch);
	if (assoc != null){
	    return assoc;
	} else {
	    String s1 = gensym(context);
	    cc.constants.put(ch, s1);
	    return s1;
	}
    }

    public Object classNode(SimpleNode node, Context context){
	CompileContext cc = (CompileContext)context;

	if (_includeLineNo){
	    if (cc.getLine() != node.beginLine){
		cc.emit("setLine(context, " + node.beginLine + ");\n");
		cc.setLine(node.beginLine);
	    }
	}

	String name = node.str;
	int n = node.jjtGetNumChildren();

	for (int i = 0; i < n; i++){
	    SimpleNode ch = node.jjtGetChild(i);
	    if (ch.id == JJTPACKAGE){
		name += ".";
		name += ch.str;
	    }
	}

	cc.emit("try { " + cc.top + " = Pnuts.loadClass(\"" + name + "\"); } catch (ClassNotFoundException e){ throw new PnutsException(e, context);}\n");
	return null;
    }

    public Object andNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;

	String s1 = gensym(context);
	String s = (String)accept(node, 0, context);
	if (s != null){
	    cc.emit(cc.top + " = " + s + ";\n");
	}

	int id = node.jjtGetChild(0).id;
	if (id != JJTTRUENODE &&
	    id != JJTFALSENODE &&
	    id != JJTLENODE &&
	    id != JJTLTNODE &&
	    id != JJTGENODE &&
	    id != JJTGTNODE &&
	    id != JJTEQUALNODE &&
	    id != JJTNOTEQNODE &&
	    id != JJTINSTANCEOFEXPRESSION &&
	    id != JJTORNODE &&
	    id != JJTANDNODE &&
	    id != JJTNOTNODE){
	    cc.emit("if (!(" + cc.top + " instanceof Boolean)){\n");
	    cc.emit( "throw new PnutsException(\"boolean.expected\", NO_PARAM, context);\n");
	    cc.emit("}\n");
	}

	cc.emit("if (((Boolean)" + cc.top + ").booleanValue()){\n");

	s = (String)accept(node, 1, context);
	if (s != null){
	    cc.emit(cc.top + " = " + s + ";\n");
	}

	id = node.jjtGetChild(1).id;
	if (id != JJTTRUENODE &&
	    id != JJTFALSENODE &&
	    id != JJTLENODE &&
	    id != JJTLTNODE &&
	    id != JJTGENODE &&
	    id != JJTGTNODE &&
	    id != JJTEQUALNODE &&
	    id != JJTNOTEQNODE &&
	    id != JJTINSTANCEOFEXPRESSION &&
	    id != JJTORNODE &&
	    id != JJTANDNODE &&
	    id != JJTNOTNODE){
	    cc.emit("if (!(" + cc.top + " instanceof Boolean)){\n");
	    cc.emit( "throw new PnutsException(\"boolean.expected\", NO_PARAM, context);\n");
	    cc.emit("}\n");
	}
	cc.emit("}\n");
	return null;
    }

    public Object orNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	String s = (String)accept(node, 0, context);
	if (s != null){	
	    cc.emit(cc.top + " = " + s + ";\n");
	}
	int id = node.jjtGetChild(0).id;
	if (id != JJTTRUENODE &&
	    id != JJTFALSENODE &&
	    id != JJTLENODE &&
	    id != JJTLTNODE &&
	    id != JJTGENODE &&
	    id != JJTGTNODE &&
	    id != JJTEQUALNODE &&
	    id != JJTNOTEQNODE &&
	    id != JJTINSTANCEOFEXPRESSION &&
	    id != JJTORNODE &&
	    id != JJTANDNODE &&
	    id != JJTNOTNODE){
	    cc.emit("if (!(" + cc.top + " instanceof Boolean)){\n");
	    cc.emit( "throw new PnutsException(\"boolean.expected\", NO_PARAM, context);\n");
	    cc.emit("}\n");
	}

	cc.emit("if (!((Boolean)" + cc.top + ").booleanValue()){\n");

	s = (String)accept(node, 1, context);
	if (s != null){
	    cc.emit(cc.top + " = " + s + ";\n");
	}
	id = node.jjtGetChild(1).id;
	if (id != JJTTRUENODE &&
	    id != JJTFALSENODE &&
	    id != JJTLENODE &&
	    id != JJTLTNODE &&
	    id != JJTGENODE &&
	    id != JJTGTNODE &&
	    id != JJTEQUALNODE &&
	    id != JJTNOTEQNODE &&
	    id != JJTINSTANCEOFEXPRESSION &&
	    id != JJTORNODE &&
	    id != JJTANDNODE &&
	    id != JJTNOTNODE){
	    cc.emit( "if (!(" + cc.top + " instanceof Boolean)){\n");
	    cc.emit(  "throw new PnutsException(\"boolean.expected\", NO_PARAM, context);\n");
	    cc.emit( "}\n");
	}
	cc.emit( "}\n");
	return null;
    }

    public Object notNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	String s = (String)accept(node, 0, context);
	String s1 = gensym(context);
	if (s == null){
	    s = cc.top;
	}
	cc.emit("Object " + s1 + " = " + s + ";\n");

	cc.emit("try {\n");
	cc.emit("if (" + s1 + " instanceof Boolean){\n");
	cc.emit(cc.top + " = (((Boolean)" + s1 + ").booleanValue()) ? Boolean.FALSE : Boolean.TRUE;\n");

	cc.emit("} else {\n");
	cc.emit("throw new PnutsException(\"boolean.expected\", NO_PARAM, context);\n");
	cc.emit("}\n");
	cc.emit("} catch (IllegalArgumentException e){\n");
	cc.emit("throw new PnutsException(\"illegal.operandType\", NO_PARAM, context);\n");
	cc.emit("}\n");
	return null;
    }

    public Object shiftLeftNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	cc.emit("try {\n");
	String s1 = gensym(context);
	String s = (String)accept(node, 0, context);
	if (s == null){
	    s = cc.top;
	}
	cc.emit("Object " + s1 + " = " + s + ";\n");

	s = (String)accept(node, 1, context);
	if (s == null){
	    s = cc.top;
	}
	cc.emit(cc.top + " = Arithmetic.shiftLeft(" + s1 + ", " + s + ");\n");

	cc.emit("} catch (IllegalArgumentException e){\n");
	cc.emit( "throw new PnutsException(\"illegal.operandType\", NO_PARAM, context);\n");
	cc.emit("}\n");
	return null;
    }

    public Object shiftRightNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	cc.emit("try {\n");
	String s1 = gensym(context);
	String s = (String)accept(node, 0, context);
	if (s == null){
	    s = cc.top;
	}
	cc.emit("Object " + s1 + " = " + s + ";\n");

	s = (String)accept(node, 1, context);
	if (s == null){
	    s = cc.top;
	}
	cc.emit(cc.top + " = Arithmetic.shiftRight(" + s1 + ", " + s + ");\n");

	cc.emit("} catch (IllegalArgumentException e){\n");
	cc.emit( "throw new PnutsException(\"illegal.operandType\", NO_PARAM, context);\n");
	cc.emit("}\n");
	return null;
    }

    public Object shiftArithmeticNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	cc.emit("try {\n");
	String s1 = gensym(context);
	String s = (String)accept(node, 0, context);
	if (s == null){
	    s = cc.top;
	}
	cc.emit("Object " + s1 + " = " + s + ";\n");

	s = (String)accept(node, 1, context);
	if (s == null){
	    s = cc.top;
	}
	cc.emit(cc.top + " = Arithmetic.shiftArithmetic(" + s1 + ", " + s + ");\n");

	cc.emit("} catch (IllegalArgumentException e){\n");
	cc.emit( "throw new PnutsException(\"illegal.operandType\", NO_PARAM, context);\n");
	cc.emit("}\n");
	return null;
    }

    public Object addNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	String s = (String)accept(node, 0, context);
	String s1 = gensym(context);
	if (s == null){
	    s = cc.top;
	}
	cc.emit("Object " + s1 + " = " + s + ";\n");

	s = (String)accept(node, 1, context);
	if (s == null){
	    s = cc.top;
	}
	return "Arithmetic.add(" + s1 + ", " + s + ")";
    }

    public Object subtractNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	String s1 = gensym(context);
	String s = (String)accept(node, 0, context);
	if (s == null){
	    s = cc.top;
	}
	cc.emit("Object " + s1 + " = " + s + ";\n");

	s = (String)accept(node, 1, context);
	if (s == null){
	    s = cc.top;
	}
	return "Arithmetic.subtract(" + s1 + ", " + s + ")";
    }

    public Object multNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	String s1 = gensym(context);
	String s = (String)accept(node, 0, context);
	if (s == null){
	    s = cc.top;
	}
	cc.emit("Object " + s1 + " = " + s + ";\n");

	s = (String)accept(node, 1, context);
	if (s == null){
	    s = cc.top;
	}
	return "Arithmetic.multiply(" + s1 + ", " + s + ")";
    }

    public Object divideNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	String s1 = gensym(context);
	String s = (String)accept(node, 0, context);
	if (s == null){
	    s = cc.top;
	}
	cc.emit("Object " + s1 + " = " + s + ";\n");

	s = (String)accept(node, 1, context);
	if (s == null){
	    s = cc.top;
	}
	return "Arithmetic.divide(" + s1 + ", " + s + ")";
    }

    public Object negativeNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	String s1 = gensym(context);
	String s = (String)accept(node, 0, context);
	if (s == null){
	    s = cc.top;
	}
	cc.emit("Object " + s1 + " = " + s + ";\n");

	cc.emit("try {\n");
	cc.emit( "if (" + s1 + " instanceof Numeric || " + s1 + " instanceof Number){\n");
	cc.emit(  cc.top + " = Arithmetic.negate(" + s1 + ");\n");
	cc.emit( "} else {\n");
	cc.emit(  "throw new PnutsException(\"illegal.operandType\", NO_PARAM, context);\n");
	cc.emit( "}\n");
	cc.emit("} catch(IllegalArgumentException e){\n");
	cc.emit( "throw new PnutsException(\"illegal.operandType\", NO_PARAM, context);\n");
	cc.emit("}\n");
	return null;
    }

    public Object modNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	String s1 = gensym(context);
	String s = (String)accept(node, 0, context);
	if (s == null){
	    s = cc.top;
	}
	cc.emit("Object " + s1 + " = " + s + ";\n");

	s = (String)accept(node, 1, context);
	cc.emit("try {\n");
	if (s == null){
	    s = cc.top;
	}
	cc.emit(cc.top + " = Arithmetic.mod(" + s1 + ", " + s + ");\n");

	cc.emit("} catch(IllegalArgumentException e){\n");
	cc.emit("throw new PnutsException(\"illegal.operandType\", NO_PARAM, context);\n");
	cc.emit("}\n");
	return null;
    }

    public Object logExorNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	cc.emit("try {\n");
	String s1 = gensym(context);
	String s = (String)accept(node, 0, context);
	if (s != null){
	    cc.emit("Object " + s1 + " = " + s + ";\n");
	} else {
	    cc.emit("Object " + s1 + " = " + cc.top + ";\n");
	}
	s = (String)accept(node, 1, context);
	if (s == null){
	    s = cc.top;
	}
	cc.emit(cc.top + " = Arithmetic.xor(" + s1 + ", " + s + ");\n");

	cc.emit("} catch (IllegalArgumentException e){\n");
	cc.emit( "throw new PnutsException(\"illegal.operandType\", NO_PARAM, context);\n");
	cc.emit("}\n");
	return null;
    }

    public Object logOrNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	cc.emit("try {\n");
	String s1 = gensym(context);
	String s = (String)accept(node, 0, context);
	if (s == null){
	    s = cc.top;
	}
	cc.emit("Object " + s1 + " = " + s + ";\n");

	s = (String)accept(node, 1, context);
	if (s == null){
	    s = cc.top;
	}
	cc.emit(cc.top + " = Arithmetic.or(" + s1 + ", " + s + ");\n");

	cc.emit("} catch (IllegalArgumentException e){\n");
	cc.emit( "throw new PnutsException(\"illegal.operandType\", NO_PARAM, context);\n");
	cc.emit("}\n");
	return null;
    }

    public Object logAndNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	cc.emit("try {\n");
	String s1 = gensym(context);
	String s = (String)accept(node, 0, context);
	if (s == null){
	    s = cc.top;
	}
	cc.emit("Object " + s1 + " = " + s + ";\n");

	s = (String)accept(node, 1, context);
	if (s == null){
	    s = cc.top;
	}
	cc.emit(cc.top + " = Arithmetic.and(" + s1 + ", " + s + ");\n");

	cc.emit("} catch (IllegalArgumentException e){\n");
	cc.emit( "throw new PnutsException(\"illegal.operandType\", NO_PARAM, context);\n");
	cc.emit("}\n");
	return null;
    }

    public Object logNotNode(SimpleNode node, Context context){
	CompileContext cc = (CompileContext)context;
	cc.emit("try {\n");
	String s = (String)accept(node, 0, context);
	if (s == null){
	    s = cc.top;
	}
	cc.emit(cc.top + " = Arithmetic.not(" + s + ");\n");
	cc.emit("} catch (IllegalArgumentException e){\n");
	cc.emit("throw new PnutsException(\"illegal.operandType\", NO_PARAM, context);\n");
	cc.emit("}\n");
	return null;
    }

    public Object ltNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	String s = (String)accept(node, 0, context);
	String s1 = gensym(context);
	if (s == null){
	    s = cc.top;
	}
	cc.emit("Object " + s1 + " = " + s + ";\n");
	s = (String)accept(node, 1, context);
	if (s == null){
	    s = cc.top;
	}
	return "((Arithmetic.compareTo(" + s1 + ", " + s + ") == Numeric.RIGHT_IS_BIGGER) ? Boolean.TRUE : Boolean.FALSE)";
    }

    public Object leNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	String s1 = gensym(context);
	String s2 = gensym(context);
	cc.emit("try {\n");
	String s = (String)accept(node, 0, context);
	if (s == null){
	    s = cc.top;
	}
	cc.emit("Object " + s1 + " = " + s + ";\n");
	s = (String)accept(node, 1, context);
	if (s == null){
	    s = cc.top;
	}
	cc.emit("int " + s2 + " = Arithmetic.compareTo(" + s1 + ", " + s + ");\n");
	cc.emit("if (" + s2 + " == Numeric.NOT_EQUAL){\n");
	cc.emit( "throw new PnutsException(\"illegal.operandType\", NO_PARAM, context);\n");
	cc.emit("}\n");
	cc.emit("if (" + s2 + " == Numeric.EQUAL || " + s2 + " == Numeric.RIGHT_IS_BIGGER){\n");
	cc.emit(cc.top + " = Boolean.TRUE;\n");
	cc.emit("} else {\n");
	cc.emit(cc.top + " = Boolean.FALSE;\n");
	cc.emit("}\n");
	cc.emit("} catch (IllegalArgumentException e){\n");
	cc.emit( "throw new PnutsException(\"illegal.operandType\", NO_PARAM, context);\n");
	cc.emit("}\n");
	return null;
    }

    public Object gtNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	String s1 = gensym(context);
	String s = (String)accept(node, 0, context);
	if (s == null){
	    s = cc.top;
	}
	cc.emit("Object " + s1 + " = " + s + ";\n");
	s = (String)accept(node, 1, context);
	if (s == null){
	    s = cc.top;
	}
	return "((Arithmetic.compareTo(" + s1 + ", " + s + ") == Numeric.LEFT_IS_BIGGER) ? Boolean.TRUE : Boolean.FALSE)";
    }

    public Object geNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	String s1 = gensym(context);
	String s2 = gensym(context);
	cc.emit("try {\n");
	String s = (String)accept(node, 0, context);
	if (s == null){
	    s = cc.top;
	}
	cc.emit("Object " + s1 + " = " + s + ";\n");

	s = (String)accept(node, 1, context);
	if (s == null){
	    s = cc.top;
	}
	cc.emit("int " + s2 + " = Arithmetic.compareTo(" + s1 + ", " + s + ");\n");
	cc.emit("if (" + s2 + " == Numeric.NOT_EQUAL){\n");
	cc.emit( "throw new PnutsException(\"illegal.operandType\", NO_PARAM, context);\n");
	cc.emit("}\n");
	cc.emit("if (" + s2 + " == Numeric.EQUAL || " + s2 + " == Numeric.LEFT_IS_BIGGER){\n");
	cc.emit(cc.top + " = Boolean.TRUE;\n");
	cc.emit("} else {\n");
	cc.emit(cc.top + " = Boolean.FALSE;\n");
	cc.emit("}\n");
	cc.emit("} catch (IllegalArgumentException e){\n");
	cc.emit( "throw new PnutsException(\"illegal.operandType\", NO_PARAM, context);\n");
	cc.emit("}\n");
	return null;
    }

    public Object equalNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;

	String s1 = gensym(context);
	String s = (String)accept(node, 0, context);
	if (s == null){
	    s = cc.top;
	}
	cc.emit("Object " + s1 + " = " + s + ";\n");
	s = (String)accept(node, 1, context);
	if (s == null){
	    s = cc.top;
	}
	return "((Arithmetic.compareTo(" + s1 + ", " + s + ") == Numeric.EQUAL) ? Boolean.TRUE : Boolean.FALSE)";
    }

    public Object notEqNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	String s1 = gensym(context);
	String s = (String)accept(node, 0, context);
	if (s == null){
	    s = cc.top;
	}
	cc.emit("Object " + s1 + " = " + s + ";\n");

	s = (String)accept(node, 1, context);
	if (s == null){
	    s = cc.top;
	}
	return "((Arithmetic.compareTo(" + s1 + ", " + s + ") == Numeric.EQUAL) ? Boolean.FALSE : Boolean.TRUE)";
    }

    public Object ifStatement(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;

	if (_includeLineNo && cc.getLine() != node.beginLine){
	    cc.emit("setLine(context, " + node.beginLine + ");\n");
	    cc.setLine(node.beginLine);
	}
	String s = (String)accept(node, 0, context);
	int id = node.jjtGetChild(0).id;
	if (id != JJTTRUENODE &&
	    id != JJTFALSENODE &&
	    id != JJTLENODE &&
	    id != JJTLTNODE &&
	    id != JJTGENODE &&
	    id != JJTGTNODE &&
	    id != JJTEQUALNODE &&
	    id != JJTNOTEQNODE &&
	    id != JJTINSTANCEOFEXPRESSION &&
	    id != JJTORNODE &&
	    id != JJTANDNODE &&
	    id != JJTNOTNODE){
	    if (s != null){
	        cc.emit(cc.top + " = " + s + ";\n");
	    }
	    cc.emit("if (!(" + cc.top + " instanceof Boolean)){\n");
	    cc.emit( "throw new PnutsException(\"boolean.expected\", NO_PARAM, context);\n");
	    cc.emit("}\n");
	    cc.emit("if (((Boolean)" + cc.top + ").booleanValue()){\n");
	} else {
	    if (s == null){
		s = cc.top;
	    }
	    cc.emit("if (((Boolean)" + s + ").booleanValue()){\n");
	}
	s = (String)accept(node, 1, context);
	if (s != null){
	    cc.emit(cc.top + " = " + s + ";\n");
	}
	cc.emit("} else {\n");
	int n = node.jjtGetNumChildren();
	int j = 0;
	for (int i = 2; i < n; i++){
	    SimpleNode _node = node.jjtGetChild(i);
	    if (_node.id == JJTELSEIFNODE){
		s = (String)accept(_node, 0, context);
		if (s != null){
		    cc.emit(cc.top + " = " + s + ";\n");
		}

		id = _node.jjtGetChild(0).id;
		if (id != JJTTRUENODE &&
		    id != JJTFALSENODE &&
		    id != JJTLENODE &&
		    id != JJTLTNODE &&
		    id != JJTGENODE &&
		    id != JJTGTNODE &&
		    id != JJTEQUALNODE &&
		    id != JJTNOTEQNODE &&
		    id != JJTINSTANCEOFEXPRESSION &&
		    id != JJTORNODE &&
		    id != JJTANDNODE &&
		    id != JJTNOTNODE){
		    if (s != null){
			cc.emit(cc.top + " = " + s + ";\n");
		    }
		    cc.emit("if (!(" + cc.top + " instanceof Boolean)){\n");
		    cc.emit( "throw new PnutsException(\"boolean.expected\", NO_PARAM, context);\n");
		    cc.emit("}\n");
		}

		cc.emit("if (((Boolean)" + cc.top + ").booleanValue()){\n");

		s = (String)accept(_node, 1, context);
		if (s != null){
		    cc.emit(cc.top + " = " + s + ";\n");
		}
	        cc.emit("} else {\n");
		j++;
	    } else if (_node.id == JJTELSENODE){
		s = (String)accept(_node, 0, context);
		if (s != null){
		    cc.emit(cc.top + " = " + s + ";\n");
		}
	    }
	}
	for (int i = 0; i <= j; i++){
	    cc.emit("}\n");
	}
	return null;
    }

    public Object functionStatement(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	SimpleNode block = null;
	SimpleNode param = null;
	SimpleNode n = node.jjtGetChild(0);
	if (n.id == JJTPARAMLIST){
	    param = n;
	    block = node.jjtGetChild(1);
	} else if (n.id == JJTBLOCK){
	    block = n;
	}
	int nargs = param.jjtGetNumChildren();
	String[] locals = new String[nargs];
	SimpleNode n0 = null;
	if (nargs == 1 &&
	    (n0 = param.jjtGetChild(0)).id == JJTINDEXNODE){
	    nargs = -1;
	    locals[0] = n0.jjtGetChild(0).str;
	} else {
	    for (int j = 0; j < nargs; j++){
		locals[j] = param.jjtGetChild(j).str;
	    }
	}

	String cls = gensym(context);
	cc.emit("class " + cls + " extends Function {\n");
	cc.emit(cls + "(String func, String[] locals, int nargs, SimpleNode node, Package pkg, Context context){\n");
	cc.emit(" super(func, locals, nargs, node, pkg, context);\n");
	cc.emit("}\n");

	cc.emit("public PnutsFunction register(PnutsFunction f){\n");
	cc.emit( "return super.register(f);\n");
	cc.emit("}\n");

	cc.emit("protected String unparse(Context context){\n");
	cc.emit(" return \"" + (String)node.accept(new UnparseVisitor(true), context) + "\";\n");
	cc.emit("}\n");

	String args = gensym(context);
	String fname = node.str;

	cc.openFrame(fname, locals, args);

	cc.emit("protected Object exec(final Object " + args + "[], Context context){\n");

	Reference fref = null;

	if (fname != null){
	    fref = cc.getReference(fname);
	    cc.emit("final Object " + fref.symbol + " = function;\n");
	}

	cc._declare("!top");
	Reference tref = cc.getReference("!top");
	cc.emit("Object " + tref.symbol + "= null;\n");

	cc._declare("!exception");
	Reference eref = cc.getReference("!exception");
	cc.emit("Hashtable " + eref.symbol + " = null; // exception handler\n");

	CompileContext tmp_context = (CompileContext)context.clone();
	block.accept(this, tmp_context);

	String[] maps = cc.env.bottom.maps;
	for (int i = 0; i < cc.env.bottom.count; i++){
	    cc.emit("final Object[] " + maps[i] + " = new Object[1];\n");
	}

	cc.emit("try {\n");
	try {
	    tmp_context.writeTo(cc.pw);
	} catch (IOException e){
	    throw new PnutsException(e, context);
	}
	cc.emit("return " + tref.symbol + ";\n");
	cc.emit("} catch (Throwable t){\n");
	cc.emit("checkException(context, t, (Hashtable)" + eref.symbol + ");\n");
	cc.emit("return null;\n");
	cc.emit("}\n");
	cc.emit("}\n");
	cc.emit("}\n");

	cc.closeFrame();

	String s2 = gensym(context);
	cc.emit("Package " + s2 + " = context.getCurrentPackage();\n");

	if (fname != null){
	    cc.emit(cc.top + " = new " + cls + "(\"" + fname + "\", new String[]{\n");
	} else {
	    cc.emit(cc.top + " = new " + cls + "(null, new String[]{\n");
	}
	
	if (locals.length > 0){
	    cc.emit("\"" + locals[0] + "\"\n");
	}
	for (int i = 1; i < locals.length; i++){
	    cc.emit(",\"" + locals[i] + "\"\n");
	}
	cc.emit("}, " + nargs + ", null, " + s2 + ", context);\n");

	String s1 = gensym(context);
	cc.emit("Object " + s1 + " = null;\n");

	Reference ref = null;
	if (fname != null){
	    ref = cc.env.getReference(fname);
	    if (ref != null && ref.offset >= 0){
		cc.emit(s1 + " = " + ref.symbol + "[" + ref.offset + "];\n");
	    } else {
		cc.emit("if (" + s2 + ".defined(((Function)" + cc.top + ").getName())){\n");
		cc.emit(s1 + " = "+ s2 + ".get(((Function)" + cc.top + ").getName(), context);\n");
		cc.emit("}\n");
		cc.emit("if (!(" + s1 + " instanceof PnutsFunction)){\n");
		cc.emit(s1 + " = null;\n");
		cc.emit("}\n");
	    }
	}

	cc.emit(cc.top + " = ((" + cls + ")" + cc.top + ").register((PnutsFunction)" + s1 + ");\n");

	if (fname == null){
	   return null;
	}

	if (cc.env.parent != null){
	    if (ref == null){
		cc.declare(fname);
		ref = cc.getReference(fname);
		cc.emit(ref.symbol + "[0] = " + cc.top + ";\n");
	    }
	} else {
	    cc.emit(s2 + ".set(\"" + fname + "\", " + cc.top + ", context);\n");
	}
	return null;
    }

    public Object applicationNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	SimpleNode argNode = node.jjtGetChild(1);
	int nargs = argNode.jjtGetNumChildren();

	String s = (String)accept(node, 0, context);
	String s1 = gensym(context);
	if (s == null){
	    s = cc.top;
	}
	cc.emit("Object " + s1 + " = " + s + ";\n");

	Reference eref = cc.getReference("!exception");

	cc.emit("if (" + s1 + " instanceof PnutsFunction){\n");

	s = (String)accept(node, 1, context);
	if (s == null){
	    s = cc.top;
	}

	String s2 = gensym(context);
	cc.emit("Object[] " + s2 + " = (Object[])" + s + ";\n");
	if (cc.env.parent != null){
	    if (cc.env.fname != null){
		Reference fref = cc.getReference(cc.env.fname);
		cc.emit("if (" + s1 + " == " + fref.symbol + " && nargs == " + nargs + "){\n");
		cc.emit(cc.top + " = exec(" + s2 + ", context);\n");
		cc.emit(" } else if (" + s1 + " == PnutsFunction.CATCH){\n");
	    } else {
		cc.emit("if (" + s1 + " == PnutsFunction.CATCH){\n");
	    }
	    cc.emit("  if (" + eref.symbol + " == null){\n");
	    cc.emit(     eref.symbol + " = new Hashtable(10);\n");
	    cc.emit("  }\n");
	    cc.emit("  if (" + s2 + "[1] == null){\n");
	    cc.emit(     eref.symbol + ".remove(" + s2 + "[0]);\n");
	    cc.emit("  } else {\n");
	    cc.emit(     eref.symbol + ".put(" + s2 + "[0], " + s2 + "[1]);\n");
	    cc.emit("  }\n");
	    cc.emit(" } else {\n");
	    cc.emit(cc.top + " = callFunction(context, (PnutsFunction)" + s1 + ", " + s2 + ");\n");
	    cc.emit("}\n");
	} else {
	    cc.emit("if (" + s1 + " == PnutsFunction.CATCH){\n");
	    cc.emit( "catchException((Class)" + s2 + "[0], (PnutsFunction)" + s2 + "[1], context);\n");
	    cc.emit(" } else {\n");
	    cc.emit(cc.top + " = callFunction(context, (PnutsFunction)" + s1 + ", " + s2 + ");\n");
	    cc.emit("}\n");
	}

	cc.emit("} else if (" + s1 + " instanceof Class){\n");  // constructor
	s = (String)accept(node, 1, context);
	if (s == null){
	    s = cc.top;
	}
	cc.emit("Object[] " + s2 + " = (Object[])" + s + ";\n");

	String s3 = gensym(context);
	boolean types_created = false;
	for (int i = 0; i < nargs; i++){
	    SimpleNode n = argNode.jjtGetChild(i);
	    if (n.id == JJTCASTEXPRESSION){
		if (!types_created){
		    cc.emit("Class[] " + s3 + " = new Class[" + nargs + "];\n");
		    types_created = true;
		}
		s = (String)accept(n, 0, context);
		if (s == null){
		    s = cc.top;
		}
		cc.emit(s3 + "[" + i + "] = (Class)" + s + ";\n");
	    }
	}
	if (types_created){
	    cc.emit(cc.top + " = newInstance(context, (Class)" + s1 + ", " + s2 + ", " + s3 + ");\n");
	} else {
	    cc.emit(cc.top + " = newInstance(context, (Class)" + s1 + ", " + s2 + ", null);\n");
	}
	cc.emit("} else {\n");
	cc.emit("throw new PnutsException(\"funcOrType.expected\", new Object[]{Pnuts.format(" + s1 + ")}, context);\n");
	cc.emit("}\n");
	return null;
    }

    public Object blockNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	int n = node.jjtGetNumChildren();

	for (int i = 0; i < n; i++) {
	    String s = (String)accept(node, i, context);
	    if (s != null){
	        cc.emit(cc.top + " = " + s + ";\n");
	    }
	}
	return null;
    }

    public Object trueNode(SimpleNode node, Context context) {
	return "Boolean.TRUE";
    }

    public Object falseNode(SimpleNode node, Context context) {
	return "Boolean.FALSE";
    }

    public Object nullNode(SimpleNode node, Context context) {
	return "null";
    }

    public Object idNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	if (_includeLineNo && cc.getLine() != node.beginLine){
	    cc.emit("setLine(context, " + node.beginLine + ");\n");
	    cc.setLine(node.beginLine);
	}
	String symbol = node.str;
	Reference ref = cc.getReference(symbol);
	if (ref != null){
	    if (ref.offset < 0){
		return ref.symbol;
	    } else {
		return ref.symbol + "[" + ref.offset + "]";
	    }
	} else {
	    return "context.getId(\"" + symbol + "\")";
	}
    }

    public Object global(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	String symbol = node.str;
	cc.emit("if (Package.getGlobalPackage().defined(\"" + symbol + "\")){\n");
	cc.emit(cc.top + "= Package.getGlobalPackage().get(\"" + symbol + "\", context);\n");
	cc.emit("} else {\n");
	cc.emit("throw new PnutsException(\"not.defined\", new Object[]{\"" + symbol + "\"}, context);\n");
	cc.emit("}\n");
	return null;
    }

    public Object listElements(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	int n = node.jjtGetNumChildren();
	String s1 = gensym(context);
	cc.emit("Object[] " + s1 + " = new Object[" + n + "];\n");
	if (n > 0){
	    String s = (String)accept(node, 0, context);
	    if (s == null){
		s = cc.top;
	    }
	    cc.emit(s1 + "[0] = " + s + ";\n");

	    for (int i = 1; i < n; i++){
		s = (String)accept(node, i, context);
		if (s == null){
		    s = cc.top;
		}
		cc.emit(s1 + "[" + i + "] = " + s + ";\n");
	    }
	}
	return s1;
    }

    public Object castExpression(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	String s1 = gensym(context);
	String s = (String)accept(node, 0, context);
	if (s == null){
	    s = cc.top;
	}
	cc.emit("Class " + s1 + " = (Class)" + s + ";\n");

	s = (String)accept(node, 1, context);
	if (s == null){
	    s = cc.top;
	}
	return "Runtime.cast(context," + s1 + ", " + s + ", true)";
    }

    public Object memberNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;

	if (_includeLineNo && cc.getLine() != node.beginLine){
	    cc.emit("setLine(context, " + node.beginLine + ");\n");
	    cc.setLine(node.beginLine);
	}
	String s = (String)accept(node, 0, context);
	if (s != null){
	    cc.emit(cc.top + " = " + s + ";\n");
	}
	cc.emit("if (" + cc.top + " == null){\n");
	cc.emit( "throw new PnutsException(new NullPointerException(), context);\n");
	cc.emit("} else if (Runtime.isArray(" + cc.top + ") && \"" + node.str + "\".equals(\"length\")){\n");
	cc.emit( cc.top + " = new Integer(Array.getLength(" + cc.top + "));\n");
	cc.emit("} else {\n");
	cc.emit( cc.top + " = Runtime.getField(context, " + cc.top + ", \"" + node.str + "\");\n");
	cc.emit("}\n");
	return null;
    }

    public Object methodNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	if (_includeLineNo && cc.getLine() != node.beginLine){
	    cc.emit("setLine(context, " + node.beginLine + ");\n");
	    cc.setLine(node.beginLine);
	}
	String s1 = gensym(context);
	String s2 = gensym(context);
	String s = (String)accept(node, 0, context);
	if (s == null){
	    s = cc.top;
	}
	cc.emit("Object " + s1 + " = " + s + ";\n");

	s = (String)accept(node, 1, context);
	if (s == null){
	    s = cc.top;
	}
	cc.emit("Object[] " + s2 + " = (Object[])" + s + ";\n");

	SimpleNode argNode = node.jjtGetChild(1);
	int nargs = argNode.jjtGetNumChildren();
	String s3 = gensym(context);
	boolean types_created = false;
	for (int i = 0; i < nargs; i++){
	    SimpleNode n = argNode.jjtGetChild(i);
	    if (n.id == JJTCASTEXPRESSION){
		if (!types_created){
		    cc.emit("Class[] " + s3 + " = new Class[" + nargs + "];\n");
		    types_created = true;
		}
		s = (String)accept(n, 0, context);
		if (s == null){
		    s = cc.top;
		}
		cc.emit(s3 + "[" + i + "] = (Class)" + s + ";\n");
	    }
	}

	cc.emit("if (" + s1 + " == null){\n");

	if (_includeLineNo && cc.getLine() != node.beginLine){
	    cc.emit( "setLine(context, " + node.beginLine + ");\n");
	    cc.setLine(node.beginLine);
	}
	cc.emit( "PnutsException pe = new PnutsException(new NullPointerException(), context);\n");
	cc.emit( "throw pe;\n");
	cc.emit("}\n");
	if (types_created){
	    cc.emit(cc.top + "= Runtime.callMethod(context, " + s1 + ".getClass(), \"" + node.str + "\", " + s2 + ", " + s3 + ", " + s1 + ");\n");
	} else {
	    cc.emit(cc.top + "= Runtime.callMethod(context, " + s1 + ".getClass(), \"" + node.str + "\", " + s2 + ", null, " + s1 + ");\n");
	}
	return null;
    }

    public Object arrayType(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;

	SimpleNode n = node;
	int count = 0;
	while (n != null && n.id == JJTARRAYTYPE){
	    count++;
	    n = n.jjtGetChild(0);
	}
	if (n.id == JJTINDEXNODE){
	    Object[] idx = parseIndex(n);
	    String s = (String)((SimpleNode)idx[0]).accept(this, context);
	    if (s != null){
		cc.emit(cc.top + " = " + s + ";\n");
	    }
	    cc.emit("if (!(" + cc.top + " instanceof Class)){\n");
	    cc.emit( "throw new PnutsException(\"classOrArray.expected\", NO_PARAM, context);\n");
 	    cc.emit("}\n");
	    cc.emit(cc.top + " = Runtime.arrayType((Class)" + cc.top + ", " + count + ");\n");

	    Object[] dim = (Object[])idx[1];
	    String dimstr[] = new String[dim.length];
	    for (int i = 0; i < dim.length; i++){
		s = (String)((SimpleNode)dim[i]).accept(this, context);
		if (s == null){
		    s = cc.top;
		}
		cc.emit("if (!(" + s + " instanceof Number)){\n");
		cc.emit("throw new PnutsException(\"number.expected\", NO_PARAM, context);\n");
		cc.emit("}\n");
		String s2 = gensym(context);
		cc.emit("int " + s2 + " = ((Number)" + s + ").intValue();\n");
		dimstr[i] = s2;
	    }

	    cc.emit(cc.top + " = Array.newInstance((Class)" + cc.top + ", new int[]{\n");
	    cc.emit(dimstr[0]);
	    for (int i = 1; i < dim.length; i++){
		cc.emit(", " + dimstr[i]);
	    }
	    cc.emit("});\n");
	} else {
	    String s = (String)n.accept(this, context);
	    if (s != null){
		cc.emit(cc.top + " = " + s + ";\n");
	    }
	    cc.emit("if (" + cc.top + " instanceof Class){\n");
	    cc.emit( cc.top + " = Runtime.arrayType((Class)" + cc.top + ", " + count + ");\n");
	    cc.emit("} else {\n");
	    cc.emit( "throw new PnutsException(\"classOrArray.expected\", NO_PARAM, context);\n");
	    cc.emit("}\n");
	}
	return null;
    }

    /**
     * @return [base_component_node, [idx_node_0, idx_node_1, ...]]
     */
    static Object[] parseIndex(SimpleNode node){
	SimpleNode c0 = node.jjtGetChild(0);
	SimpleNode c1 = node.jjtGetChild(1);
	if (c0.id != JJTINDEXNODE){
	    return new Object[]{c0, new Object[]{c1}};
	} else {
	    Object[] r = parseIndex(c0);
	    Object[] d = (Object[])r[1];
	    Object[] a = new Object[d.length + 1];
	    System.arraycopy(d, 0, a, 0, d.length);
	    a[d.length] = c1;
	    return new Object[]{r[0], a};
	}
    }

    public Object indexNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	Object[] idx = parseIndex(node);
	String s = (String)((SimpleNode)idx[0]).accept(this, context);
	String s1 = gensym(context);
	if (s == null){
	    s = cc.top;
	}
	cc.emit("Object " + s1 + " = " + s + ";\n");

	Object[] dim = (Object[])idx[1];

	cc.emit("if (" + s1 + " instanceof Class){\n");

	String dimstr[] = new String[dim.length];

	for (int i = 0; i < dim.length; i++){
	    s = (String)((SimpleNode)dim[i]).accept(this, context);
	    if (s == null){
		s = cc.top;
	    }
	    cc.emit("if (!(" + s + " instanceof Number)){\n");
	    cc.emit("throw new PnutsException(\"number.expected\", NO_PARAM, context);\n");
	    cc.emit("}\n");
	    String s2 = gensym(context);
	    cc.emit("int " + s2 + " = ((Number)" + s + ").intValue();\n");
	    dimstr[i] = s2;
	}
	if (dim.length > 1){
	    cc.emit(cc.top + " = Array.newInstance((Class)" + s1 + ", new int[]{");
	    cc.emit(dimstr[0]);
	    for (int i = 1; i < dim.length; i++){
		cc.emit(", " + dimstr[i]);
	    }
	    cc.emit("});\n");
	} else {
	    cc.emit(cc.top + " = Array.newInstance((Class)" + s1 + ", " + dimstr[0] + ");\n");
	}

	cc.emit("} else {\n");

	for (int i = 0; i < dim.length; i++){
	    s = (String)((SimpleNode)dim[i]).accept(this, context);
	    if (s == null){
		cc.emit(s + " = " + cc.top + ";\n");
	    }
	    cc.emit("if (!(" + s + " instanceof Number)){\n");
	    cc.emit("throw new PnutsException(\"number.expected\", NO_PARAM, context);\n");
	    cc.emit("}\n");
	    cc.emit(s1 + " = getIndex(" + s1 + ", ((Number)" + s + ").intValue(), context);\n");
	}
	cc.emit(cc.top + " = " + s1 + ";\n");
	cc.emit("}\n");

	return null;
    }

    public Object instanceofExpression(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;

	String s1 = gensym(context);
	String s = (String)accept(node, 0, context);
	if (s == null){
	    s = cc.top;
	}
	cc.emit("Object " + s1 + " = " + s + ";\n");

	s = (String)accept(node, 1, context);
	if (s != null){
	    cc.emit(cc.top + " = " + s + ";\n");
	}
	cc.emit("if (" + cc.top + " instanceof Class){\n");
	cc.emit( cc.top + " = " + "((Class)" + cc.top + ").isInstance(" + s1 + ") ? Boolean.TRUE : Boolean.FALSE;\n");
	cc.emit("} else if (" + cc.top + " == null){\n");
	cc.emit( cc.top + " = Boolean.FALSE;\n");
	cc.emit("} else {\n");
	cc.emit( "throw new PnutsException(\"class.expected\", NO_PARAM, context);\n");
	cc.emit("}\n");
	return null;
    }

    public Object assignment(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENT);
    }

    public Object assignmentTA(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENTTA);
    }

    public Object assignmentMA(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENTMA);
    }

    public Object assignmentDA(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENTDA);
    }

    public Object assignmentPA(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENTPA);
    }

    public Object assignmentSA(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENTSA);
    }

    public Object assignmentLA(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENTLA);
    }

    public Object assignmentRA(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENTRA);
    }

    public Object assignmentRAA(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENTRAA);
    }

    public Object assignmentAA(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENTAA);
    }

    public Object assignmentEA(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENTEA);
    }

    public Object assignmentOA(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENTOA);
    }

    public Object preIncrNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	String assoc = (String)cc.constants.get(ONE);
	if (assoc == null){
	    assoc = gensym(context);
	    cc.constants.put(ONE, assoc);
	}
	_assign(node.jjtGetChild(0), context, JJTASSIGNMENTPA, assoc, true);
	return null;
    }

    public Object preDecrNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	String assoc = (String)cc.constants.get(ONE);
	if (assoc == null){
	    assoc = gensym(context);
	    cc.constants.put(ONE, assoc);
	}
	_assign(node.jjtGetChild(0), context, JJTASSIGNMENTSA, assoc, true);
	return null;
    }

    public Object postIncrNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	String s1 = gensym(context);
	String s = (String)accept(node, 0, context);
	if (s == null){
	    s = cc.top;
	}
	cc.emit("Object " + s1 + " = " + s + ";\n");

	String assoc = (String)cc.constants.get(ONE);
	if (assoc == null){
	    assoc = gensym(context);
	    cc.constants.put(ONE, assoc);
	}
	_assign(node.jjtGetChild(0), context, JJTASSIGNMENTPA, assoc, true);
	cc.emit(cc.top + " = " + s1 + ";\n");
	return null;
    }

    public Object postDecrNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	String s1 = gensym(context);
	String s = (String)accept(node, 0, context);
	if (s == null){
	    s = cc.top;
	}
	cc.emit("Object " + s1 + " = " + s + ";\n");

	String assoc = (String)cc.constants.get(ONE);
	if (assoc == null){
	    assoc = gensym(context);
	    cc.constants.put(ONE, assoc);
	}
	_assign(node.jjtGetChild(0), context, JJTASSIGNMENTSA, assoc, true);
	cc.emit(cc.top + " = " + s1 + ";\n");
	return null;
    }

    private Object assign(SimpleNode node, Context context, int id){
	CompileContext cc = (CompileContext)context;
	String s = (String)accept(node, 1, context);

	if (s != null){
 	    _assign(node.jjtGetChild(0), context, id, s, false);
	} else {
 	    _assign(node.jjtGetChild(0), context, id, cc.top, false);
	}
	return null;
    }

    private Object _assign(SimpleNode lhs, Context context, int id, String rhs){
	return _assign(lhs, context, id, rhs, false);
    }

    private Object _assign(SimpleNode lhs, Context context, int id, String rhs, boolean mutable){
	CompileContext cc = (CompileContext)context;
	if (lhs.id == JJTIDNODE){
	    if (cc.env.parent != null && !mutable){
		if (id == JJTASSIGNMENT){
		    Reference ref = cc.getReference(lhs.str);
		    if (ref != null){
			if (ref.offset >= 0){
			    cc.emit(ref.symbol + "[" + ref.offset + "] = " + rhs + ";\n");
			} else {
			    cc.declare(lhs.str);
			    ref = cc.getReference(lhs.str);
			    cc.emit(ref.symbol + "[0] = " + rhs + ";\n");
			}
		    } else {
			cc.declare(lhs.str);
			ref = cc.getReference(lhs.str);
			cc.emit(ref.symbol + "[0] = " + rhs + ";\n");
		    }
		    cc.emit(cc.top + " = " + rhs + ";\n");
		} else {
		    Reference ref = cc.getReference(lhs.str);
		    if (ref != null){
			if (ref.offset >= 0){
			    String s2 = gensym(context);
			    cc.emit("Object " + s2 + " = " + ref.symbol + "[" + ref.offset + "];\n");
			    cc.emit(cc.top + " = " + ref.symbol + "[" + ref.offset + "] = " + compute(s2, rhs, id) + ";\n");
			} else {
			    throw new RuntimeException("never happen");
			}
		    } else {
			throw new RuntimeException("never happen");
		    }
		}
	    } else {
		Reference ref = cc.getReference(lhs.str);
		if (ref != null){

		    if (id != JJTASSIGNMENT){
			String s2 = gensym(context);
			if (ref.offset < 0){
			    cc.emit("Object " + s2 + " = " + ref.symbol + ";\n");
			} else {
			    cc.emit("Object " + s2 + " = " + ref.symbol + "[" + ref.offset + "];\n");
			}
			rhs = compute(s2, rhs, id);
		    }
		    if (ref.offset < 0){
			cc.emit(cc.top + " = " + ref.symbol + " = " + rhs + ";\n");
		    } else {
			cc.emit(cc.top + " = " + ref.symbol + "[" + ref.offset + "] = " + rhs + ";\n");
		    }
		} else {
		    if (id != JJTASSIGNMENT){
			String s2 = gensym(context);
			cc.emit("Object " + s2 + " = context.getCurrentPackage().get(\"" + lhs.str + "\");\n");
			cc.emit(cc.top + " = " + compute(s2, rhs, id) + ";\n");
		    } else if (!rhs.equals(cc.top)){
			cc.emit(cc.top + " = " + rhs + ";\n");
		    }
		    cc.emit("context.getCurrentPackage().set(\"" + lhs.str + "\", " + cc.top + ", context);\n");
		}
	    }
	} else if (lhs.id == JJTGLOBAL){
	    if (id != JJTASSIGNMENT){
		String s2 = gensym(context);
		cc.emit("Object " + s2 + " = Package.getGlobalPackage().get(\"" + lhs.str + "\", context);\n");
		cc.emit(cc.top + " = " + compute(s2, rhs, id) + ";\n");
	    } else if (!rhs.equals(cc.top)){
		cc.emit(cc.top + " = " + rhs + ";\n");
	    }
	    cc.emit("Package.getGlobalPackage().set(\"" + lhs.str + "\", " + rhs + ", context);\n");
	} else if (lhs.id == JJTINDEXNODE){
	    String s2 = gensym(context);
	    String s1 = gensym(context);
	    cc.emit("Object " + s1 + " = " + rhs + ";\n");
	    String s = (String)accept(lhs, 0, context);
	    if (s == null){
		s = cc.top;
	    }
	    cc.emit("Object " + s2 + " = " + s + ";\n");

	    String s3 = gensym(context);
	    s = (String)accept(lhs, 1, context);
	    if (s == null){
		s = cc.top;
	    }
	    cc.emit("Object " + s3 + " = " + s + ";\n");

	    cc.emit("if (!(" + s3 + " instanceof Number)){\n");
	    cc.emit( "throw new PnutsException(\"number.expected\", NO_PARAM, context);\n");
	    cc.emit("}\n");
	    String s4 = gensym(context);
	    cc.emit("int " + s4 + " = ((Number)" + s3 + ").intValue();\n");

	    cc.emit("if (" + s2 + " instanceof Indexed){\n");

	    if (id == JJTASSIGNMENT){
		cc.emit("((Indexed)" + s2 + ").set(" + s4 + ", " + s1 + ");\n");
	        cc.emit(cc.top + " = " + s1 + ";\n");
	    } else {
		String s6 = gensym(context);
		cc.emit("Object " + s6 + " = ((Indexed)" + s2 + ").get(" + s4 + ");\n");
		cc.emit(cc.top + " = " + compute(s6, s1, id) + ";\n");
		cc.emit("((Indexed)" + s2 + ").set(" + s4 + ", " + cc.top + ");\n");
	    }

	    cc.emit("} else if (" + s2 + " instanceof String){\n");

	    if (id != JJTASSIGNMENT){
		cc.emit("throw new PnutsException(\"illegal.assign\", NO_PARAM, context);\n");
	    } else {
		cc.emit("if (!(" + s1 + " instanceof Character)){\n");
		cc.emit("throw new PnutsException(\"illegal.assign\", NO_PARAM, context);\n");
		cc.emit("}\n");

	        cc.emit(cc.top + " = assignRange(" + s2 + ", (Number)" + s3 + ", (Number)" + s3 + ", " + s1 + ");\n");
	        SimpleNode ch = lhs.jjtGetChild(0);
		if (ch.id == JJTIDNODE ||
		    ch.id == JJTGLOBAL ||
		    ch.id == JJTINDEXNODE ||
		    ch.id == JJTMEMBERNODE ||
		    ch.id == JJTSTATICMEMBERNODE){
		    _assign(ch, context, id, cc.top, true);
		}
	    }
	    
	    String s5 = gensym(context);
	    if (id == JJTASSIGNMENT){
		cc.emit("} else if (" + s2 + " instanceof Object[]){");
		cc.emit("((Object[])" + s2 + ")[" + s4 + "] = " + s1 + ";\n");
		cc.emit(cc.top + " = " + s1 + ";\n");
		cc.emit("} else if (" + s2 + " instanceof int[]){");
		cc.emit("((int[])" + s2 + ")[" + s4 + "] = ((Number)" + s1 + ").intValue();\n");
		cc.emit(cc.top + " = " + s1 + ";\n");
		cc.emit("} else if (Runtime.isArray(" + s2 + ")){\n");
		cc.emit("Array.set(" + s2 + ", " + s4 + ", " + s1 + ");\n");
		cc.emit(cc.top + " = " + s1 + ";\n");
		cc.emit("} else {\n");
		cc.emit( "throw new PnutsException(\"illegal.assign\", NO_PARAM, context);\n");
		cc.emit("}\n");
	    } else {
		cc.emit("} else if (Runtime.isArray(" + s2 + ")){\n");
		cc.emit("Object " + s5 + " = Array.get(" + s2 + ", " + s4 + ");\n");
		cc.emit(cc.top + " = " + compute(s5, s1, id) + ";\n");
		cc.emit("Array.set(" + s2 + ", " + s4 + ", " + cc.top + ");\n");
		cc.emit("} else {\n");
		cc.emit( "throw new PnutsException(\"illegal.assign\", NO_PARAM, context);\n");
		cc.emit("}\n");
	    }
	} else if (lhs.id == JJTSTATICMEMBERNODE){
	    SimpleNode c1 = lhs.jjtGetChild(0);
	    String pkgName = getPackageName(c1);
	    String s2 = gensym(context);
	    cc.emit("Package " + s2 + " = Package.find(\"" + pkgName + "\");\n");
	    cc.emit("if (" + s2 + " != null){\n");
	    cc.emit( s2 + ".set(\"" + lhs.str + "\", " + rhs + ", context);\n");
	    cc.emit( cc.top + " = " + rhs + ";\n");
	    cc.emit("} else {\n");

	    String s5 = gensym(context);
	    cc.emit("Object " + s5 + " = " + rhs + ";\n");

	    String s3 = gensym(context);
	    String s = (String)c1.accept(this, context);
	    if (s == null){
		s = cc.top;
	    }
	    cc.emit("Object " + s3 + " = " + s + ";\n");

	    cc.emit("if (" + s3 + " instanceof Class){\n");
	    if (id != JJTASSIGNMENT){
		String s4 = gensym(context);
		cc.emit("Object " + s4 + " = Runtime.getStaticField(context, (Class)" + s3 + ", \"" + lhs.str + "\");\n");
		cc.emit(s5 + " = " + compute(s4, s5, id) + ";\n");
	    }
	    cc.emit("Runtime.putStaticField(context, (Class)" + s3 + ", \"" + lhs.str + "\", " + s5 + ");\n");
	    cc.emit(cc.top + " = " + s5 + ";\n");
	    cc.emit("} else {\n");
	    cc.emit( "throw new PnutsException(\"package.notFound\", NO_PARAM, context);\n");
	    cc.emit("}\n");
	    cc.emit("}\n");
	} else if (lhs.id == JJTMEMBERNODE){
	    String s1 = gensym(context);
	    cc.emit("Object " + s1 + " = " + rhs + ";\n");

	    String s2 = gensym(context);
	    String s = (String)accept(lhs, 0, context);
	    if (s == null){
		s = cc.top;
	    }
	    cc.emit("Object " + s2 + " = " + s + ";\n");

	    if (id != JJTASSIGNMENT){
		String s3 = gensym(context);
		cc.emit("Object " + s3 + " = Runtime.getField(" + s2 + ", \"" + lhs.str + "\");\n");
		cc.emit(s1 + " = " + compute(s3, s1, id) + ";\n");
	    }
	    cc.emit("Runtime.putField(context, " + s2 + ", \"" + lhs.str + "\", " + s1 + ");\n");
	    cc.emit(cc.top + " = " + s1 + ";\n");
	} else if (lhs.id == JJTRANGENODE){
	    String s1 = gensym(context);
	    cc.emit("Object " + s1 + " = " + rhs + ";\n");

	    String s2 = gensym(context);
	    String s = (String)accept(lhs, 0, context);
	    if (s == null){
		s = cc.top;
	    }
	    cc.emit("Object " + s2 + " = " + s + ";\n");

	    s = (String)accept(lhs, 1, context);
	    String s3 = gensym(context);
	    if (s == null){
		s = cc.top;
	    }
	    cc.emit("Object " + s3 + " = " + s + ";\n");

	    String s4 = gensym(context);
	    if (lhs.jjtGetNumChildren() >= 3){
		s = (String)accept(lhs, 2, context);
		if (s == null){
		    s = cc.top;
		}
		cc.emit("Object " + s4 + " = " + s + ";\n");

	    } else {
		cc.emit("Object " + s4 + " = null;\n");
	    }
	    cc.emit("if (!(" + s3 + " instanceof Number) || (" + s4 + " != null && !(" + s4 + " instanceof Number))){\n");
	    cc.emit("throw new PnutsException(\"number.expected\", NO_PARAM, context);\n");
	    cc.emit("}\n");
	    cc.emit(cc.top + " = assignRange(" + s2 + ", (Number)" + s3 + ", (Number)" + s4 + ", " + s1 + ");\n");
	    SimpleNode ch = lhs.jjtGetChild(0);
	    if (ch.id == JJTIDNODE ||
		ch.id == JJTGLOBAL ||
		ch.id == JJTINDEXNODE ||
		ch.id == JJTMEMBERNODE ||
		ch.id == JJTSTATICMEMBERNODE){
		cc.emit("if (" + s2 + " instanceof String){\n");
		_assign(ch, context, id, cc.top, true);
		cc.emit("}\n");
	    }
	} else {
	    throw new RuntimeException("never happen");
	}
	return null;
    }

    String compute(String s1, String s2, int id){
	switch (id){
	case JJTASSIGNMENTTA:
	    return "Arithmetic.multiply(" + s1 + ", " + s2 + ")";
	case JJTASSIGNMENTMA:
	    return "Arithmetic.mod((Number)" + s1 + ", (Number)" + s2 + ")";
	case JJTASSIGNMENTDA:
	    return "Arithmetic.divide(" + s1 + ", " + s2 + ")";
	case JJTASSIGNMENTPA:
	    return "Arithmetic.add(" + s1 + ", " + s2 + ")";
	case JJTASSIGNMENTSA:
	    return "Arithmetic.subtract(" + s1 + ", " + s2 + ")";
	case JJTASSIGNMENTLA:
	    return "Arithmetic.shiftLeft((Number)" + s1 + ", (Number)" + s2 + ")";
	case JJTASSIGNMENTRA:
	    return "Arithmetic.shiftRight((Number)" + s1 + ", (Number)" + s2 + ")";
	case JJTASSIGNMENTRAA:
	    return "Arithmetic.shiftArithmetic((Number)" + s1 + ", (Number)" + s2 + ")";
	case JJTASSIGNMENTAA:
	    return "Arithmetic.and((Number)" + s1 + ", (Number)" + s2 + ")";
	case JJTASSIGNMENTEA:
	    return "Arithmetic.xor((Number)" + s1 + ", (Number)" + s2 + ")";
	case JJTASSIGNMENTOA:
	    return "Arithmetic.or((Number)" + s1 + ", (Number)" + s2 + ")";
	default:
	    throw new RuntimeException("never happen");
	}
    }

    public Object staticMethodNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	if (_includeLineNo && cc.getLine() != node.beginLine){
	    cc.emit("setLine(context, " + node.beginLine + ");\n");
	    cc.setLine(node.beginLine);
	}
	String s1 = gensym(context);
	String s = (String)accept(node, 1, context);
	if (s == null){
	    s = cc.top;
	}
	cc.emit("Object[] " + s1 + " = (Object[])" + s + ";\n");
	String s2 = gensym(context);
//	cc.emit("Class[] " + s2 + " = null;\n");

	SimpleNode c1 = node.jjtGetChild(0);
	String pkgName = getPackageName(c1);

	SimpleNode argNode = node.jjtGetChild(1);
	int nargs = argNode.jjtGetNumChildren();

	String s3 = gensym(context);
	String s4 = gensym(context);
	cc.emit("Package " + s3 + " = Package.find(\"" + pkgName + "\");\n");
	cc.emit("if (" + s3 + " != null){\n");
	cc.emit( "Object " + s4 + " = " + s3 + ".get(\"" + node.str + "\", context);\n");

	cc.emit("if (" + s4 + " instanceof PnutsFunction){\n");
	cc.emit( cc.top + " = callFunction(context, (PnutsFunction)" + s4 + ", " + s1 + ");\n");
	cc.emit("} else if (" + s4 + " instanceof Class){\n");

	boolean types_created = false;
	for (int i = 0; i < nargs; i++){
	    SimpleNode n = argNode.jjtGetChild(i);
	    if (n.id == JJTCASTEXPRESSION){
		if (!types_created){
		    cc.emit("Class[]" + s2 + " = new Class[" + nargs + "];\n");
		    types_created = true;
		}
		s = (String)accept(n, 0, context);
		if (s == null){
		    s = cc.top;
		}
		cc.emit(s2 + "[" + i + "] = (Class)" + s + ";\n");
	    }
	}
	if (types_created){
	    cc.emit( cc.top + " = newInstance(context, (Class)" + s4 + ", " + s1 + ", " + s2 + ");\n");
	} else {
	    cc.emit( cc.top + " = newInstance(context, (Class)" + s4 + ", " + s1 + ", null);\n");
	}
	cc.emit("}\n");

	cc.emit("} else {\n");

	String s5 = gensym(context);
	s = (String)accept(node, 0, context);
	if (s == null){
	    s = cc.top;
	}
	cc.emit("Object " + s5 + " = " + s + ";\n");

	cc.emit("if ("+ s5 + " instanceof Class){\n");

	types_created = false;
	for (int i = 0; i < nargs; i++){
	    SimpleNode n = argNode.jjtGetChild(i);
	    if (n.id == JJTCASTEXPRESSION){
		if (!types_created){
		    cc.emit("Class[] " + s2 + " = new Class[" + nargs + "];\n");
		    types_created = true;
		}
		s = (String)accept(n, 0, context);
		if (s == null){
		    s = cc.top;
		}
		cc.emit(s2 + "[" + i + "] = (Class)" + s + ";\n");
	    }
	}

	if (types_created){
	    cc.emit( cc.top + " = Runtime.callMethod(context, (Class)" + s5 + ", \"" + node.str + "\", " + s1 + ", " + s2 + ", null);\n");
	} else {
	    cc.emit( cc.top + " = Runtime.callMethod(context, (Class)" + s5 + ", \"" + node.str + "\", " + s1 + ", null, null);\n");
	}
	cc.emit("} else {\n");
	cc.emit( "throw new PnutsException(\"illegal.staticCall\", new Object[]{\"" + pkgName + "\", \"" + node.str + "\" , new Integer(" + nargs + ")}, context);\n");
	cc.emit("}\n");

	cc.emit("}\n");
	return null;
    }

    public Object staticMemberNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	if (_includeLineNo && cc.getLine() != node.beginLine){
	    cc.emit("setLine(context, " + node.beginLine + ");\n");
	    cc.setLine(node.beginLine);
	}
	SimpleNode c1 = node.jjtGetChild(0);
	String pkgName = getPackageName(c1);
	String s1 = gensym(context);
	cc.emit("Package " + s1 + " = Package.find(\"" + pkgName + "\");\n");
	cc.emit("if (" + s1 + " != null){\n");
	cc.emit( "if (" + s1 + ".defined(\"" + node.str + "\", context)){\n");
	cc.emit(  cc.top + " = " + s1 + ".get(\"" + node.str + "\", context);\n");
	cc.emit( "} else {\n");
	cc.emit(  "throw new PnutsException(\"not.defined\", new Object[]{\"" + node.str + "\"}, context);\n");
	cc.emit( "}\n");
	cc.emit("} else {\n");
	String s2 = gensym(context);
	String s = (String)c1.accept(this, context);
	if (s == null){
	    s = cc.top;
	}
	cc.emit("Object " + s2 + " = " + s + ";\n");

	cc.emit("if (" + s2 + " instanceof Class){\n");
	cc.emit( cc.top + " = Runtime.getStaticField(context, (Class)" + s2 + ",\"" + node.str + "\");\n");
	cc.emit("} else {\n");
	cc.emit( "throw new PnutsException(\"packageOrClass.expected\", NO_PARAM, context);\n");
	cc.emit("}\n");
	cc.emit("}\n");
	return null;
    }

    public Object rangeNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	String s1 = gensym(context);
	String s2 = gensym(context);
	String s3 = gensym(context);
	String s = (String)accept(node, 0, context);
	if (s == null){
	    s = cc.top;
	}
	cc.emit("Object " + s1 + " = " + s + ";\n");

	s = (String)accept(node, 1, context);
	if (s == null){
	    s = cc.top;
	}
	cc.emit("Object " + s2 + " = " + s + ";\n");

	cc.emit("Object " + s3 + " = null;\n");
	if (node.jjtGetNumChildren() >= 3){
	    s = (String)accept(node, 2, context);
	    if (s == null){
		s = cc.top;
	    }
	    cc.emit(s3 + " = " + s + ";\n");
	}
	cc.emit("if (" + s2 + " instanceof Number && (" + s3 + " == null || " + s3 + " instanceof Number)){\n");
	String s4 = gensym(context);
	String s5 = gensym(context);
	cc.emit("int " + s4 + " = ((Number)" + s2 + ").intValue();\n");
	cc.emit("int " + s5 + " = 0;\n");
	cc.emit("if (" + s3 + " != null){\n");
	cc.emit(s5 + " = ((Number)" + s3 + ").intValue();\n");
	cc.emit("if (" + s5 + " < 0){\n");
	cc.emit("throw new ArrayIndexOutOfBoundsException(String.valueOf(" + s5 + "));\n");
	cc.emit("}\n");
	cc.emit("}\n");
	cc.emit("if (" + s1 + " instanceof String){\n");
	String s6 = gensym(context);
	cc.emit("String " + s6 + " = (String)" + s1 + ";\n");
	cc.emit("if (" + s3 + " != null){\n");
	cc.emit(cc.top + " = " + s6 + ".substring(" + s4 + ", " + s5 + " + 1);\n");
	cc.emit("} else {\n");
	cc.emit(cc.top + " = " + s6 + ".substring(" + s4 + ");\n");
	cc.emit("}\n");
	cc.emit("} else {\n");
	String s7 = gensym(context);
	cc.emit("int " + s7 + " = Array.getLength(" + s1 + ");\n");
	cc.emit("if (" + s3 + " == null){\n");
	cc.emit(s5 + " = " + s7 + " - 1;\n");
	cc.emit("} else if (" + s5 + " > " + s7 + " - 1){\n");
	cc.emit("throw new ArrayIndexOutOfBoundsException(String.valueOf(" + s5 + "));\n");
	cc.emit("}\n");
	String s8 = gensym(context);
	cc.emit("int " + s8 + " = " + s5 + " - " + s4 + " + 1;\n");
	cc.emit("if (" + s4 + " + " + s8 + " > " + s7 + "){\n");
	cc.emit(s8 + " = " + s7 + " - " + s4 + ";\n");
	cc.emit("}\n");
	String s9 = gensym(context);
	cc.emit("Class " + s9 + " = " + s1 + ".getClass().getComponentType();\n");
	String s10 = gensym(context);
	cc.emit("Object " + s10 + " = Array.newInstance(" + s9 + ", " + s8 + ");\n");
	cc.emit("if (" + s8 + " > 0){\n");
	cc.emit("System.arraycopy(" + s1 + ", " + s4 + ", " + s10 + ", 0, " + s8 + ");\n");
	cc.emit("}\n");
	cc.emit(cc.top + " = " + s10 + ";\n");
	cc.emit("}\n");
	cc.emit("} else {\n");
	cc.emit("throw new PnutsException(\"number.expected\", NO_PARAM, context);\n");
	cc.emit("}\n");
	return null;
    }

    public Object forStatement(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	SimpleNode initNode = null;
	SimpleNode condNode = null;
	SimpleNode updateNode = null;
	SimpleNode blockNode = null;

	int j = 0;
	SimpleNode n = node.jjtGetChild(j);
	if (n.id == JJTFORINIT){
	    initNode = n;
	    j++;
	}
	n = node.jjtGetChild(j);
	if (n.id != JJTFORUPDATE && n.id != JJTBLOCK){
	    condNode = n;
	    j++;
	}
	n = node.jjtGetChild(j);
	if (n.id == JJTFORUPDATE){
	    updateNode = n;
	    j++;
	}

	blockNode = node.jjtGetChild(j);

	String s1 = gensym(context);
	cc.emit("Object " + s1 + " = null;\n");
	if (initNode != null){
	    int num = initNode.jjtGetNumChildren();
	    String[] env = new String[num];
	    for (int i = 0; i < env.length; i++){
		SimpleNode sn = initNode.jjtGetChild(i);
		env[i] = sn.str;
	    }
	    cc.openScope(env);
	    for (int i = 0; i < env.length; i++){
		SimpleNode sn = initNode.jjtGetChild(i);
		String s = (String)accept(sn, 0, context);
		Reference ref = cc.getReference(env[i]);
		if (s == null){
		    s = cc.top;
		}
		cc.emit("Object[] " + ref.symbol + " = new Object[]{" + s + "};\n");
	    }
	} else {
	    cc.openScope(new String[]{});
	}

	String s2 = gensym(context);
	cc.emit("Object " + s2 + " = null;\n");
	String s = null;

	if (condNode != null){
	    s = (String)condNode.accept(this, context);
	    if (s == null){
		cc.emit(s2 + " = " + cc.top + ";\n");
	    } else {
		cc.emit(s2 + " = " + s + ";\n");
	    }

	    int id = condNode.id;
	    if (id != JJTTRUENODE &&
		id != JJTFALSENODE &&
		id != JJTLENODE &&
		id != JJTLTNODE &&
		id != JJTGENODE &&
		id != JJTGTNODE &&
		id != JJTEQUALNODE &&
		id != JJTNOTEQNODE &&
		id != JJTINSTANCEOFEXPRESSION &&
		id != JJTORNODE &&
		id != JJTANDNODE &&
		id != JJTNOTNODE){
		cc.emit("if (!(" + s2 + " instanceof Boolean)){\n");
		cc.emit( "throw new PnutsException(\"boolean.expected\", NO_PARAM, context);\n");
		cc.emit("}\n");
	    }
	} else {
	    cc.emit(s2 + " = Boolean.TRUE;\n");
	}

	cc.emit("try {\n");
	cc.emit( "while (true){\n");

	cc.emit( "if (!((Boolean)" + s2 + ").booleanValue()){\n");

	cc.emit( cc.top + " = null;\n");
	cc.emit( "break;\n");
	cc.emit( "}\n");
	s = (String)blockNode.accept(this, context);
	if (s != null){
	    cc.emit(cc.top + " = " + s + ";\n");
	}

	if (updateNode != null){
	    for (int i = 0; i < updateNode.jjtGetNumChildren(); i++){
		accept(updateNode, i, context);
	    }
	}
	if (condNode != null){
	    s = (String)condNode.accept(this, context);
	    if (s != null){
		cc.emit(s2 + " = " + s + ";\n");
	    } else {
		cc.emit(s2 + " = " + cc.top + ";\n");
	    }

	    int id = condNode.id;
	    if (id != JJTTRUENODE &&
		id != JJTFALSENODE &&
		id != JJTLENODE &&
		id != JJTLTNODE &&
		id != JJTGENODE &&
		id != JJTGTNODE &&
		id != JJTEQUALNODE &&
		id != JJTNOTEQNODE &&
		id != JJTINSTANCEOFEXPRESSION &&
		id != JJTORNODE &&
		id != JJTANDNODE &&
		id != JJTNOTNODE){
		cc.emit("if (!(" + s2 + " instanceof Boolean)){\n");
		cc.emit( "throw new PnutsException(\"boolean.expected\", NO_PARAM, context);\n");
		cc.emit("}\n");
	    }
	}

	cc.emit("}\n");
	cc.emit("} catch (ClassCastException e){\n");
	cc.emit( "throw new PnutsException(\"boolean.expected\", NO_PARAM, context);\n");
	cc.emit("}\n");
	cc.closeScope();
	return null;
    }

    public Object breakNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	if (node.jjtGetNumChildren() > 0){
	    String s = (String)accept(node, 0, context);
	    cc.emit(cc.top + " = " + s + ";\n");
	} else {
	    cc.emit(cc.top + " = null;\n");
	}
	SimpleNode parent = node.jjtGetParent();
	while (parent != null){
	    if (parent.id == JJTFORSTATEMENT ||
		parent.id == JJTWHILESTATEMENT ||
		parent.id == JJTFOREACHSTATEMENT ||
		parent.id == JJTSWITCHSTATEMENT){
		cc.emit("if (true) break;\n");
		break;
	    }
	    parent = parent.jjtGetParent();
	}
	return null;
    }

    public Object continueNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	SimpleNode parent = node.jjtGetParent();
	while (parent != null){
	    if (parent.id == JJTFORSTATEMENT ||
		parent.id == JJTWHILESTATEMENT ||
		parent.id == JJTFOREACHSTATEMENT){
		cc.emit("if (true) continue;\n");
		break;
	    }
	    parent = parent.jjtGetParent();
	}
	return null;
    }

    public Object returnNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	String s1 = gensym(context);
	SimpleNode parent = node.jjtGetParent();
	if (node.jjtGetNumChildren() > 0){
	    String s = (String)accept(node, 0, context);
	    if (s != null){
		cc.emit(cc.top + " = " + s + ";\n");
	    }
	}
	if (cc.env.parent != null){
	    cc.emit("if (true) return " + cc.top + ";\n");
	    return null;
	}
	String s2 = gensym(context);
	cc.emit("doReturn(" + cc.top + ");\n");
	return null;
    }

    public Object whileStatement(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	if (_includeLineNo && cc.getLine() != node.beginLine){
	    cc.emit("setLine(context, " + node.beginLine + ");\n");
	    cc.setLine(node.beginLine);
	}
	String s1 = gensym(context);
	cc.openScope(new String[]{});

	cc.emit("while (true){\n");
	String s = (String)accept(node, 0, context);

	int id = node.jjtGetChild(0).id;
	if (id != JJTTRUENODE &&
	    id != JJTFALSENODE &&
	    id != JJTLENODE &&
	    id != JJTLTNODE &&
	    id != JJTGENODE &&
	    id != JJTGTNODE &&
	    id != JJTEQUALNODE &&
	    id != JJTNOTEQNODE &&
	    id != JJTINSTANCEOFEXPRESSION &&
	    id != JJTORNODE &&
	    id != JJTANDNODE &&
	    id != JJTNOTNODE){
	    if (s != null){
	        cc.emit(cc.top + " = " + s + ";\n");
	    }
	    cc.emit("if (!(" + cc.top + " instanceof Boolean)){\n");
	    cc.emit( "throw new PnutsException(\"boolean.expected\", NO_PARAM, context);\n");
	    cc.emit("}\n");
	} else {
	    if (s == null){
		s = cc.top;
	    }
	}
	cc.emit( "if (!((Boolean)" + s + ").booleanValue()){\n");

	cc.emit(  cc.top + " = null;\n");
	cc.emit(  "break;\n");
	cc.emit( "}\n");

	s = (String)accept(node, 1, context);
	if (s != null){
	    cc.emit(cc.top + " = " + s + ";\n");
	}

	cc.emit("}\n");
	cc.closeScope();
	return null;
    }

    public Object foreachStatement(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	String s1 = gensym(context);
	String s2 = gensym(context);
	cc.openScope(new String[]{node.str});
	Reference count = cc.getReference(node.str);

	cc.emit("Object[] " + count.symbol + " = new Object[1];\n");
	String s = (String)accept(node, 0, context);
	if (s != null){
	    cc.emit(cc.top + " = " + s + ";\n");
	}
	cc.emit("if (" + cc.top + " != null){\n");
	cc.emit( "Object " + s1 + " = " + cc.top + ";\n");
	cc.emit( "if (Runtime.isArray(" + s1 + ")){\n");
	cc.emit(  "for (int " + s2 + " = 0; " + s2 + " < Array.getLength(" + s1 + "); " + s2 + "++){\n");
	cc.emit(   count.symbol + "[0] = Array.get(" + s1 + ", " + s2 + ");\n");
	s = (String)accept(node, 1, context);
	if (s != null){
	    cc.emit(cc.top + " = " + s + ";\n");
	}
	cc.emit(  "}\n");
	cc.emit( "} else if (" + s1 + " instanceof Enumeration){\n");
	cc.emit(  "for (Enumeration " + s2 + " = (Enumeration)" + s1 + "; " + s2 + ".hasMoreElements(); ){\n");
	cc.emit(   count.symbol + "[0] = " + s2 + ".nextElement();\n");
	s = (String)accept(node, 1, context);
	if (s != null){
	    cc.emit(cc.top + " = " + s + ";\n");
	}
	cc.emit(  "}\n");
	if (Pnuts.isJava2()){
	    cc.emit( "} else if (" + s1 + " instanceof Iterator){\n");
	    cc.emit(  "for (Iterator " + s2 + " = (Iterator)" + s1 + "; " + s2 + ".hasNext(); ){\n");
	    cc.emit(   count.symbol + "[0] = " + s2 + ".next();\n");
	    s = (String)accept(node, 1, context);
	    if (s != null){
	        cc.emit(cc.top + " = " + s + ";\n");
	    }
	    cc.emit("}\n");
	}
	cc.emit( "}\n");
	cc.emit("}\n");
	cc.closeScope();
	return null;
    }

    public Object switchStatement(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	int n = node.jjtGetNumChildren();
	String s = (String)accept(node, 0, context);
	String s1 = gensym(context);
	if (s == null){
	    s = cc.top;
	}
	cc.emit("Object " + s1 + " = " + s + ";\n");

	cc.openScope(new String[]{});

	cc._declare("!matched");
	Reference matched = cc.getReference("!matched");
	cc.emit("boolean " + matched.symbol + " = false;\n");

	String s2 = gensym(context);
	cc.emit("boolean " + s2 + " = true;\n");
	cc.emit("while (" + s2 + "){\n");
	cc.emit(s2 + " = false;\n");
	for (int i = 1; i < n; i++){
	    SimpleNode _node = node.jjtGetChild(i);
	    if (_node.jjtGetNumChildren() == 1){ // case
		i++;
		s = (String)accept(_node, 0, context);
		if (s != null){
		    cc.emit(cc.top + " = " + s + ";\n");
		}
		cc.emit("if (" + matched.symbol + " || " +
		    "Arithmetic.compareTo(" + s1 + ", " + cc.top + ") == Numeric.EQUAL){\n");
		s = (String)accept(node, i, context);
		if (s != null){
		    cc.emit(cc.top + " = " + s + ";\n");
		}
		cc.emit(matched.symbol + " = true;\n");
		cc.emit("}\n");
	    } else { // default
		cc.emit(matched.symbol + " = true;\n");
		s = (String)accept(node, ++i, context);
		if (s != null){
		    cc.emit(cc.top + " = " + s + ";\n");
		}
	    }

	}
	cc.emit("break;\n");
	cc.emit("}\n");
	cc.closeScope();
	return null;
    }

    public Object switchBlock(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;

	int n = node.jjtGetNumChildren();
	String s1 = gensym(context);
	for (int i = 0; i < n; i++){
	    cc.emit("try {\n");
	    String s = (String)accept(node, i, context);
	    if (s != null){
		cc.emit(cc.top + " = " + s + ";\n");
	    }
	    cc.emit("} catch (PnutsException " + s1 + "){\n");
	    cc.emit( "if (" + s1 + ".getThrowable() instanceof ThreadDeath){\n");
	    cc.emit(  "break;\n");
	    cc.emit( "}\n");
	    cc.emit( "throw " + s1 + ";\n");
	    cc.emit("}\n");
	}
	return null;
    }

    static String getPackageName(SimpleNode node){
	if (node.jjtGetNumChildren() > 0){
	    SimpleNode c1 = node.jjtGetChild(0);
	    return getPackageName(c1) + "::" + node.str;
	} else {
	    return node.str;
	}
    }

    protected Object accept(SimpleNode node, int idx, Context context){
	return node.jjtGetChild(idx).accept(this, context);
    }
}
