/*
 * @(#)LocaleAdapter.java	1.3 99/06/03
 *
 * Copyright (c) 1997,1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package pnuts.util;

import java.util.ResourceBundle;
import java.util.Enumeration;
import java.util.Locale;
import java.util.Date;
import java.util.MissingResourceException;
import java.text.MessageFormat;
import java.text.NumberFormat;
import java.text.DateFormat;
import pnuts.lang.AbstractData;
import pnuts.lang.Property;
import pnuts.lang.Runtime;
import java.lang.reflect.InvocationTargetException;

/**
 * A locale related utility class for Pnuts
 *
 * @version	1.3
 *
 * <pre>e.g.
 * locale = class pnuts.util.LocaleAdapter
 * LC = locale()
 * LC = locale(Locale::ITALY)
 * txt = LC.label("open") // ResourceBundle::getBundle("label").getObject("open", Locale::ITALY)
 * txt = LC.err("E1", 20) // MessageFormat::format(ResourceBundle::getBundle("err").getString("E1", Locale::ITALY), [20])
 * fmt1 = LC.number(12345)
 * fmt2 = LC.currency(12345)
 * fmt3 = LC.percent(0.12)
 * fmt4 = LC.date("full")
 * fmt4 = LC.date()
 * </pre>
 */
public class LocaleAdapter implements AbstractData {

    static String FULL = "full".intern();
    static String LONG = "long".intern();
    static String SHORT = "short".intern();
    static String MEDIUM = "medium".intern();
    static String DEFAULT = "default".intern();

    private final static String NUMBER = "number".intern();
    private final static String CURRENCY = "currency".intern();
    private final static String PERCENT = "percent".intern();
    private final static String TIME = "time".intern();
    private final static String DATE = "date".intern();
    private final static String DATETIME = "datetime".intern();
    private final static String GETLOCALE = "getLocale".intern();

    static boolean isKeyword(String sym){
	String s = sym.intern();
	return
	    s == NUMBER || s == CURRENCY || s == PERCENT ||
	    s == TIME || s == DATE || s == DATETIME || s == GETLOCALE;
    }

    Locale locale;

    public LocaleAdapter(){
	this(Locale.getDefault());
    }

    public LocaleAdapter(Locale locale){
	this.locale = locale;
    }

    public Object get(String name){
	try {
	    return ResourceBundle.getBundle(name, locale);
	} catch (MissingResourceException e){
	    return null;
	}
    }

    public void set(String name, Object value){
	throw new RuntimeException("cannot set a resource bundle");
    }

    public Object invoke(String name, Object arg[]){
	try {
	    ResourceBundle b = ResourceBundle.getBundle(name, locale);
	    String key = (String)arg[0];
	    if (arg.length > 1){
		String s[] = new String[arg.length - 1];
		for (int i = 0; i < s.length; i++){
		    s[i] = arg[i + 1].toString();
		}
		return MessageFormat.format(b.getString(key), s);
	    } else {
		return b.getObject(key);
	    }
	} catch (MissingResourceException e){
	    if (isKeyword(name)){
		try {
		    return Runtime.callMethod(getClass(), name, arg, this);
		} catch (IllegalAccessException e1){
		    System.err.println(e1);
		} catch (IllegalArgumentException e2){
		    System.err.println(e2);
		} catch (InvocationTargetException e3){
		    Throwable t = e3.getTargetException();
		    if (t instanceof RuntimeException){
			throw (RuntimeException)t;
		    }
		    if (t instanceof Error){
			throw (Error)t;
		    }
		    throw new Error("never happen");
		}
	    }
	    return null;
	}
    }

    public Locale getLocale(){
	return locale;
    }

    static int getStyle(String str){
	String s = str.intern();
	if (s == FULL){
	    return DateFormat.FULL;
	} else if (s == LONG){
	    return DateFormat.LONG;
	} else if (s == MEDIUM){
	    return DateFormat.MEDIUM;
	} else if (s == SHORT){
	    return DateFormat.SHORT;
	} else if (s == DEFAULT){
	    return DateFormat.DEFAULT;
	} else {
	    return DateFormat.DEFAULT;
	}
    }

    NumberFormat getNumberFormat(int imin, int imax, int fmin, int fmax){
	NumberFormat fmt = NumberFormat.getNumberInstance(locale);
	setScale(fmt, imin, imax, fmin, fmax);
	return fmt;
    }

    NumberFormat getCurrencyFormat(int imin, int imax, int fmin, int fmax){
	NumberFormat fmt = NumberFormat.getCurrencyInstance(locale);
	setScale(fmt, imin, imax, fmin, fmax);
	return fmt;
    }

    NumberFormat getPercentFormat(int imin, int imax, int fmin, int fmax){
	NumberFormat fmt = NumberFormat.getPercentInstance(locale);
	setScale(fmt, imin, imax, fmin, fmax);
	return fmt;
    }

    static void setScale(NumberFormat fmt, int imin, int imax, int fmin, int fmax){
	if (imin >= 0){
	    fmt.setMinimumIntegerDigits(imin);
	}
	if (imax >= 0){
	    fmt.setMaximumIntegerDigits(imax);
	}
	if (fmin >= 0){
	    fmt.setMinimumFractionDigits(fmin);
	}
	if (fmax >= 0){
	    fmt.setMaximumFractionDigits(fmax);
	}
    }

    public String number(double num){
	return number(num, -1, -1, -1, -1);
    }

    public String number(long num){
	return number(num, -1, -1, -1, -1);
    }

    public String number(long num, int fmin){
	return number(num, -1, -1, fmin, -1);
    }

    public String number(double num, int fmin){
	return number(num, -1, -1, fmin, -1);
    }

    public String number(long num, int imin, int imax, int fmin, int fmax){
	return getNumberFormat(imin, imax, fmin, fmax).format(num);
    }

    public String number(double num, int imin, int imax, int fmin, int fmax){
	return getNumberFormat(imin, imax, fmin, fmax).format(num);
    }

    public String currency(double num){
	return getCurrencyFormat(-1, -1, -1, -1).format(num);
    }

    public String currency(long num){
	return getCurrencyFormat(-1, -1, -1, -1).format(num);
    }

    public String currency(double num, int fmin){
	return currency(num, -1, -1, fmin, -1);
    }

    public String currency(long num, int fmin){
	return currency(num, -1, -1, fmin, -1);
    }

    public String currency(long num, int imin, int imax, int fmin, int fmax){
	return getCurrencyFormat(imin, imax, fmin, fmax).format(num);
    }

    public String currency(double num, int imin, int imax, int fmin, int fmax){
	return getCurrencyFormat(imin, imax, fmin, fmax).format(num);
    }

    public String percent(double num){
	return percent(num, -1, -1, -1, -1);
    }

    public String percent(long num){
	return percent(num, -1, -1, -1, -1);
    }

    public String percent(double num, int fmin){
	return percent(num, -1, -1, fmin, -1);
    }

    public String percent(long num, int fmin){
	return percent(num, -1, -1, fmin, -1);
    }

    public String percent(double num, int imin, int imax, int fmin, int fmax){
	return getPercentFormat(imin, imax, fmin, fmax).format(num);
    }

    public String percent(long num, int imin, int imax, int fmin, int fmax){
	return getPercentFormat(imin, imax, fmin, fmax).format(num);
    }

    public String date(){
	return date(new Date());
    }

    public String date(Date d){
	return date(d, DEFAULT);
    }

    public String date(String style){
	return date(new Date(), style);
    }

    public String date(Date d, String style){
	return DateFormat.getDateInstance(getStyle(style), locale).format(d);
    }

    public String time(){
	return time(new Date());
    }

    public String time(Date d){
	return time(d, DEFAULT);
    }

    public String time(String style){
	return time(new Date(), style);
    }

    public String time(Date d, String style){
	return DateFormat.getTimeInstance(getStyle(style), locale).format(d);
    }

    public String datetime(){
	return datetime(new Date());
    }

    public String datetime(Date d){
	return datetime(d, DEFAULT, DEFAULT);
    }

    public String datetime(String style){
	return datetime(new Date(), style, style);
    }

    public String datetime(String style1, String style2){
	return datetime(new Date(), style1, style2);
    }

    public String datetime(Date d, String style1, String style2){
	return DateFormat.getDateTimeInstance(getStyle(style1),
					      getStyle(style2),
					      locale).format(d);
    }

    public String toString(){
	return getClass().getName() + "(" + locale.toString() + ")";
    }    
}
